﻿Imports Microsoft.DirectX
Imports Microsoft.DirectX.DirectDraw

Public Class DirectDrawEngine

    Public Interpolation As Boolean

    Private Dev As Device
    Private Target As Form
    Private BackgroundSurface As Surface
    Private BackgroundDesc As SurfaceDescription
    Private ForegroundSurface As Surface
    Private ForegroundDesc As SurfaceDescription
    Private Primary As Surface
    Private PrimaryDesc As SurfaceDescription
    Private PrimaryClipper As Clipper
    Private FullScreen As Boolean

    Public Sub New(ByVal TargetForm As Form, ByVal TargetFullScreen As Boolean)
        Target = TargetForm
        Dev = New Device(CreateFlags.Default)
        FullScreen = TargetFullScreen

        If (FullScreen) Then
            Dev.SetCooperativeLevel(Target, CooperativeLevelFlags.Fullscreen)
        Else
            Dev.SetCooperativeLevel(Target, CooperativeLevelFlags.Normal)
        End If

        PrimaryDesc = New SurfaceDescription
        With PrimaryDesc
            .SurfaceCaps.PrimarySurface = True
        End With
        Primary = New Surface(PrimaryDesc, Dev)

        If (Not FullScreen) Then
            PrimaryClipper = New Clipper(Dev)
            PrimaryClipper.Window = Target
            Primary.Clipper = PrimaryClipper
        End If

        BackgroundDesc = New SurfaceDescription
        With BackgroundDesc
            .SurfaceCaps.OffScreenPlain = True
        End With

        Resized()
    End Sub

    Private Sub RestoreSurfaces()
        Try
            Primary.Restore()
            BackgroundSurface.Restore()
        Catch ex As WrongModeException
            ' we are in a different mode, can't do
        End Try
    End Sub

    Public Sub Resized()
        Dim ClientRect As Rectangle

        BackgroundDesc = New SurfaceDescription
        With BackgroundDesc
            .SurfaceCaps.OffScreenPlain = True
        End With

        If (FullScreen) Then
            ClientRect = Screen.PrimaryScreen.Bounds
        Else
            ClientRect = Target.RectangleToScreen(Target.ClientRectangle)
        End If

        Try
        Catch ex As Exception
            RestoreSurfaces()
        End Try
    End Sub

    Public Sub Update(ByVal BackgroundData As Bitmap, ByVal Rect As Rectangle)
        Dim TargetRect As Rectangle
        Dim Effects As New DrawEffects
        Dim BackgroundGraphics As Graphics = Nothing
        Dim DE As New DrawEffects

        If (FullScreen) Then
            TargetRect = Screen.PrimaryScreen.Bounds
        Else
            TargetRect = Target.RectangleToScreen(Rect)
        End If

        If (BackgroundSurface IsNot Nothing) Then
            BackgroundSurface.Dispose()
            BackgroundSurface = Nothing
        End If

        Try
            BackgroundSurface = New Surface(BackgroundData, BackgroundDesc, Dev)
            Primary.Draw(TargetRect, BackgroundSurface, 0)
        Catch ex As UnsupportedModeException
            ' do nothing, something else is using the screen
        Catch ex As SurfaceLostException
            RestoreSurfaces()
        Catch ex As InvalidRectangleException
            ' this is OK, we just can't draw this time
        Catch ex As Exception
            ' generic catch-all
        End Try

        If (BackgroundSurface IsNot Nothing) Then
            BackgroundSurface.Dispose()
            BackgroundSurface = Nothing
        End If
    End Sub

    Protected Overrides Sub Finalize()
        MyBase.Finalize()
        Dev.Dispose()
        Dev = Nothing
    End Sub
End Class