﻿Imports System.IO

Public Class GameForm
    Public WithEvents Environment As Roton.Environment
    Private Video As Roton.VideoGDI
    Private Audio As Roton.AudioDirectX9
    Private HUDVideo As Roton.VideoGDI
    Private Buffer As Bitmap
    Private KeyboardInput As Roton.InputForm
    Private XScale As Integer
    Private YScale As Integer
    Private XZoom As Integer = 1
    Private YZoom As Integer = 1
    Private XCamera As Integer
    Private YCamera As Integer
    Private DisplaySize As Size
    Private HudWidth As Integer
    Private HudHeight As Integer
    Private Status As ToolStripStatusLabel
    Private CameraBoundWidth As Integer
    Private CameraBoundHeight As Integer
    Private Resized As Boolean
    Private HideSideBar As Boolean = False
    Private Cheat As String = ""
    Private WithEvents GameTimer As Roton.Timer
    Private DDraw As DirectDrawEngine
    Private EnableDD As Boolean
    Private RecordingDemo As Boolean
    Private PlayingDemo As Boolean
    Private DemoStream As Stream
    Private DemoRecorder As Roton.DemoRecorder
    Private DemoInput As Roton.DemoInput
    Private Action As PlayerAction = PlayerAction.Normal
    Private Loaded As Boolean
    Private FlagKeys As New List(Of FlagKey)

    Private Enum PlayerAction As Integer
        Normal
        Cheating
        QuittingGame
        QuittingZZT
        [Error]
    End Enum

    Public Class FlagKey
        Public Key As Keys
        Public Flag As String

        Public Sub New()
            Key = Keys.None
            Flag = ""
        End Sub

        Public Sub New(NewKey As Keys, NewFlag As String)
            Key = NewKey
            Flag = NewFlag
        End Sub
    End Class

    Public Sub New(ByVal NewEnvironment As Roton.Environment, ByVal Font() As Byte)
        ' This call is required by the designer.
        ' Add any initialization after the InitializeComponent() call.
        InitializeComponent()
        Environment = NewEnvironment

        ' load user settings
        LoadUserSettings()

        ' set gamespeed
        SetGameSpeed(4)

        ' the form controls are initialized
        Loaded = True

        ' set up environment
        InitEnvironment()

        ' set up audio
        Audio = New Roton.AudioDirectX9(Me)
        With Audio
            .LoadSoundEffectStream(0, My.Resources.drum0)
            .LoadSoundEffectStream(1, My.Resources.drum1)
            .LoadSoundEffectStream(2, My.Resources.drum2)
            .LoadSoundEffectStream(3, My.Resources.player)
            .LoadSoundEffectStream(4, My.Resources.drum4)
            .LoadSoundEffectStream(5, My.Resources.drum5)
            .LoadSoundEffectStream(6, My.Resources.drum6)
            .LoadSoundEffectStream(7, My.Resources.drum7)
            .LoadSoundEffectStream(8, My.Resources.drum8)
            .LoadSoundEffectStream(9, My.Resources.drum9)
        End With
        Environment.Audio = Audio

        ' set up video
        SetStyle(ControlStyles.AllPaintingInWmPaint, True)
        SetStyle(ControlStyles.UserPaint, True)
        SetStyle(ControlStyles.Opaque, True)
        Video = New Roton.VideoGDI
        Video.SetFont(Font)
        Video.SetViewport(Environment.Parameters.ViewWidth, Environment.Parameters.ViewHeight)
        Video.SetBrightMode(mnuViewBright.Checked)

        HudWidth = 20
        HudHeight = Environment.Parameters.ViewHeight
        HUDVideo = New Roton.VideoGDI
        HUDVideo.SetFont(Font)
        HUDVideo.SetSize(HudWidth, HudHeight)
        HUDVideo.SetViewport(HudWidth, HudHeight)

        Environment.Video = Video
        If Environment.Parameters.DisplayWide Then
            XScale = 2
            YScale = 1
        Else
            XScale = 1
            YScale = 1
        End If
        Buffer = New Bitmap((Video.Bitmap.Width * XScale) + HUDVideo.Bitmap.Width, Video.Bitmap.Height, Imaging.PixelFormat.Format32bppPArgb)
        DisplaySize = Buffer.Size
        UpdateSize()

        ' set up input
        KeyboardInput = New Roton.InputForm
        Environment.Input = KeyboardInput

        ' set up messenger
        Environment.Messenger = New Roton.MessengerVideo(Video, Environment.Parameters)

        ' set up scroll
        Environment.Scroll = New Roton.ScrollVideo(Video)

        ' set up camera stuff
        CameraBoundWidth = Video.ViewWidth \ 3
        CameraBoundHeight = Video.ViewHeight \ 3

        ' add help files
        AddHelpResource()

        ' set up debug menu
        BuildDebugBoardList()

        ' set up flag keys
        FlagKeys.Add(New FlagKey(Keys.I, "I"))
    End Sub

    Private Sub LoadUserSettings()
        mnuView2x.Checked = My.Settings.View2xZoom
        mnuViewBright.Checked = My.Settings.ViewBright
        mnuViewDirectDraw.Checked = My.Settings.ViewDirectDraw
        mnuViewInterpolated.Checked = My.Settings.ViewInterpolated
        mnuViewAutoMaximize.Checked = My.Settings.ViewAutoMaximize
        mnuOptionsAlerts.Checked = My.Settings.OptionsDisableAlerts
        mnuOptionsSounds.Checked = My.Settings.OptionsDisableAnnoyingSounds
    End Sub

    Private Sub SaveUserSettings()
        My.Settings.View2xZoom = mnuView2x.Checked
        My.Settings.ViewBright = mnuViewBright.Checked
        My.Settings.ViewDirectDraw = mnuViewDirectDraw.Checked
        My.Settings.ViewInterpolated = mnuViewInterpolated.Checked
        My.Settings.ViewAutoMaximize = mnuViewAutoMaximize.Checked
        My.Settings.OptionsDisableAlerts = mnuOptionsAlerts.Checked
        My.Settings.OptionsDisableAnnoyingSounds = mnuOptionsSounds.Checked
        My.Settings.Save()
    End Sub

    Private Sub BuildDebugBoardList()
        Dim Index As Integer

        mnuDebugBoards.DropDownItems.Clear()
        For Index = 0 To Environment.Boards.Count - 1
            mnuDebugBoards.DropDownItems.Add(Index & ": " & Environment.Boards(Index).Name)
        Next
    End Sub

    Private Sub UpdateTitle()
        Me.Text = Environment.World.Name & " - " & Environment.CurrentBoard.Name
    End Sub

    Private Sub BoardChangeHandler(B As Roton.Board)
        XCamera = Environment.PlayerX - (Video.ViewWidth \ 2)
        YCamera = Environment.PlayerY - (Video.ViewHeight \ 2)
        UpdateTitle()
    End Sub

    Private Sub WorldChangeHandler(W As Roton.World)
        BuildDebugBoardList()
        UpdateTitle()
    End Sub

    Private Sub ShowGameSpeed()
        mnuSpeed.ShowDropDown()
    End Sub

    Private Sub InitEnvironment()
        ' set up event handlers
        AddHandler Environment.UpdateStatus, AddressOf UpdateHud
        AddHandler Environment.RedrawStatus, AddressOf RedrawHud
        AddHandler Environment.QuitGame, AddressOf QuitGameHandler
        AddHandler Environment.QuitZZT, AddressOf QuitZZTHandler
        AddHandler Environment.EnterCheat, AddressOf BeginCheat
        AddHandler Environment.BoardChanged, AddressOf BoardChangeHandler
        AddHandler Environment.WorldChanged, AddressOf WorldChangeHandler
        AddHandler Environment.SetGameSpeed, AddressOf ShowGameSpeed
    End Sub

    Private Sub InitD3D(ByVal IsFullScreen As Boolean)
        EnableDD = True
        If (DDraw IsNot Nothing) Then
            DDraw = Nothing
        End If
        DDraw = New DirectDrawEngine(Me, False)
    End Sub

    Private Sub UnInitD3D()
        DDraw = Nothing
        EnableDD = False
    End Sub

    Private Sub UpdateD3D()
        Dim DisplayWidth As Integer
        Dim DisplayHeight As Integer
        Dim DisplayRatio As Double = ClientSize.Width / ClientSize.Height
        Dim ImageRatio As Double
        Dim ImageWidth As Integer
        Dim ImageHeight As Integer
        Dim DisplayX As Integer
        Dim DisplayY As Integer
        Dim TopMargin As Integer
        Dim LeftMargin As Integer
        Dim RightMargin As Integer
        Dim BottomMargin As Integer
        Dim Source As Bitmap
        Dim G As Graphics

        If (HideSideBar) Then
            Source = Video.Bitmap
            ImageWidth = Source.Width
            ImageHeight = Source.Height
            If (Environment.Parameters.DisplayWide) Then
                ImageWidth *= 2
            End If
        Else
            Source = Buffer
            ImageWidth = Source.Width
            ImageHeight = Source.Height
        End If

        ImageRatio = CDbl(ImageWidth) / CDbl(ImageHeight)

        If (WindowState = FormWindowState.Minimized) Then
            Exit Sub
        End If

        If (DisplayRatio > ImageRatio) Then
            DisplayHeight = ClientSize.Height
            DisplayWidth = CInt(CDbl(DisplayHeight) * ImageRatio)
        Else
            DisplayWidth = ClientSize.Width
            DisplayHeight = CInt(CDbl(DisplayWidth) / ImageRatio)
        End If

        DisplayX = (ClientSize.Width - DisplayWidth) \ 2
        DisplayY = (ClientSize.Height - DisplayHeight) \ 2

        TopMargin = DisplayY
        LeftMargin = DisplayX
        RightMargin = ClientRectangle.Width - (DisplayX + DisplayWidth)
        BottomMargin = ClientRectangle.Height - (DisplayY + DisplayHeight)

        ' draw black border around the outside
        G = Graphics.FromHwnd(Handle)
        G.PixelOffsetMode = Drawing2D.PixelOffsetMode.Half
        G.InterpolationMode = Drawing2D.InterpolationMode.NearestNeighbor
        G.FillRectangle(Brushes.Black, New Rectangle(0, 0, ClientSize.Width, TopMargin))
        G.FillRectangle(Brushes.Black, New Rectangle(0, TopMargin, LeftMargin, DisplayHeight))
        G.FillRectangle(Brushes.Black, New Rectangle((DisplayX + DisplayWidth), TopMargin, RightMargin, DisplayHeight))
        G.FillRectangle(Brushes.Black, New Rectangle(0, (TopMargin + DisplayHeight), ClientSize.Width, BottomMargin))

        If (DDraw Is Nothing) Then
            If (mnuViewInterpolated.Checked) Then
                G.InterpolationMode = Drawing2D.InterpolationMode.Low
            End If
            G.DrawImage(Source, DisplayX, DisplayY, DisplayWidth, DisplayHeight)
        Else
            DDraw.Update(Source, New Rectangle(DisplayX, DisplayY, DisplayWidth, DisplayHeight))
        End If
        G.Dispose()
    End Sub

    Private Sub ResizeD3D()
        Dim G As Graphics
        If (DDraw Is Nothing) Then
            G = Graphics.FromHwnd(Handle)
            G.FillRectangle(Brushes.Black, ClientRectangle)
            G.Dispose()
        Else
            DDraw.Resized()
        End If
    End Sub

    Private Sub SwitchDD()
        If (mnuViewDirectDraw.Checked) Then
            InitD3D(False)
        Else
            UnInitD3D()
        End If
    End Sub

    Private Sub AddHelpResource()
        Dim HelpResource As New IO.MemoryStream(My.Resources.ZZT)
        Environment.Cache.ImportHelpLibrary(HelpResource, HelpResource.Length)
        HelpResource.Dispose()
    End Sub

    Private Function HudNumber(ByVal InNumber As Integer) As String
        Dim Ret As String = InNumber & " "
        If (InNumber >= 0) Then
            Ret = " " & Ret
        End If
        Return Ret
    End Function

    Private Sub QuitGameHandler()
        If (Environment.World.Health > 0) Then
            Action = PlayerAction.QuittingGame
            QuitGameHud()
            HUDVideo.Update()
            DrawBuffer()
            Invalidate()
        Else
            'immediately quit if health = 0
            Environment.ConfirmQuitGame()
            Action = PlayerAction.Normal
        End If
    End Sub

    Private Sub QuitZZTHandler()
        Action = PlayerAction.QuittingZZT
        QuitZZTHud()
        HUDVideo.Update()
        DrawBuffer()
        Invalidate()
    End Sub

    Private Sub UpdateHud(ByVal W As Roton.World)
        Dim Index As Integer
        Dim StoneString As String = ""

        Select Case Environment.State.PlayerElement.ID
            Case Roton.ElementID.Player
                ' special counter
                If (W.Stones >= 0) Then
                    For Index = 0 To (W.Flags.Count - 1)
                        If (W.Flags(Index).StartsWith("Z")) Then
                            StoneString = W.Flags(Index).Substring(1) & ":"
                        End If
                    Next
                    HUDVideo.WriteOverlay(12 - StoneString.Length, 12, StoneString & W.Stones & " ", &H1E)
                Else
                    HUDVideo.WriteOverlay(2, 12, "                 ", &H1E)
                End If

                ' time
                If (Environment.CurrentBoard.TimeLimit > 0) Then
                    HUDVideo.WriteOverlay(7, 5, "Time:" & Environment.CurrentBoard.TimeLimit - W.TimePassed & " ", &H1E)
                Else
                    HUDVideo.WriteOverlay(7, 5, "           ", &H1E)
                End If

                ' keys
                For Index = 0 To 6
                    If (W.Keys(Index)) Then
                        HUDVideo.PlotOverlay(12 + Index, 11, New Roton.AnsiChar(12, &H10 Or (Index + 9)))
                    Else
                        HUDVideo.PlotOverlay(12 + Index, 11, New Roton.AnsiChar(32, &H10))
                    End If
                Next

                ' torch cycles
                If (W.TorchCycles > 0) Then
                    For Index = 2 To 5
                        If (((W.TorchCycles * 5) \ 200) >= Index) Then
                            HUDVideo.PlotOverlay(13 + Index, 8, New Roton.AnsiChar(177, &H16))
                        Else
                            HUDVideo.PlotOverlay(13 + Index, 8, New Roton.AnsiChar(176, &H16))
                        End If
                    Next
                Else
                    HUDVideo.WriteOverlay(15, 8, "    ", &H1C)
                End If

                ' stats
                HUDVideo.WriteOverlay(12, 6, W.Health & " ", &H1E)
                HUDVideo.WriteOverlay(12, 7, W.Ammo & " ", &H1E)
                HUDVideo.WriteOverlay(12, 8, W.Torches & " ", &H1E)
                HUDVideo.WriteOverlay(12, 9, W.Gems & " ", &H1E)
                HUDVideo.WriteOverlay(12, 10, W.Score & " ", &H1E)

                ' be quiet/noisy
                If (Environment.State.GameQuiet) Then
                    HUDVideo.WriteOverlay(6, 15, "Be noisy", &H1F)
                Else
                    HUDVideo.WriteOverlay(6, 15, "Be quiet", &H1F)
                End If

            Case Roton.ElementID.Monitor
                HUDVideo.WriteOverlay(9, 8, W.Name, &H1F)
                HUDVideo.WriteOverlay(6, 22, "          ", &H1F)
                HUDVideo.PlotOverlay(7 + Environment.State.GameSpeed, 22, New Roton.AnsiChar(31, &H1F))
        End Select

        If (Environment.State.GamePaused) Then
            HUDVideo.WriteOverlay(4, 4, "Pausing...", &H1F)
        End If

        HUDVideo.Update()
    End Sub

    Private Sub CheatHud()
        HUDVideo.WriteOverlay(2, 3, "            ", &HF)
        HUDVideo.WriteOverlay(2, 3, Cheat & "_", &HF)
    End Sub

    Private Sub QuitGameHud()
        HUDVideo.WriteOverlay(2, 4, "End this game? _", &H1F)
    End Sub

    Private Sub QuitZZTHud()
        HUDVideo.WriteOverlay(2, 5, "Quit ZZT? _      ", &H1F)
    End Sub

    Private Sub RedrawHud()
        Dim X As Integer
        Dim Y As Integer

        For X = 0 To HudWidth - 1
            For Y = 0 To HudHeight - 1
                HUDVideo.PlotOverlay(X, Y, New Roton.AnsiChar(32, &H1E))
            Next
        Next

        For X = 0 To 6 Step 2
            HUDVideo.PlotOverlay(9 - X, 0, New Roton.AnsiChar(196, &H1F))
            HUDVideo.PlotOverlay(10 + X, 0, New Roton.AnsiChar(196, &H1F))
            HUDVideo.PlotOverlay(9 - X, 2, New Roton.AnsiChar(196, &H1F))
            HUDVideo.PlotOverlay(10 + X, 2, New Roton.AnsiChar(196, &H1F))
        Next

        HUDVideo.WriteOverlay(2, 1, "      Lyon      ", &H70)

        Select Case Environment.State.PlayerElement.ID
            Case Roton.ElementID.Player
                ' text and values
                HUDVideo.WriteOverlay(5, 6, "Health:", &H1E)
                HUDVideo.WriteOverlay(7, 7, "Ammo:", &H1E)
                HUDVideo.WriteOverlay(4, 8, "Torches:", &H1E)
                HUDVideo.WriteOverlay(7, 9, "Gems:", &H1E)
                HUDVideo.WriteOverlay(6, 10, "Score:", &H1E)
                HUDVideo.WriteOverlay(7, 11, "Keys:", &H1E)

                HUDVideo.WriteOverlay(2, 15, " B ", &H30)
                HUDVideo.WriteOverlay(7, 18, " " & Chr(24) & Chr(25) & Chr(26) & Chr(27), &H30)
                HUDVideo.WriteOverlay(1, 19, " Shift " & Chr(24) & Chr(25) & Chr(26) & Chr(27), &H70)
                HUDVideo.WriteOverlay(2, 21, " S ", &H70)
                HUDVideo.WriteOverlay(2, 22, " P ", &H30)
                HUDVideo.WriteOverlay(2, 23, " Q ", &H70)

                If (Environment.Parameters.TorchesEnabled) Then
                    HUDVideo.WriteOverlay(2, 14, " T ", &H70)
                    HUDVideo.WriteOverlay(6, 14, "Torch", &H1F)
                End If

                HUDVideo.WriteOverlay(2, 16, " H ", &H70)
                If (Environment.Parameters.HintsEnabled) Then
                    HUDVideo.WriteOverlay(6, 16, "Hint", &H1F)
                Else
                    HUDVideo.WriteOverlay(6, 16, "Help", &H1F)
                End If
                HUDVideo.WriteOverlay(13, 18, "Move", &H1F)
                HUDVideo.WriteOverlay(13, 19, "Shoot", &H1F)
                HUDVideo.WriteOverlay(6, 21, "Save game", &H1F)
                HUDVideo.WriteOverlay(6, 22, "Pause", &H1F)
                HUDVideo.WriteOverlay(6, 23, "Quit", &H1F)

                HUDVideo.PlotOverlay(2, 6, New Roton.AnsiChar(2, &H1F))
                HUDVideo.PlotOverlay(2, 7, New Roton.AnsiChar(132, &H1B))
                HUDVideo.PlotOverlay(2, 8, New Roton.AnsiChar(157, &H16))
                HUDVideo.PlotOverlay(2, 9, New Roton.AnsiChar(4, &H1B))
                HUDVideo.PlotOverlay(2, 11, New Roton.AnsiChar(12, &H1F))
            Case Roton.ElementID.Monitor
                HUDVideo.WriteOverlay(2, 5, "Pick a command:", &H1B)

                HUDVideo.WriteOverlay(2, 7, " W ", &H30)
                HUDVideo.WriteOverlay(2, 11, " P ", &H70)
                HUDVideo.WriteOverlay(2, 12, " R ", &H30)
                HUDVideo.WriteOverlay(2, 13, " Q ", &H70)
                HUDVideo.WriteOverlay(2, 16, " A ", &H30)
                HUDVideo.WriteOverlay(2, 17, " - ", &H70)
                HUDVideo.WriteOverlay(2, 18, " - ", &H30)
                HUDVideo.WriteOverlay(2, 21, " S ", &H70)

                HUDVideo.WriteOverlay(6, 7, "World:", &H1E)
                HUDVideo.WriteOverlay(6, 11, "Play", &H1F)
                HUDVideo.WriteOverlay(6, 12, "Restore game", &H1E)
                HUDVideo.WriteOverlay(6, 13, "Quit", &H1E)
                HUDVideo.WriteOverlay(6, 16, "About", &H1E)
                HUDVideo.WriteOverlay(6, 17, "High Scores", &H18)
                HUDVideo.WriteOverlay(6, 18, "Board Editor", &H18)
                HUDVideo.WriteOverlay(6, 21, "Game speed:", &H1E)
                HUDVideo.WriteOverlay(6, 23, "F....:....S", &H1E)
        End Select
    End Sub

    Private Sub UpdateSize()
        If (Not Loaded) Then
            Exit Sub
        End If
        If (HideSideBar) Then
            DisplaySize = New Size(Video.Bitmap.Width, Video.Bitmap.Height)
            If (Environment.Parameters.DisplayWide) Then
                DisplaySize.Width *= 2
            End If
        Else
            DisplaySize = New Size(Buffer.Width, Buffer.Height)
        End If
        If (WindowState = FormWindowState.Normal) Then
            With DisplaySize
                ClientSize = New Size(.Width * XZoom, .Height * YZoom)
            End With
        End If
        ResizeD3D()
        Invalidate()
    End Sub

    Protected Overrides Sub OnPaint(ByVal e As System.Windows.Forms.PaintEventArgs)
        MyBase.OnPaint(e)
        UpdateD3D()
    End Sub

    Protected Overrides Sub OnPaintBackground(ByVal e As System.Windows.Forms.PaintEventArgs)
        'MyBase.OnPaintBackground(e)
    End Sub

    Protected Sub SaveScreenshot()
        Dim SFD As New SaveFileDialog
        With SFD
            .Filter = "PNG screenshots|*.png"
            If (.ShowDialog = Windows.Forms.DialogResult.OK) Then
                Buffer.Save(.FileName, Imaging.ImageFormat.Png)
            End If
        End With
        SFD.Dispose()
    End Sub

    Protected Overrides Function ProcessCmdKey(ByRef msg As System.Windows.Forms.Message, ByVal keyData As System.Windows.Forms.Keys) As Boolean
        Dim EventArgs As KeyEventArgs = New KeyEventArgs(keyData)
        KeyboardInput.KeyDown(EventArgs)
        Select Case keyData
            Case (Keys.Enter Or Keys.Control)
                If (WindowState = FormWindowState.Maximized) Then
                    WindowState = FormWindowState.Normal
                Else
                    WindowState = FormWindowState.Maximized
                End If
                Return True
            Case (Keys.Tab Or Keys.Control)
                ' let the main app take care of this key
                Return False
            Case Keys.F11
                ToggleSideBar()
                Return True
        End Select

        Select Case EventArgs.KeyCode
            Case Keys.Up, Keys.Down, Keys.Left, Keys.Right,
                Keys.ShiftKey, Keys.LShiftKey, Keys.RShiftKey, Keys.Tab
                'these keys should always be processed by the game window
                Return True
        End Select
        Return MyBase.ProcessCmdKey(msg, keyData)
    End Function

    Private Sub ToggleSideBar()
        HideSideBar = Not HideSideBar
        UpdateSize()
    End Sub

    Private Sub GameForm_Click(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.Click
        Focus()
    End Sub

    Private Sub GameForm_DoubleClick(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.DoubleClick
        SaveScreenshot()
    End Sub

    Private Sub GameForm_FormClosed(ByVal sender As Object, ByVal e As System.Windows.Forms.FormClosedEventArgs) Handles Me.FormClosed
        Environment = Nothing
        Video = Nothing
        Audio.Stop()
        Audio = Nothing
        KeyboardInput = Nothing
        DemoInput = Nothing
        If (DemoStream IsNot Nothing) Then
            DemoStream.Close()
            DemoStream = Nothing
        End If
    End Sub

    Private Sub GameForm_FormClosing(ByVal sender As Object, ByVal e As System.Windows.Forms.FormClosingEventArgs) Handles Me.FormClosing
        MainForm.ActiveEnvironment = Nothing
    End Sub

    Private Sub GameForm_GotFocus(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.GotFocus
        MainForm.ActiveEnvironment = Environment
        ResizeD3D()
    End Sub

    Private Sub GameForm_KeyDown(ByVal sender As Object, ByVal e As System.Windows.Forms.KeyEventArgs) Handles Me.KeyDown
        Dim Index As Integer
        Dim Pressed As Boolean

        Select Case Action
            Case PlayerAction.Normal
                For Index = 0 To FlagKeys.Count - 1
                    With FlagKeys(Index)
                        If (.Key = e.KeyCode) Then
                            Pressed = True
                            If (Environment.FindFlag(.Flag)) Then
                                KeyboardInput.Cheat = "-" & .Flag
                            ElseIf (Environment.World.Flags.Count < Environment.Parameters.MaxFlags) Then
                                KeyboardInput.Cheat = "+" & .Flag
                            End If
                        End If
                    End With
                Next
                If (Not Pressed) Then
                    KeyboardInput.KeyDown(e)
                End If
            Case PlayerAction.Cheating
            Case PlayerAction.QuittingGame
            Case PlayerAction.QuittingZZT
        End Select
    End Sub

    Private Sub BeginCheat()
        Action = PlayerAction.Cheating
        CheatHud()
        HUDVideo.Update()
        DrawBuffer()
        Invalidate()
    End Sub

    Private Sub GameForm_KeyPress(ByVal sender As Object, ByVal e As System.Windows.Forms.KeyPressEventArgs) Handles Me.KeyPress
        Select Case Action
            Case PlayerAction.Normal
            Case PlayerAction.Cheating
                Select Case e.KeyChar
                    Case Chr(8)
                        'backspace
                        If (Cheat.Length > 0) Then
                            Cheat = Cheat.Substring(0, Cheat.Length - 1)
                        End If
                    Case Chr(13)
                        'enter
                        Action = PlayerAction.Normal
                        Select Case Cheat.ToUpper
                            Case "DEBUG"
                                mnuDebug.Visible = True
                            Case Else
                                KeyboardInput.Cheat = Cheat
                        End Select
                        Cheat = ""
                    Case Chr(27)
                        'escape
                        Action = PlayerAction.Normal
                        Cheat = ""
                    Case Else
                        Cheat &= e.KeyChar
                End Select
                If (Cheat.Length > 11) Then
                    Cheat = Cheat.Substring(0, 11)
                End If
                If (Action = PlayerAction.Cheating) Then
                    CheatHud()
                Else
                    RedrawHud()
                    UpdateHud(Environment.World)
                End If
                HUDVideo.Update()
                DrawBuffer()
                Invalidate()
            Case PlayerAction.QuittingGame
                Select Case e.KeyChar.ToString.ToUpper
                    Case "Y"
                        Environment.ConfirmQuitGame()
                        Action = PlayerAction.Normal
                    Case "N", Chr(27)
                        Action = PlayerAction.Normal
                End Select
                RedrawHud()
                UpdateHud(Environment.World)
                HUDVideo.Update()
                DrawBuffer()
                Invalidate()
            Case PlayerAction.QuittingZZT
                Select Case e.KeyChar.ToString.ToUpper
                    Case "Y"
                        Environment.ConfirmQuitZZT()
                        Action = PlayerAction.Normal
                    Case "N", Chr(27)
                        Action = PlayerAction.Normal
                End Select
                RedrawHud()
                UpdateHud(Environment.World)
                HUDVideo.Update()
                DrawBuffer()
                Invalidate()
        End Select
    End Sub

    Private Sub GameForm_KeyUp(ByVal sender As Object, ByVal e As System.Windows.Forms.KeyEventArgs) Handles Me.KeyUp
        KeyboardInput.KeyUp(e)
    End Sub

    Private Sub UpdateCamera()
        Dim CameraBoundLeft As Integer
        Dim CameraBoundRight As Integer
        Dim CameraBoundTop As Integer
        Dim CameraBoundBottom As Integer

        CameraBoundLeft = XCamera + CameraBoundWidth
        CameraBoundTop = YCamera + CameraBoundHeight
        CameraBoundRight = ((XCamera + Video.ViewWidth) - CameraBoundWidth) - 1
        CameraBoundBottom = ((YCamera + Video.ViewHeight) - CameraBoundHeight) - 1

        If (Environment.PlayerX < CameraBoundLeft) Then
            XCamera = Environment.PlayerX - CameraBoundWidth
        End If
        If (Environment.PlayerX > CameraBoundRight) Then
            XCamera = ((Environment.PlayerX - Video.ViewWidth) + CameraBoundWidth) + 1
        End If
        If (Environment.PlayerY < CameraBoundTop) Then
            YCamera = Environment.PlayerY - CameraBoundHeight
        End If
        If (Environment.PlayerY > CameraBoundBottom) Then
            YCamera = ((Environment.PlayerY - Video.ViewHeight) + CameraBoundHeight) + 1
        End If

        Video.SetCamera(XCamera, YCamera)
    End Sub

    Private Sub GameTimerTick(ByVal e As System.EventArgs)
        Tick()
    End Sub

    Private Sub DrawBuffer()
        Dim G As Graphics = Graphics.FromImage(Buffer)
        G.InterpolationMode = Drawing2D.InterpolationMode.NearestNeighbor
        G.PixelOffsetMode = Drawing2D.PixelOffsetMode.Half
        G.DrawImage(Video.Bitmap, 0, 0, Video.Bitmap.Width * XScale, Video.Bitmap.Height * YScale)
        G.DrawImageUnscaled(HUDVideo.Bitmap, Video.Bitmap.Width * XScale, 0)
        G.Dispose()
    End Sub

    Private Sub Tick()
        Dim Updated As Boolean

        If (Not Focused) AndAlso (Not mnuOptionsBackground.Checked) Then
            Exit Sub
        End If

        If (Action = PlayerAction.Normal) AndAlso (Not (Environment Is Nothing)) Then
            If (Debugger.IsAttached) Then
                Updated = Environment.Execute
            Else
                Try
                    Updated = Environment.Execute
                Catch ex As Exception
                    Action = PlayerAction.Error
                    If MsgBox("Game: " & Environment.World.Name & vbCrLf &
                           vbCrLf &
                           ex.Message & vbCrLf &
                           vbCrLf &
                           "Would you like to copy the stack trace to your clipboard?",
                           MsgBoxStyle.YesNo,
                           ex.GetType.ToString) = MsgBoxResult.Yes Then
                        Clipboard.Clear()
                        Clipboard.SetText(ex.GetType.ToString & vbCrLf & ex.StackTrace)
                    End If
                End Try
            End If
            If (Not (Environment Is Nothing)) Then
                If (Updated) Then
                    If (PlayingDemo) Then
                        HUDVideo.WriteOverlay(1, 3, DemoStream.Position & "/" & DemoStream.Length, &H30)
                        HUDVideo.Update()
                    End If
                    UpdateCamera()
                    Video.Update()
                    DrawBuffer()
                    Invalidate(ClientRectangle)
                End If
            End If
        End If
        If (Environment Is Nothing) OrElse (Environment.State.QuitZZT) Then
            Close()
        End If
    End Sub

    Private Sub SetGameSpeed(ByVal Index As Integer)
        Dim ItemIndex As Integer
        Dim X As ToolStripMenuItem

        If (Environment IsNot Nothing) Then
            Environment.State.GameSpeed = Index
        End If

        For ItemIndex = 0 To mnuSpeed.DropDownItems.Count - 1
            X = mnuSpeed.DropDownItems(ItemIndex)
            X.Checked = (ItemIndex = Index)
        Next

        If (Loaded) Then
            UpdateHud(Environment.World)
        End If
    End Sub

    Private Sub mnuSpeed0_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuSpeed0.Click
        SetGameSpeed(0)
    End Sub

    Private Sub mnuSpeed1_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuSpeed1.Click
        SetGameSpeed(1)
    End Sub

    Private Sub mnuSpeed2_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuSpeed2.Click
        SetGameSpeed(2)
    End Sub

    Private Sub mnuSpeed3_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuSpeed3.Click
        SetGameSpeed(3)
    End Sub

    Private Sub mnuSpeed4_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuSpeed4.Click
        SetGameSpeed(4)
    End Sub

    Private Sub mnuSpeed5_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuSpeed5.Click
        SetGameSpeed(5)
    End Sub

    Private Sub mnuSpeed6_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuSpeed6.Click
        SetGameSpeed(6)
    End Sub

    Private Sub mnuSpeed7_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuSpeed7.Click
        SetGameSpeed(7)
    End Sub

    Private Sub mnuSpeed8_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuSpeed8.Click
        SetGameSpeed(8)
    End Sub

    Private Sub mnuView2x_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuView2x.Click
        If (mnuView2x.Checked) Then
            XZoom = 2
            YZoom = 2
        Else
            XZoom = 1
            YZoom = 1
        End If
        UpdateSize()
    End Sub

    Private Sub mnuViewResetSize_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuViewResetSize.Click
        UpdateSize()
        Resized = False
    End Sub

    Private Sub GameForm_Load(sender As Object, e As System.EventArgs) Handles Me.Load
        Icon = My.Resources.Lyon
    End Sub

    Private Sub GameForm_Resize(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.Resize
        mnuViewResetSize.Enabled = (WindowState = FormWindowState.Normal)
        mnuView2x.Enabled = (WindowState = FormWindowState.Normal)
        UpdateSize()
    End Sub

    Private Sub InitTimer(TPS As Double)
        If (GameTimer IsNot Nothing) Then
            RemoveHandler GameTimer.Tick, AddressOf GameTimerTick
            GameTimer = Nothing
        End If
        GameTimer = New Roton.Timer(1000000.0# / TPS)
        AddHandler GameTimer.Tick, AddressOf GameTimerTick
    End Sub

    Private Sub GameForm_Shown(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.Shown
        InitTimer(71.2#)
        tmrTick.Enabled = True
        If (mnuViewAutoMaximize.Checked) Then
            WindowState = FormWindowState.Maximized
        End If
        SwitchDD()
    End Sub

    Private Sub tmrTick_Tick(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles tmrTick.Tick
        GameTimer.Proc()
    End Sub

    Protected Overrides Sub Finalize()
        UnInitD3D()
        MyBase.Finalize()
    End Sub

    Public Sub RecordDemo()
        Dim DemoWriter As BinaryWriter
        If (Not RecordingDemo) AndAlso (Not PlayingDemo) Then
            DemoStream = New MemoryStream
            DemoWriter = New BinaryWriter(DemoStream)
            DemoWriter.Write(CType(Environment.RNGSeed, Int32))
            DemoRecorder = New Roton.DemoRecorder(KeyboardInput, DemoStream)
            Environment.Input = DemoRecorder
            RecordingDemo = True
        End If
    End Sub

    Public Sub PlayDemo(ByVal SourceStream As Stream)
        Dim DemoReader As BinaryReader
        If (Not RecordingDemo) AndAlso (Not PlayingDemo) Then
            DemoStream = SourceStream
            DemoReader = New BinaryReader(DemoStream)
            Environment.RNGSeed = DemoReader.ReadInt32
            DemoInput = New Roton.DemoInput(DemoStream, KeyboardInput)
            Environment.Input = DemoInput
            PlayingDemo = True
        End If
    End Sub

    Public Sub StopDemo()
        If (RecordingDemo) Then
            DemoRecorder.Close()
            DemoStream.Flush()
            RecordingDemo = False
        ElseIf (PlayingDemo) Then
            PlayingDemo = False
            DemoStream.Close()
            DemoStream = Nothing
        End If
        Environment.Input = KeyboardInput
    End Sub

    Public Function ExportDemo() As Byte()
        Dim Ret() As Byte
        DemoStream.Position = 0
        ReDim Ret(DemoStream.Length - 1)
        DemoStream.Read(Ret, 0, DemoStream.Length)
        Return Ret
    End Function

    Private Sub mnuViewDirectDraw_Click(sender As System.Object, e As System.EventArgs) Handles mnuViewDirectDraw.Click
        SwitchDD()
    End Sub

    Private Sub mnuDebugBoards_DropDownItemClicked(sender As Object, e As System.Windows.Forms.ToolStripItemClickedEventArgs) Handles mnuDebugBoards.DropDownItemClicked
        Dim Index As Integer

        Index = mnuDebugBoards.DropDownItems.IndexOf(e.ClickedItem)
        Environment.EditorSetBoard(Index)
    End Sub

    Private Sub mnuViewBright_Click(sender As System.Object, e As System.EventArgs) Handles mnuViewBright.Click
        Video.SetBrightMode(mnuViewBright.Checked)
    End Sub

    Private Sub mnuOptionsSaveSettings_Click(sender As System.Object, e As System.EventArgs) Handles mnuOptionsSaveSettings.Click
        SaveUserSettings()
    End Sub

    Private Sub mnuViewInterpolated_Click(sender As System.Object, e As System.EventArgs) Handles mnuViewInterpolated.Click
        If (DDraw IsNot Nothing) Then
            DDraw.Interpolation = mnuViewInterpolated.Checked
        End If
    End Sub

    Private Sub mnuOptionsAlerts_Click(sender As System.Object, e As System.EventArgs) Handles mnuOptionsAlerts.Click
        If (mnuOptionsAlerts.Checked) Then
            Environment.Alerts.Disable()
        End If
    End Sub

    Private Sub mnuOptionsSounds_Click(sender As System.Object, e As System.EventArgs) Handles mnuOptionsSounds.Click
        If (mnuOptionsSounds.Checked) Then
            With Environment.Parameters
                .WaterMessage = ""
            End With
            With Environment.SoundBank
                .Water = {}
            End With
        End If
    End Sub

    Private Sub SuperspeedToolStripMenuItem_Click(sender As System.Object, e As System.EventArgs) Handles SuperspeedToolStripMenuItem.Click
        InitTimer(1000000)
        SetGameSpeed(0)
    End Sub
End Class
