﻿Public Class MainForm
    Private WithEvents _ActiveEnvironment As Roton.Environment
    Private _DefaultFont() As Byte = My.Resources.vga_rom
    Private SFDOpen As Boolean
    Private OFDOpen As Boolean
    Private Z2WebSite As String = "http://zzt.org"
    Private Z2ThreadSite As String = Z2WebSite & "/fora/viewtopic.php?f=9&t=3320"

    Public Property ActiveEnvironment As Roton.Environment
        Get
            Return _ActiveEnvironment
        End Get
        Set(ByVal value As Roton.Environment)
            _ActiveEnvironment = value
            mnuFileSave.Enabled = Not (_ActiveEnvironment Is Nothing)
        End Set
    End Property

    Private Sub mnuFileOpen_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuFileOpen.Click
        Dim Env As Roton.Environment = LoadWorld(1)

        If (Not (Env Is Nothing)) Then
            StartGame(Env, OFDWorld.FileName).Show()
        End If
    End Sub

    Private Sub mnuFileExit_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuFileExit.Click
        Application.Exit()
    End Sub

    Private Sub MainForm_DragDrop(ByVal sender As Object, ByVal e As System.Windows.Forms.DragEventArgs) Handles Me.DragDrop
        Dim Files() As String = e.Data.GetData("FileDrop")
        Dim Env As New List(Of Roton.Environment)
        Dim KS As Integer = e.KeyState
        Dim Index As Integer
        Dim Forms As New List(Of Form)

        ' ctrl=8
        ' alt=32
        ' shift=4

        For Each File In Files
            Env.Add(LoadWorldFile(File))
        Next

        ' if ctrl is pressed, open in editor
        If ((e.KeyState And 8) <> 0) Then
            For Index = 0 To Env.Count - 1
                If (Not (Env(Index) Is Nothing)) Then
                    'StartEditor(Env(Index), Files(Index)).Show()
                End If
            Next
        Else
            For Index = 0 To Env.Count - 1
                If (Not (Env(Index) Is Nothing)) Then
                    StartGame(Env(Index), Files(Index))
                End If
            Next
        End If
    End Sub

    Private Sub MainForm_DragOver(ByVal sender As Object, ByVal e As System.Windows.Forms.DragEventArgs) Handles Me.DragOver
        e.Effect = DragDropEffects.All
    End Sub

    Private Sub MainForm_Load(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.Load
        ActiveEnvironment = Nothing
        Icon = My.Resources.Lyon
    End Sub

    Private Sub mnuFileSave_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuFileSave.Click
        FileSave(_ActiveEnvironment)
    End Sub

    Private Sub FileSave(ByVal E As Roton.Environment)
        Dim FileName As String
        Dim FileStream As IO.FileStream

        If (SFDOpen) Then
            Exit Sub
        End If
        SFDOpen = True

        With E
            If (Not .State.EditorMode) AndAlso (Not .State.TitleScreen) Then
                .State.GamePaused = True
            End If
            If (.World.SavedGame) Then
                SFDWorld.Filter = "Saved games|*.sav|All files|*.*"
                SFDWorld.DefaultExt = "sav"
            Else
                Select Case (.Type)
                    Case Roton.EnvironmentGameType.ZZT
                        SFDWorld.Filter = "ZZT worlds|*.zzt|All files|*.*"
                        SFDWorld.DefaultExt = "zzt"
                    Case Roton.EnvironmentGameType.SuperZZT
                        SFDWorld.Filter = "Super ZZT worlds|*.szt|All files|*.*"
                        SFDWorld.DefaultExt = "szt"
                    Case Else
                        SFDWorld.Filter = "All files|*.*"
                        SFDWorld.DefaultExt = ""
                End Select
            End If
        End With

        With SFDWorld
            If (.ShowDialog = Windows.Forms.DialogResult.OK) Then
                FileName = .FileName
                FileStream = New IO.FileStream(FileName, IO.FileMode.Create, IO.FileAccess.Write, IO.FileShare.Read)
                Roton.File.Save(FileStream, E)
                FileStream.Dispose()
            End If
        End With

        SFDOpen = False
    End Sub

    Private Function GetFont(ByVal FileName As String) As Byte()
        Dim BaseName As String
        Dim CharExtensions() As String = {".com", ".chr"}
        Dim Ret() As Byte = {}
        Dim Index As Integer
        Dim Source As IO.Stream

        With My.Computer.FileSystem
            BaseName = FileName
            If (.GetName(BaseName.Contains("."))) Then
                BaseName = BaseName.Substring(0, BaseName.LastIndexOf("."))
            End If

            For Index = 0 To CharExtensions.Length - 1
                If (.FileExists(BaseName & CharExtensions(Index))) Then
                    Source = New IO.FileStream(BaseName & CharExtensions(Index), IO.FileMode.Open, IO.FileAccess.Read, IO.FileShare.Read)
                    Ret = Roton.File.LoadCharacterFile(Source, Source.Length)
                    Source.Dispose()
                    If (Ret.Length > 0) Then
                        Exit For
                    End If
                End If
            Next
        End With

        If (Ret.Length = 0) Then
            Ret = _DefaultFont
        End If

        Return Ret
    End Function

    Private Function LoadWorldGetWorlds(ByVal SourceList() As String)
        Dim Extensions() As String = Roton.EnvironmentParameters.EnumerateWorldExtensions
        Dim ThisFile As String
        Dim Index As Integer
        Dim ExtensionIndex As Integer
        Dim Ret As New List(Of String)

        For Index = 0 To SourceList.Length - 1
            ThisFile = SourceList(Index).ToUpper
            For ExtensionIndex = 0 To Extensions.Length - 1
                If (ThisFile.EndsWith("." & Extensions(ExtensionIndex).ToUpper)) Then
                    Ret.Add(ThisFile)
                    Exit For
                End If
            Next
        Next

        Ret.Sort()
        Return Ret.ToArray
    End Function

    Private Function LoadWorld(ExtensionIndex As Integer) As Roton.Environment
        Dim Env As Roton.Environment = Nothing

        If (OFDOpen) Then
            Return Nothing
        End If
        OFDOpen = True

        With OFDWorld
            .FilterIndex = ExtensionIndex
            If (.ShowDialog = Windows.Forms.DialogResult.OK) Then
                Env = LoadWorldFile(.FileName)
            End If
        End With

        OFDOpen = False
        Return Env
    End Function

    Private Function LoadWorldFile(FileName As String) As Roton.Environment
        Dim FileStream As IO.FileStream
        Dim Env As Roton.Environment = Nothing
        Dim WorldFiles() As String
        Dim Cache As New Roton.FileCache

        Cache.Path = My.Computer.FileSystem.GetFileInfo(FileName).DirectoryName

        FileStream = New IO.FileStream(FileName, IO.FileMode.Open, IO.FileAccess.Read, IO.FileShare.Read)
        If (FileName.ToUpper.EndsWith(".ZIP")) Then
            Cache.ImportZipArchive(FileStream)
            WorldFiles = LoadWorldGetWorlds(Cache.FileList)
            If (WorldFiles.Length > 1) Then
                If (MultiWorld.ShowDialog(WorldFiles) = DialogResult.OK) Then
                    FileName = WorldFiles(MultiWorld.SelectedIndex)
                Else
                    FileName = ""
                End If
            ElseIf (WorldFiles.Length = 1) Then
                FileName = WorldFiles(0)
            Else
                FileName = ""
            End If
            If (FileName <> "") Then
                Env = Roton.File.Load(Cache.LoadFileStream(FileName))
            End If
        Else
            Env = Roton.File.Load(FileStream)
        End If

        If (Env IsNot Nothing) Then
            Env.Cache = Cache
        End If

        FileStream.Dispose()

        Return Env
    End Function

    Private Function StartGame(ByVal Env As Roton.Environment, ByVal FileName As String) As Form
        Dim Game As Form

        AddHandler Env.LoadGame, AddressOf LoadGameHandler
        AddHandler Env.LoadWorld, AddressOf LoadWorldHandler
        AddHandler Env.SaveGame, AddressOf SaveGameHandler
        AddHandler Env.SaveWorld, AddressOf SaveWorldHandler

        Game = New GameForm(Env, GetFont(FileName))
        Game.MdiParent = Me
        Game.Show()

        Return Game
    End Function

    Private Function StartEditor(ByVal Env As Roton.Environment, ByVal FileName As String) As Form
        'Dim Editor As Form

        'Editor = New EditorForm(Env, GetFont(FileName), New ToolStripStatusLabel, mnuSettingsDirectDraw.Checked)
        'Editor.MdiParent = Me
        'Return Editor

        Return Nothing
    End Function

    Private Sub mnuFileEdit_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuFileEdit.Click
        Dim Env As Roton.Environment = LoadWorld(1)

        If (Not (Env Is Nothing)) Then
            'StartEditor(Env, OFDWorld.FileName).Show()
        End If
    End Sub

    Private Sub mnuFileNewZZT_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuFileNewZZT.Click
        'StartEditor(Roton.File.Create(Roton.EnvironmentGameType.ZZT), "").Show()
    End Sub

    Private Sub mnuFileNewSuperZZT_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuFileNewSuperZZT.Click
        'StartEditor(Roton.File.Create(Roton.EnvironmentGameType.SuperZZT), "").Show()
    End Sub

    Private Sub mnuArrangeTileH_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuArrangeH.Click
        LayoutMdi(MdiLayout.TileHorizontal)
    End Sub

    Private Sub mnuArrangeTileV_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuArrangeV.Click
        LayoutMdi(MdiLayout.TileVertical)
    End Sub

    Private Sub mnuArrangeCascade_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuArrangeCascade.Click
        LayoutMdi(MdiLayout.Cascade)
    End Sub

    Private Sub mnuFileRestore_Click(sender As System.Object, e As System.EventArgs) Handles mnuFileRestore.Click
        Dim Env As Roton.Environment = LoadWorld(2)

        If (Not (Env Is Nothing)) Then
            StartGame(Env, OFDWorld.FileName)
        End If
    End Sub

    Protected Overrides Function ProcessCmdKey(ByRef msg As System.Windows.Forms.Message, ByVal keyData As System.Windows.Forms.Keys) As Boolean
        Select Case keyData
            Case (Keys.Enter Or Keys.Alt)
                If (WindowState = FormWindowState.Maximized) Then
                    WindowState = FormWindowState.Normal
                Else
                    WindowState = FormWindowState.Maximized
                End If
                Return True
            Case (Keys.Control Or Keys.Tab)
                Return MyBase.ProcessCmdKey(msg, Keys.Tab)
            Case Keys.F12
                MenuStrip1.Visible = Not MenuStrip1.Visible
                Return True
        End Select
        Return MyBase.ProcessCmdKey(msg, keyData)
    End Function

    Sub SaveGameHandler(ByVal E As Roton.Environment)
        FileSave(E)
    End Sub

    Sub SaveWorldHandler(ByVal E As Roton.Environment)
        FileSave(E)
    End Sub

    Sub LoadGameHandler(ByVal E As Roton.Environment)
        Dim Ret As Roton.Environment
        Ret = LoadWorld(3)
        If (Ret IsNot Nothing) Then
            E.ApplyEnvironment(Ret)
        End If
    End Sub

    Sub LoadWorldHandler(ByVal E As Roton.Environment)
        Dim Ret As Roton.Environment
        Ret = LoadWorld(1)
        If (Ret IsNot Nothing) Then
            E.ApplyEnvironment(Ret)
        End If
    End Sub

    Private Sub MainForm_Shown(sender As Object, e As System.EventArgs) Handles Me.Shown
        Dim Args = My.Application.CommandLineArgs
        Dim Index As Integer
        Dim StartupWorlds As New List(Of Roton.Environment)
        Dim StartupFiles As New List(Of String)
        Dim Env As Roton.Environment

        For Index = 0 To Args.Count - 1
            Env = LoadWorldFile(Args(Index))
            If (Env IsNot Nothing) Then
                StartupWorlds.Add(Env)
                StartupFiles.Add(Args(Index))
            End If
        Next

        For Index = 0 To StartupWorlds.Count - 1
            StartGame(StartupWorlds(Index), StartupFiles(Index)).Show()
        Next
    End Sub

    Private Sub mnuHelpAbout_Click(sender As System.Object, e As System.EventArgs) Handles mnuHelpAbout.Click
        Dim AboutMessage As String =
            Application.ProductName & " is a ZZT/SuperZZT simulator." & vbCrLf & vbCrLf &
            "Code by SaxxonPike. (saxxonpike@gmail.com)" & vbCrLf &
            "For more information, check out http://www.zzt.org"
        MsgBox(AboutMessage, MsgBoxStyle.Information, "About " & Application.ProductName)
    End Sub

    Private Sub mnuHelpZ2Forum_Click(sender As System.Object, e As System.EventArgs) Handles mnuHelpZ2Forum.Click
        System.Diagnostics.Process.Start(Z2ThreadSite)
    End Sub

    Private Sub mnuHelpWebsite_Click(sender As System.Object, e As System.EventArgs) Handles mnuHelpWebsite.Click
        System.Diagnostics.Process.Start(Z2WebSite)
    End Sub

End Class