﻿' RotonAudioDirectX9
'
' A built-in audio plugin for Roton that uses DirectX 9.

Imports Microsoft.DirectX.DirectSound
Imports System.IO
Imports System.Windows.Forms

Public Class AudioDirectX9
    Implements Audio

    Private Dev As Device
    Private Desc As BufferDescription
    Private Sounds(9) As Buffer
    Private Tone As Buffer
    Private Played As Buffer
    Private PlayingTone As Boolean

    Public Sub New(ByVal OwnerForm As Form)
        Dev = New Device
        Dev.SetCooperativeLevel(OwnerForm, CooperativeLevel.Normal)
        Desc = New BufferDescription
        With Desc
            .CanGetCurrentPosition = False
            .Control3D = False
            .ControlFrequency = False
            .ControlPan = False
            .ControlPositionNotify = False
            .ControlVolume = True
            .LocateInSoftware = True
        End With
    End Sub

    Protected Overrides Sub Finalize()
        MyBase.Finalize()
        Dev.Dispose()
        Dev = Nothing
    End Sub

    Public Sub PlaySound(ByVal SoundIndex As Integer) Implements Audio.PlaySound
        [Stop]()
        Played = Sounds(SoundIndex)
        Played.Play(0, BufferPlayFlags.Default)
    End Sub

    Public Sub PlayTone(ByVal Tone As Integer) Implements Audio.PlayTone
        Dim Mem As New MemoryStream
        Dim Writer As New BinaryWriter(Mem)
        Dim I32 As Int32
        Dim I16 As Int16
        Dim HighSample As Byte = &HC0
        Dim LowSample As Byte = &H40
        Dim Freq As Int32 = 44100
        Dim Octave As Integer = (Tone >> 4)
        Dim Note As Integer = (Tone And &HF) + (Octave * 12)
        Dim Length As Integer = CInt(Freq / (440 * (2 ^ ((Note - 45) / 12))))
        Dim PeakLength As Integer = Length >> 1
        Dim Wave(Length) As Byte
        Dim Index As Integer

        For Index = 0 To PeakLength - 1
            Wave(Index) = HighSample
        Next

        For Index = PeakLength To Length - 1
            Wave(Index) = LowSample
        Next

        I32 = &H46464952
        Writer.Write(I32)   '"RIFF"
        I32 = Length + 36
        Writer.Write(I32)   'RIFF length
        I32 = &H45564157
        Writer.Write(I32)   '"WAVE"
        I32 = &H20746D66
        Writer.Write(I32)   '"fmt "
        I32 = &H10
        Writer.Write(I32)   'format chunk length
        I16 = 1
        Writer.Write(I16)   'format
        I16 = 1
        Writer.Write(I16)   'channels
        Writer.Write(Freq)  'samples / sec
        Writer.Write(Freq)  'bytes / sec
        I16 = 1
        Writer.Write(I16)   'bytes / sample
        I16 = 8
        Writer.Write(I16)   'bits / sample
        I32 = &H61746164
        Writer.Write(I32)   '"data"
        I32 = Length
        Writer.Write(I32)   'data length
        Writer.Write(Wave)
        Mem.Position = 0
        [Stop]()
        Played = New Buffer(Mem, Desc, Dev)
        Played.Volume = -600
        Played.Play(0, BufferPlayFlags.Looping)
        PlayingTone = True
        Mem.Dispose()
    End Sub

    Public WriteOnly Property SoundEffect(ByVal Index As Integer) As Byte() Implements Audio.SoundEffect
        Set(ByVal value As Byte())
            Dim Mem As New MemoryStream(Index)
            Sounds(Index) = New Buffer(Mem, Desc, Dev)
            Sounds(Index).Volume = -600
            Mem.Dispose()
        End Set
    End Property

    Public Sub LoadSoundEffectStream(ByVal Index As Integer, ByVal SourceStream As Stream)
        Sounds(Index) = New Buffer(SourceStream, Desc, Dev)
    End Sub

    Public Sub [Stop]() Implements Audio.Stop
        If Not (Played Is Nothing) Then
            Played.Stop()
            If (PlayingTone) Then
                Played.Dispose()
                PlayingTone = False
            End If
            Played = Nothing
        End If
    End Sub
End Class
