﻿' RotonBehavior
'
' Contains common code for actions to be taken by action and interaction code.

Imports System.Windows.Forms

Partial Public Class Environment

    ' dseg:0062
    Private Shared ReadOnly XVectorList4() As Integer = {
        0, 0, -1, 1
    }

    ' dseg:006A
    Private Shared ReadOnly YVectorList4() As Integer = {
        -1, 1, 0, 0
    }

    ' dseg:0042
    Private Shared ReadOnly XVectorList8() As Integer = {
        -1, 0, 1, 1, 1, 0, -1, -1
    }

    ' dseg:0052
    Private Shared ReadOnly YVectorList8() As Integer = {
        1, 1, 1, 0, -1, -1, -1, 0
    }

    Protected Enum RadiusType As Integer
        Update = 0
        Explode = 1
        Cleanup = 2
    End Enum

    Private Sub RaiseQuitGame()
        RaiseEvent QuitGame()
    End Sub

    Private Sub RaiseQuitZZT()
        RaiseEvent QuitZZT()
    End Sub

    Public Sub ConfirmQuitGame()
        State.QuitGame = True
    End Sub

    Public Sub ConfirmQuitZZT()
        State.QuitZZT = True
    End Sub

    Public Sub ApplyEnvironment(ByVal NewEnv As Environment)
        Library = NewEnv.Library
        World = NewEnv.World
        Boards = NewEnv.Boards
        State = NewEnv.State
        Parameters = NewEnv.Parameters
        Scroll.Reset()
        State.StartBoard = World.Board
        StopSound()
        RaiseEvent WorldChanged(World)
    End Sub

    Protected Sub ReadInput()
        Input.Poll()
        State.KeyArrow = False
        State.KeyVectorX = 0
        State.KeyVectorY = 0
        If (Input.Cheat <> "") Then
            Cheat(Input.Cheat)
            State.KeyShift = False
            State.KeyPressed = 0
        Else
            State.KeyShift = Input.GetShift
            State.KeyPressed = Input.GetKey
            Select Case CType(State.KeyPressed, Windows.Forms.Keys)
                Case Keys.Left
                    State.KeyArrow = True
                    State.KeyVectorX = -1
                    State.KeyVectorY = 0
                Case Keys.Up
                    State.KeyArrow = True
                    State.KeyVectorX = 0
                    State.KeyVectorY = -1
                Case Keys.Down
                    State.KeyArrow = True
                    State.KeyVectorX = 0
                    State.KeyVectorY = 1
                Case Keys.Right
                    State.KeyArrow = True
                    State.KeyVectorX = 1
                    State.KeyVectorY = 0
            End Select
            If (State.KeyVectorX <> 0) OrElse (State.KeyVectorY <> 0) Then
                State.LastKeyVectorX = State.KeyVectorX
                State.LastKeyVectorY = State.KeyVectorY
            End If
        End If
    End Sub

    Protected Sub RemoveItem(ByVal X As Integer, ByVal Y As Integer)
        If (Parameters.PickupToFloor) Then
            RemoveItemFloor(X, Y)
        Else
            ElementAt(X, Y) = Library.GetElementFromID(ElementID.Empty)
        End If
        UpdateAt(X, Y)
    End Sub

    Protected Sub RemoveItemFloor(ByVal X As Integer, ByVal Y As Integer)
        Dim Color As Integer
        Dim Index As Integer
        Dim TargetTile As Tile
        Dim TargetX As Integer
        Dim TargetY As Integer

        For Index = 0 To 3
            TargetX = X + XVectorList4(Index)
            TargetY = Y + YVectorList4(Index)
            TargetTile = TileAt(TargetX, TargetY)
            If (TargetTile.Element.Cycle >= 0) Then
                With ThingAt(TargetX, TargetY)
                    TargetTile.Element = .UnderElement
                    TargetTile.Color = .UnderColor
                End With
            End If
            Select Case TargetTile.Element.ID
                Case ElementID.Empty, ElementID.SliderNS, ElementID.SliderEW, ElementID.Boulder
                    Color = 0
                    Exit For
                Case ElementID.Floor
                    Color = TargetTile.Color
            End Select
        Next

        If (Color = 0) Then
            ElementAt(X, Y) = Library.GetElementFromID(ElementID.Empty)
        Else
            TileAt(X, Y) = New Tile(Library.GetElementFromID(ElementID.Floor), Color)
        End If
    End Sub

    Protected Sub Attack(ByVal Source As Integer, ByVal TargetX As Integer, ByVal TargetY As Integer)
        Dim SourceE As Element
        Dim TargetE As Element
        Dim T As Thing = CurrentBoard.Things(Source)

        With T
            SourceE = ElementAt(.X, .Y)
        End With
        TargetE = ElementAt(TargetX, TargetY)

        If (Source = 0) AndAlso (World.EnergyCycles > 0) Then
            World.Score += TargetE.Score
            StatusUpdate()
        Else
            HarmThing(Source)
        End If

        If (Source > 0) AndAlso (Source <= State.ActIndex) Then
            State.ActIndex -= 1
        End If

        If (TargetE.ID = ElementID.Player) AndAlso (World.EnergyCycles > 0) Then
            World.Score += SourceE.Score
            StatusUpdate()
        Else
            DestroyAt(TargetX, TargetY)
            PlaySound(2, SoundBank.EnemySuicide)
        End If
    End Sub

    Protected Sub MoveThing(ByVal Index As Integer, ByVal TargetX As Integer, ByVal TargetY As Integer)
        Dim SourceX As Integer
        Dim SourceY As Integer
        Dim SourceTile As Tile
        Dim TargetTile As Tile
        Dim GlowX As Integer
        Dim GlowY As Integer
        Dim GlowUpdate As Boolean
        Dim T As Thing = CurrentBoard.Things(Index)

        SourceX = T.X
        SourceY = T.Y
        SourceTile = TileAt(SourceX, SourceY)
        TargetTile = TileAt(TargetX, TargetY)
        ElementAt(SourceX, SourceY) = T.UnderElement
        T.UnderElement = TargetTile.Element
        ElementAt(TargetX, TargetY) = SourceTile.Element
        ColorAt(SourceX, SourceY) = T.UnderColor
        T.UnderColor = TargetTile.Color
        If (SourceTile.Element.ID = ElementID.Player) AndAlso (Parameters.ForcePlayerColor) Then
            ColorAt(TargetX, TargetY) = SourceTile.Color
        ElseIf (TargetTile.Element.ID = ElementID.Empty) Then
            ColorAt(TargetX, TargetY) = SourceTile.Color And &HF
        Else
            ColorAt(TargetX, TargetY) = (TargetTile.Color And &H70) Or (SourceTile.Color And &HF)
        End If
        T.X = TargetX
        T.Y = TargetY
        UpdateAt(SourceX, SourceY)
        UpdateAt(TargetX, TargetY)

        If (Index = 0) AndAlso (CurrentBoard.IsDark) Then
            If ((TargetY - SourceY) ^ 2) + ((TargetX - SourceX) ^ 2) = 1 Then
                For GlowX = TargetX - 11 To TargetX + 11
                    For GlowY = TargetY - 8 To TargetY + 8
                        If (GlowX >= 0) AndAlso (GlowX < Width) AndAlso (GlowY >= 0) AndAlso (GlowY < Height) Then
                            GlowUpdate = WithinRadius(TargetX, TargetY, GlowX, GlowY)
                            GlowUpdate = GlowUpdate Xor WithinRadius(SourceX, SourceY, GlowX, GlowY)
                            If (GlowUpdate) Then
                                UpdateAt(GlowX, GlowY)
                            End If
                        End If
                    Next
                Next
            Else
                Radius(SourceX, SourceY, RadiusType.Update)
                Radius(TargetX, TargetY, RadiusType.Update)
            End If
        End If
    End Sub

    Protected Sub Radius(ByVal X As Integer, ByVal Y As Integer, ByVal RadType As RadiusType)
        Dim GlowX As Integer
        Dim GlowY As Integer
        Dim E As Element
        Dim Index As Integer

        For GlowX = X - 7 To X + 7
            For GlowY = Y - 4 To Y + 4
                If (GlowX >= 0) AndAlso (GlowX < Width) AndAlso (GlowY >= 0) AndAlso (GlowY < Height) Then
                    Select Case RadType
                        Case RadiusType.Explode
                            If WithinRadius(X, Y, GlowX, GlowY) Then
                                E = ElementAt(GlowX, GlowY)
                                Index = ThingIndexAt(GlowX, GlowY)
                                If (E.Code <> "") AndAlso (Index > 0) Then
                                    'BOMBED
                                    SendLabel(-Index, OOPCodeSet(KnownCommand.Bombed), False)
                                End If
                                If (E.Destructible) OrElse (E.ID = ElementID.Star) Then
                                    DestroyAt(GlowX, GlowY)
                                    E = ElementAt(GlowX, GlowY)
                                End If
                                If (E.ID = ElementID.Empty) OrElse (E.ID = ElementID.Breakable) Then
                                    TileAt(GlowX, GlowY) = New Tile(Library.GetElementFromID(ElementID.Breakable), RNG.Next(7) + 9)
                                End If
                            End If
                        Case RadiusType.Cleanup
                            If WithinRadius(X, Y, GlowX, GlowY) Then
                                E = ElementAt(GlowX, GlowY)
                                If E.ID = ElementID.Breakable Then
                                    ElementAt(GlowX, GlowY) = Library.GetElementFromID(ElementID.Empty)
                                End If
                            End If
                    End Select
                    UpdateAt(GlowX, GlowY)
                End If
            Next
        Next
    End Sub

    Protected ReadOnly Property AbsDiff(ByVal Val1 As Integer, ByVal Val2 As Integer) As Integer
        Get
            Return Math.Abs(Val1 - Val2)
        End Get
    End Property

    Protected Sub HarmThing(ByVal Index As Integer)
        Dim E As Element
        Dim Color As Integer
        Dim T As Thing = CurrentBoard.Things(Index)
        Dim X As Integer
        Dim Y As Integer

        If (Index = 0) Then
            If (World.Health > 0) Then
                World.Health -= 10
                StatusUpdate()
                SetMessage(&H64, Parameters.OuchMessage)
                Color = ColorAt(T.X, T.Y)
                ColorAt(T.X, T.Y) = (Library.GetElementFromID(ElementID.Player).Color And &HF) Or &H70
                UpdateAt(T.X, T.Y)
                ColorAt(T.X, T.Y) = Color
                If (World.Health > 0) Then
                    World.TimePassed = 0
                    State.TimeLimitTicks = 0
                    If (CurrentBoard.RestartZap) Then
                        PlaySound(4, SoundBank.TimeOut)
                        ElementAt(T.X, T.Y) = Library.GetElementFromID(ElementID.Empty)
                        UpdateAt(T.X, T.Y)
                        X = T.X
                        Y = T.Y
                        T.X = CurrentBoard.EnterX
                        T.Y = CurrentBoard.EnterY
                        Radius(X, Y, RadiusType.Update)
                        Radius(T.X, T.Y, RadiusType.Update)
                        State.GamePaused = True
                    End If
                    PlaySound(4, SoundBank.Ouch)
                Else
                    PlaySound(5, SoundBank.GameOver)
                End If
            End If
        Else
            E = ElementAt(T.X, T.Y)
            Select Case E.ID
                Case ElementID.Bullet
                    PlaySound(3, SoundBank.BulletDie)
                Case ElementID.Object
                Case Else
                    PlaySound(3, SoundBank.EnemyDie)
            End Select
            RemoveThing(Index)
        End If
    End Sub

    Protected Sub DestroyAt(ByVal X As Integer, ByVal Y As Integer)
        Dim Index As Integer = ThingIndexAt(X, Y)
        If (Index = -1) Then
            RemoveItem(X, Y)
        Else
            HarmThing(Index)
        End If
    End Sub

    Protected Sub RemoveThing(ByVal ThingIndex As Integer)
        Dim I As Integer
        Dim Thing As Thing = CurrentBoard.Things(ThingIndex)

        If (ThingIndex > 0) AndAlso (ThingIndex < CurrentBoard.Things.Count) Then
            With Thing
                .Code = Nothing
                If (ThingIndex < State.ActIndex) Then
                    State.ActIndex -= 1
                End If
                TileAt(.X, .Y) = New Tile(.UnderElement, .UnderColor)
                If (.Y >= 0) Then
                    UpdateAt(.X, .Y)
                End If
            End With
            For I = 0 To CurrentBoard.Things.Count - 1
                With CurrentBoard.Things(I)
                    If (.Follower > ThingIndex) Then
                        .Follower -= 1
                    ElseIf (.Follower = ThingIndex) Then
                        .Follower = -1
                    End If
                    If (.Leader > ThingIndex) Then
                        .Leader -= 1
                    ElseIf (.Leader = ThingIndex) Then
                        .Leader = -1
                    End If
                End With
            Next
            CurrentBoard.Things.RemoveAt(ThingIndex)
        End If
    End Sub

    Protected Function SpawnProjectile(ByVal ProjElement As Element, ByVal X As Integer, ByVal Y As Integer, ByVal Direction As Vector, ByVal EnemyOwned As Boolean) As Boolean
        Dim E As Element
        Dim TargetX As Integer
        Dim TargetY As Integer
        Dim Success As Boolean

        TargetX = X + Direction.X
        TargetY = Y + Direction.Y
        E = ElementAt(TargetX, TargetY)
        If (E.Floor) OrElse (E.ID = ElementID.Water) OrElse (E.ID = ElementID.Lava) Then
            SpawnThing(TargetX, TargetY, ProjElement, ProjElement.Color, ProjElement.Cycle, Nothing)
            With CurrentBoard.Things(ThingCount - 1)
                If (EnemyOwned) Then
                    .P1 = 1
                Else
                    .P1 = 0
                End If
                .XStep = Direction.X
                .YStep = Direction.Y
                .P2 = &H64
                Success = True
            End With
        ElseIf (E.ID = ElementID.Breakable) OrElse (E.Destructible AndAlso ((E.ID = ElementID.Player AndAlso World.EnergyCycles = 0) OrElse (EnemyOwned = False))) Then
            DestroyAt(TargetX, TargetY)
            PlaySound(2, SoundBank.BulletDie)
            Success = True
        Else
            Success = False
        End If
        Return Success
    End Function

    Protected Sub SpawnThing(ByVal X As Integer, ByVal Y As Integer, ByVal E As Element, ByVal Col As Integer, ByVal Cycle As Integer, ByVal Source As Thing)
        Dim T As Thing
        If (ThingCount < Parameters.MaxThings) Then
            T = New Thing
            If (Not (Source Is Nothing)) Then
                T.XStep = Source.XStep
                T.YStep = Source.YStep
                T.P1 = Source.P1
                T.P2 = Source.P2
                T.P3 = Source.P3
                T.Follower = Source.Follower
                T.Leader = Source.Leader
                If (Source.Code Is Nothing) Then
                    T.Code = Nothing
                Else
                    T.Code = New CodeContainer(Source.Code.Code)
                    T.Length = T.Code.Code.Length
                End If
            End If
            With T
                T.X = X
                T.Y = Y
                T.Cycle = Cycle
                T.CurrentInstruction = 0
                With TileAt(X, Y)
                    T.UnderElement = .Element
                    T.UnderColor = .Color
                End With
            End With
            CurrentBoard.Things.Add(T)
            If (ElementAt(X, Y).EditorFloor) Then
                ColorAt(X, Y) = (ColorAt(X, Y) And &H70) + (Col And &HF)
            Else
                ColorAt(X, Y) = Col
            End If
            ElementAt(X, Y) = E
            UpdateAt(X, Y)
        End If
    End Sub

    Protected Sub Push(ByVal X As Integer, ByVal Y As Integer, ByVal V As Vector)
        Dim TargetX As Integer = X + V.X
        Dim TargetY As Integer = Y + V.Y
        Dim E As Element

        E = ElementAt(X, Y)
        If (E.Pushable) OrElse
            ((V.X = 0) AndAlso (E.ID = ElementID.SliderNS)) OrElse
            ((V.Y = 0) AndAlso (E.ID = ElementID.SliderEW)) Then

            If (V.X = 0) AndAlso (V.Y = 0) Then
                ' stack overflow
                Throw PascalError.Runtime202
                Exit Sub
            End If

            E = ElementAt(TargetX, TargetY)
            If (E.ID = ElementID.Transporter) Then
                PushTransporter(X, Y, V)
            ElseIf (E.ID <> ElementID.Empty) Then
                Push(TargetX, TargetY, V)
            End If
            E = ElementAt(TargetX, TargetY)
            If (Not E.Floor) Then
                If (E.Destructible) AndAlso (E.ID <> ElementID.Player) Then
                    DestroyAt(TargetX, TargetY)
                    E = ElementAt(TargetX, TargetY)
                End If
            End If
            If (E.Floor) Then
                Move(X, Y, TargetX, TargetY)
            End If
        End If
    End Sub

    Protected Sub PushTransporter(ByVal X As Integer, ByVal Y As Integer, ByVal V As Vector)
        Dim TargetX As Integer = X + V.X
        Dim TargetY As Integer = Y + V.Y
        Dim SearchX As Integer
        Dim SearchY As Integer
        Dim DestX As Integer
        Dim DestY As Integer
        Dim TransEnd As Boolean
        Dim Found As Boolean

        Dim T As Thing
        Dim TargetTile As Tile

        T = ThingAt(TargetX, TargetY)
        If (T.XStep = V.X) AndAlso (T.YStep = V.Y) Then
            SearchX = T.X
            SearchY = T.Y
            DestX = -1
            TransEnd = False
            Found = True
            Do
                SearchX += V.X
                SearchY += V.Y
                TargetTile = TileAt(SearchX, SearchY)
                If (TargetTile.Element.ID = ElementID.BoardEdge) Then
                    TransEnd = True
                Else
                    If (Found) Then
                        Found = False
                        If (Not TargetTile.Element.Floor) Then
                            Push(SearchX, SearchY, V)
                            TargetTile = TileAt(SearchX, SearchY)
                        End If
                        If (TargetTile.Element.Floor) Then
                            TransEnd = True
                            DestX = SearchX
                            DestY = SearchY
                        Else
                            DestX = -1
                        End If
                    End If
                End If
                If (TargetTile.Element.ID = ElementID.Transporter) Then
                    With ThingAt(SearchX, SearchY)
                        If (.XStep = -V.X) AndAlso (.YStep = -V.Y) Then
                            Found = True
                        End If
                    End With
                End If
            Loop Until (TransEnd)
            If (DestX <> -1) Then
                Move(T.X - T.XStep, T.Y - T.YStep, DestX, DestY)
                PlaySound(3, SoundBank.Transporter)
            End If
        End If
    End Sub

    Protected Sub Move(ByVal X As Integer, ByVal Y As Integer, ByVal TargetX As Integer, ByVal TargetY As Integer)
        Dim Index As Integer
        Dim T As Tile

        Index = ThingIndexAt(X, Y)
        If (Index < 0) Then
            T = TileAt(X, Y)
            TileAt(TargetX, TargetY) = T
            UpdateAt(TargetX, TargetY)
            RemoveItem(X, Y)
            UpdateAt(X, Y)
        Else
            MoveThing(Index, TargetX, TargetY)
        End If
    End Sub

    Protected Sub MoveRiver(ByVal Index As Integer)
        Dim XVector As Integer = 0
        Dim YVector As Integer = 0
        Dim TargetX As Integer
        Dim TargetY As Integer
        Dim Thing As Thing = CurrentBoard.Things(Index)

        With Thing
            Select Case .UnderElement.ID
                Case ElementID.WaterN
                    YVector = -1
                Case ElementID.WaterS
                    YVector = 1
                Case ElementID.WaterW
                    XVector = -1
                Case ElementID.WaterE
                    XVector = 1
            End Select
            If (XVector <> 0) OrElse (YVector <> 0) Then
                TargetX = .X + XVector
                TargetY = .Y + YVector
                If (Index = 0) Then
                    Interact(TargetX, TargetY, 0, XVector, YVector)
                End If
                If (ElementAt(TargetX, TargetY).Floor) Then
                    MoveThing(Index, TargetX, TargetY)
                End If
            End If
        End With
    End Sub

    Protected Sub InitBoard(ByVal BoardNumber As Integer)
        If (BoardNumber >= Boards.Count) Then
            BoardNumber = World.Board
        End If
        World.Board = BoardNumber
    End Sub

    Public Sub SetBoard(ByVal BoardNumber As Integer)
        Messenger.Clear()
        ElementAt(PlayerX, PlayerY) = Library.GetElementFromID(ElementID.Player)
        ColorAt(PlayerX, PlayerY) = ElementAt(PlayerX, PlayerY).Color
        InitBoard(BoardNumber)
        RaiseEvent BoardChanged(CurrentBoard)
    End Sub

    Protected Sub RedrawBoard()
        Dim X As Integer
        Dim Y As Integer
        For Y = 0 To Height - 1
            For X = 0 To Width - 1
                UpdateAt(X, Y)
            Next
        Next
    End Sub

    Protected Sub UpdateBorder()
        Dim X As Integer
        For X = 0 To Width - 1
            UpdateAt(X, 0)
            UpdateAt(X, Height - 1)
        Next
        For X = 1 To Height - 2
            UpdateAt(0, X)
            UpdateAt(Width - 1, X)
        Next
    End Sub

    Protected Sub BoardFade()
        Dim X As Integer
        Dim Y As Integer

        State.FadeMode = EnvironmentFadeMode.FadeOut
        State.FadeIndex = 0
        RedrawBoard()

        For X = 0 To Parameters.BoardWidth - 1
            For Y = 0 To Parameters.BoardHeight - 1
                Video.CancelUpdate(X, Y)
            Next
        Next

        If (State.FadeMatrix Is Nothing) Then
            BuildFadeMatrix()
        End If
    End Sub

    Protected Sub EnterBoard()
        With CurrentBoard
            .EnterX = PlayerX
            .EnterY = PlayerY
            State.TimeLimitTicks = 0
        End With
        World.TimePassed = 0
    End Sub

    Protected Sub EnterBoardPostFade()
        With CurrentBoard
            If (.IsDark) AndAlso (Alerts.Dark) Then
                SetMessage(&HC8, Parameters.DarkMessage)
                Alerts.Dark = False
            End If
        End With
        StatusUpdate()
    End Sub

    Protected Sub SetMessage(ByVal Ticks As Integer, ByVal Message As String)
        Ticks \= ((State.GameWaitTime) + 1)
        If (Messenger.TicksLeft > 0) Then
            UpdateBorder()
        End If
        Messenger.SetMessage(Ticks, Message)
    End Sub

    Protected Sub ConveyorMove(ByVal X As Integer, ByVal Y As Integer, ByVal RotationDirection As Integer)
        Dim FullyPushable As Boolean
        Dim Index As Integer
        Dim RotationIndex As Integer
        Dim RotationStart As Integer
        Dim RotationEnd As Integer
        Dim TempThing As Integer
        Dim TempThingX As Integer
        Dim TempThingY As Integer
        Dim TempTile As Tile = Nothing
        Dim SourceX As Integer
        Dim SourceY As Integer
        Dim TargetX As Integer
        Dim TargetY As Integer
        If RotationDirection >= 0 Then
            RotationStart = 7
            RotationEnd = 0
        Else
            RotationStart = 0
            RotationEnd = 7
        End If
        FullyPushable = True
        For Index = 0 To 7
            If Not ElementAt(X + XVectorList8(Index), Y + YVectorList8(Index)).Pushable Then
                FullyPushable = False
                Exit For
            End If
        Next
        If FullyPushable Then
            SourceX = X + XVectorList8(RotationStart)
            SourceY = Y + YVectorList8(RotationStart)
            TempThing = ThingIndexAt(SourceX, SourceY)
            TempThingX = SourceX
            TempThingY = SourceY
            If TempThing >= 0 Then
                With CurrentBoard.Things(TempThing)
                    .X = 0
                    .Y = 0
                End With
            End If
            TempTile = TileAt(TempThingX, TempThingY)
            TileAt(TempThingX, TempThingY) = New Tile(Library.GetElementFromID(ElementID.Empty), 0)
        End If
        For RotationIndex = RotationStart To RotationEnd Step -RotationDirection
            SourceX = X + XVectorList8(RotationIndex And 7)
            SourceY = Y + YVectorList8(RotationIndex And 7)
            TargetX = X + XVectorList8((RotationIndex + RotationDirection) And 7)
            TargetY = Y + YVectorList8((RotationIndex + RotationDirection) And 7)
            If ElementAt(SourceX, SourceY).Pushable Then
                If ElementAt(TargetX, TargetY).ID = ElementID.Empty Then
                    Move(SourceX, SourceY, TargetX, TargetY)
                End If
            End If
        Next
        If FullyPushable Then
            TargetX = X + XVectorList8((RotationStart + RotationDirection) And 7)
            TargetY = Y + YVectorList8((RotationStart + RotationDirection) And 7)
            If TempThing >= 0 Then
                With CurrentBoard.Things(TempThing)
                    .X = TargetX
                    .Y = TargetY
                End With
            End If
            TileAt(TargetX, TargetY) = TempTile
        End If
        For Index = 0 To 7
            UpdateAt(X + XVectorList8(Index), Y + YVectorList8(Index))
        Next
    End Sub

    Protected Function EraseTile(ByVal X As Integer, ByVal Y As Integer) As Boolean
        Dim Index As Integer
        Index = ThingIndexAt(X, Y)
        With ElementAt(X, Y)
            If (.ID = ElementID.Empty) OrElse (Index = 0) Then
                Return False
            ElseIf (Index > 0) Then
                RemoveThing(Index)
            ElseIf (Not .EditorFloor) Then
                ElementAt(X, Y) = Library.GetElementFromID(ElementID.Empty)
            End If
            Return True
        End With
    End Function

    Protected Sub PlotTile(ByVal X As Integer, ByVal Y As Integer, ByVal T As Tile)
        Dim Color As Integer

        If (ElementAt(X, Y).ID <> ElementID.Player) Then
            Color = T.Color
            If (T.Element.Color >= &HF0) Then
                If (Color = 0) Then
                    Color = ColorAt(X, Y)
                End If
                If (Color = 0) Then
                    Color = &HF
                End If
                If (T.Element.Color = &HFE) Then
                    Color = ((Color - 8) << 4) + &HF
                End If
            Else
                Color = T.Element.Color
            End If
            If (ElementAt(X, Y).ID = T.Element.ID) Then
                ColorAt(X, Y) = Color
            Else
                DestroyAt(X, Y)
                If (T.Element.Cycle < 0) Then
                    TileAt(X, Y) = New Tile(T.Element, Color)
                Else
                    SpawnThing(X, Y, T.Element, Color, T.Element.Cycle, Nothing)
                End If
            End If
            UpdateAt(X, Y)
        End If
    End Sub

    Public ReadOnly Property PlayerX() As Integer
        Get
            Return CurrentBoard.Things(0).X
        End Get
    End Property

    Public ReadOnly Property PlayerY() As Integer
        Get
            Return CurrentBoard.Things(0).Y
        End Get
    End Property

    Public ReadOnly Property Player As Thing
        Get
            Return CurrentBoard.Things(0)
        End Get
    End Property

    Protected ReadOnly Property ThingCount As Integer
        Get
            Return CurrentBoard.Things.Count
        End Get
    End Property

    Protected Property TileAt(ByVal X As Integer, ByVal Y As Integer) As Tile
        Get
            Return CurrentBoard.TileAt(X, Y)
        End Get
        Set(ByVal value As Tile)
            CurrentBoard.TileAt(X, Y) = value
        End Set
    End Property

    Protected Property ColorAt(ByVal X As Integer, ByVal Y As Integer) As Integer
        Get
            Return CurrentBoard.ColorAt(X, Y)
        End Get
        Set(ByVal value As Integer)
            CurrentBoard.ColorAt(X, Y) = value
        End Set
    End Property

    Protected Property ElementAt(ByVal X As Integer, ByVal Y As Integer) As Element
        Get
            Return CurrentBoard.ElementAt(X, Y)
        End Get
        Set(ByVal value As Element)
            CurrentBoard.ElementAt(X, Y) = value
        End Set
    End Property

    Protected ReadOnly Property ThingIndexAt(ByVal X As Integer, ByVal Y As Integer) As Integer
        Get
            Return CurrentBoard.ThingIndexAt(X, Y)
        End Get
    End Property

    Protected ReadOnly Property ThingAt(ByVal X As Integer, ByVal Y As Integer) As Thing
        Get
            Return CurrentBoard.ThingAt(X, Y)
        End Get
    End Property

    Protected ReadOnly Property Width As Integer
        Get
            Return CurrentBoard.Width
        End Get
    End Property

    Protected ReadOnly Property Height As Integer
        Get
            Return CurrentBoard.Height
        End Get
    End Property

    Protected ReadOnly Property WithinRadius(ByVal CenterX As Integer, ByVal CenterY As Integer, ByVal X As Integer, ByVal Y As Integer) As Boolean
        Get
            Dim XR As Integer = (CenterX - X)
            Dim YR As Integer = (CenterY - Y)
            Return ((XR * XR) + ((YR * YR) * 2) < 50)
        End Get
    End Property

    Protected Function Random(ByVal Range As Integer) As Integer
        Return RNG.Next(Range)
    End Function

    Protected ReadOnly Property AdjacentIDs(ByVal X As Integer, ByVal Y As Integer, ByVal ID As ElementID) As Integer
        Get
            Return CurrentBoard.AdjacentIDs(X, Y, ID)
        End Get
    End Property

    Private Sub BuildFadeMatrix()
        Dim Index As Integer
        Dim Count As Integer = 0
        Dim TempRandom As New Random
        Dim X As Integer
        Dim Y As Integer
        Dim Width As Integer = Parameters.ViewWidth
        Dim Height As Integer = Parameters.ViewHeight
        Dim Ret As New List(Of EnvironmentGameState.FadePoint)
        Dim Max As Integer = (Width * Height)

        Do
            X = TempRandom.Next(Width)
            Y = TempRandom.Next(Height)
            For Index = 0 To Count - 1
                With Ret(Index)
                    If (.X = X) AndAlso (.Y = Y) Then
                        Continue Do
                    End If
                End With
            Next
            Ret.Add(New EnvironmentGameState.FadePoint(X, Y))
            Count += 1
        Loop While (Count < Max)

        State.FadeMatrix = Ret.ToArray
    End Sub

End Class

