﻿' RotonBoard
'
' Contains code for the storage and manipulation of a board's information and tiles.



Public Structure Tile
    Public Element As Element
    Public Color As Integer

    Public Sub New(ByVal NewElement As Element, ByVal NewColor As Integer)
        Element = NewElement
        Color = NewColor
    End Sub
End Structure

Public Class ThingList
    Private InternalList() As Thing
    Private _Count As Integer

    Public Sub New(ByVal TotalThings As Integer)
        ReDim InternalList(TotalThings - 1)
        Clear()
    End Sub

    Public ReadOnly Property Count As Integer
        Get
            Return _Count
        End Get
    End Property

    Default Public Property Item(ByVal Index As Integer) As Thing
        Get
            Return InternalList(Index)
        End Get
        Set(ByVal value As Thing)
            InternalList(Index) = value
        End Set
    End Property

    Public Sub RemoveAt(ByVal Index As Integer)
        Dim LoopIndex As Integer

        If (Index < _Count) AndAlso (Index >= 0) Then
            For LoopIndex = Index To _Count - 2
                InternalList(LoopIndex).CopyFrom(InternalList(LoopIndex + 1))
            Next
            _Count -= 1
        End If
    End Sub

    Public Sub Add(ByVal NewThing As Thing)
        InternalList(_Count) = NewThing
        _Count += 1
    End Sub

    Public Function IndexOf(ByVal Thing As Thing) As Integer
        Dim Index As Integer
        For Index = 0 To _Count - 1
            If (Thing.Equals(InternalList(Index))) Then
                Return Index
            End If
        Next
        Return -1
    End Function

    Public Sub Clear()
        Dim Num As Integer = InternalList.Length - 1
        Dim Index As Integer
        ReDim InternalList(Num)
        _Count = 0

        For Index = 0 To Num
            InternalList(Index) = New Thing
        Next
    End Sub
End Class

Public Class Board
    Public ReadOnly Width As Integer
    Public ReadOnly Height As Integer
    Public ReadOnly Area As Integer

    Public Name As String
    Public Tiles() As Tile
    Public MaxShots As Integer
    Public IsDark As Boolean
    Public ExitN As Integer
    Public ExitS As Integer
    Public ExitW As Integer
    Public ExitE As Integer
    Public RestartZap As Boolean
    Public Message As String
    Public EnterX As Integer
    Public EnterY As Integer
    Public TimeLimit As Integer
    'Public Things As New List(Of Thing)
    Public Things As ThingList

    Private PlayerElement As Element
    Private EdgeElement As Element
    Private EmptyElement As Element

    Public Sub New(ByVal NewWidth As Integer, ByVal NewHeight As Integer, ByVal NewThingCount As Integer, ByVal NewEdgeElement As Element, ByVal NewEmptyElement As Element, ByVal NewPlayerElement As Element)
        SetDefaults()
        Width = NewWidth
        Height = NewHeight
        Area = Width * Height
        ReDim Tiles(Area - 1)
        Things = New ThingList(NewThingCount)

        EdgeElement = NewEdgeElement
        EmptyElement = NewEmptyElement
        PlayerElement = NewPlayerElement
    End Sub

    Public Sub SetDefaults()
        Name = "Untitled"
        MaxShots = 255
        IsDark = False
        ExitN = 0
        ExitS = 0
        ExitW = 0
        ExitE = 0
        RestartZap = False
        Message = ""
        EnterX = 0
        EnterY = 0
        TimeLimit = 0
    End Sub

    Public Sub Clear()
        Dim Index As Integer
        Dim X As Integer
        Dim Y As Integer
        Dim T As Thing

        ReDim Tiles(Area - 1)
        Things.Clear()
        For Index = 0 To Area - 1
            Tiles(Index) = New Tile(EmptyElement, 0)
        Next
        If (Not (PlayerElement Is Nothing)) Then
            X = Width \ 2
            Y = Height \ 2
            T = New Thing
            With T
                .X = X
                .Y = Y
                .Cycle = PlayerElement.Cycle
                .Leader = 0
                .Follower = 0
                .CurrentInstruction = 0
                .UnderColor = EmptyElement.Color
                .UnderElement = EmptyElement
                TileAt(X, Y) = New Tile(PlayerElement, PlayerElement.Color)
            End With
            Things.Add(T)
        End If

    End Sub

    Public Property TileAt(ByVal X As Integer, ByVal Y As Integer) As Tile
        Get
            If (X < 0) OrElse (X >= Width) OrElse (Y < 0) OrElse (Y >= Height) Then
                Return New Tile(EdgeElement, 0)
            End If
            Return Tiles(X + (Y * Width))
        End Get
        Set(ByVal value As Tile)
            If (X < 0) OrElse (X >= Width) OrElse (Y < 0) OrElse (Y >= Height) Then
                Exit Property
            End If
            Tiles(X + (Y * Width)) = value
        End Set
    End Property

    Public Property ColorAt(ByVal X As Integer, ByVal Y As Integer) As Integer
        Get
            If (X < 0) OrElse (X >= Width) OrElse (Y < 0) OrElse (Y >= Height) Then
                Return 0
            End If
            Return Tiles(X + (Y * Width)).Color
        End Get
        Set(ByVal value As Integer)
            If (X < 0) OrElse (X >= Width) OrElse (Y < 0) OrElse (Y >= Height) Then
                Exit Property
            End If
            Tiles(X + (Y * Width)).Color = value
        End Set
    End Property

    Public Property ElementAt(ByVal X As Integer, ByVal Y As Integer) As Element
        Get
            If (X < 0) OrElse (X >= Width) OrElse (Y < 0) OrElse (Y >= Height) Then
                Return EdgeElement
            End If
            Return Tiles(X + (Y * Width)).Element
        End Get
        Set(ByVal value As Element)
            If (X < 0) OrElse (X >= Width) OrElse (Y < 0) OrElse (Y >= Height) Then
                Exit Property
            End If
            Tiles(X + (Y * Width)).Element = value
        End Set
    End Property

    Public ReadOnly Property ThingIndexAt(ByVal X As Integer, ByVal Y As Integer) As Integer
        Get
            Dim Index As Integer
            For Index = 0 To Things.Count - 1
                With Things(Index)
                    If (.X = X) AndAlso (.Y = Y) Then
                        Return Index
                    End If
                End With
            Next
            Return -1
        End Get
    End Property

    Public ReadOnly Property ThingAt(ByVal X As Integer, ByVal Y As Integer) As Thing
        Get
            Dim Index As Integer
            Dim T As Thing

            Index = ThingIndexAt(X, Y)
            If (Index >= 0) Then
                Return Things(Index)
            Else
                T = New Thing
                T.UnderElement = EmptyElement
                Return T
            End If
        End Get
    End Property

    Public ReadOnly Property AdjacentIDs(ByVal X As Integer, ByVal Y As Integer, ByVal ID As ElementID) As Integer
        Get
            Dim Index As Integer
            Index = 0
            If (Y <= 0) OrElse (ElementAt(X, Y - 1).ID = ID) OrElse (ThingAt(X, Y - 1).UnderElement.ID = ID) Then
                Index += 1
            End If
            If (Y >= Height - 1) OrElse (ElementAt(X, Y + 1).ID = ID) OrElse (ThingAt(X, Y + 1).UnderElement.ID = ID) Then
                Index += 2
            End If
            If (X <= 0) OrElse (ElementAt(X - 1, Y).ID = ID) OrElse (ThingAt(X - 1, Y).UnderElement.ID = ID) Then
                Index += 4
            End If
            If (X >= Width - 1) OrElse (ElementAt(X + 1, Y).ID = ID) OrElse (ThingAt(X + 1, Y).UnderElement.ID = ID) Then
                Index += 8
            End If
            Return Index
        End Get
    End Property
End Class

