﻿' RotonDemo
'
' Contains code that handles the recording and playback of game demos.

Imports System.IO



' 2 bytes per poll
' - 1 byte is the current key pressed
' - 1 byte is flags
'
' flag bits:
' - bit 0 = shift is pressed
' - bit 1 = a cheat string follows
'
' if all bits are set in the flag, end the demo

Public Class DemoRecorder
    Implements Input

    Private Base As Input
    Private DemoWriter As BinaryWriter

    Public Sub New(ByVal SourceInput As Input, ByVal TargetStream As Stream)
        Base = SourceInput
        DemoWriter = New BinaryWriter(TargetStream)
    End Sub

    Public Property Cheat As String Implements Input.Cheat
        Get
            Return Base.Cheat
        End Get
        Set(ByVal value As String)
            Base.Cheat = value
        End Set
    End Property

    Public Function GetKey() As Byte Implements Input.GetKey
        Return Base.GetKey
    End Function

    Public Function GetShift() As Boolean Implements Input.GetShift
        Return Base.GetShift
    End Function

    Public Sub Poll() Implements Input.Poll
        Dim KeyOutput As Byte
        Dim ShiftOutput As Byte = 0

        Base.Poll()
        KeyOutput = Base.GetKey
        If (Base.GetShift) Then
            ShiftOutput += CByte(1)
        End If
        If (Base.Cheat <> "") Then
            ShiftOutput += CByte(2)
        End If
        DemoWriter.Write(KeyOutput)
        DemoWriter.Write(ShiftOutput)
        If ((ShiftOutput And 2) = 2) Then
            DemoWriter.Write(Base.Cheat)
        End If
    End Sub

    Public Sub Close()
        Dim CloseByte As Byte
        CloseByte = &HFF
        DemoWriter.Write(CloseByte)
        DemoWriter.Write(CloseByte)
    End Sub
End Class

Public Class DemoInput
    Implements Input

    Private Base As Stream
    Private Reader As BinaryReader
    Private CurrentKey As Byte
    Private CurrentShift As Boolean
    Private CurrentCheat As String
    Private FallbackInput As Input
    Private Ended As Boolean

    Public Sub New(ByVal Source As Stream, ByVal Fallback As Input)
        Base = Source
        Reader = New BinaryReader(Base)
        FallbackInput = Fallback
        Ended = False
    End Sub

    Public Property Cheat As String Implements Input.Cheat
        Get
            If (Ended) Then
                Return FallbackInput.Cheat
            Else
                Return CurrentCheat
            End If
        End Get
        Set(ByVal value As String)
            If (Ended) Then
                FallbackInput.Cheat = value
            End If
        End Set
    End Property

    Public Function GetKey() As Byte Implements Input.GetKey
        If (Ended) Then
            Return FallbackInput.GetKey
        Else
            Return CurrentKey
        End If
    End Function

    Public Function GetShift() As Boolean Implements Input.GetShift
        If (Ended) Then
            Return FallbackInput.GetShift
        Else
            Return CurrentShift
        End If
    End Function

    Public Sub Poll() Implements Input.Poll
        Dim Flags As Byte

        If (Not Ended) Then
            Try
                CurrentKey = Reader.ReadByte
                Flags = Reader.ReadByte
            Catch ex As EndOfStreamException
                Ended = True
            End Try

            If (Flags = &HFF) Then
                Ended = True
            End If
        End If

        If (Ended) Then
            FallbackInput.Poll()
        Else
            CurrentShift = ((Flags And 1) = 1)
            If ((Flags And 2) = 2) Then
                CurrentCheat = Reader.ReadString
            Else
                CurrentCheat = ""
            End If
        End If
    End Sub
End Class
