﻿' RotonDraw
'
' Contains special draw code for all elements.



Partial Public Class Environment

    Private Shared ReadOnly LineChars() As Byte = {
        &HF9, &HD0, &HD2, &HBA,
        &HB5, &HBC, &HBB, &HB9,
        &HC6, &HC8, &HC9, &HCC,
        &HCD, &HCA, &HCB, &HCE
    }

    Private Shared ReadOnly StarChars() As Byte = {
        &HB3, &H2F, &HC4, &H5C
    }

    Private Shared ReadOnly TransHChars() As Byte = {
        &H28, &H3C, &H28, &HB3,
        &H29, &H3E, &H29, &HB3
    }

    Private Shared ReadOnly TransVChars() As Byte = {
        &H5E, &H7E, &H5E, &H2D,
        &H76, &H5F, &H76, &H2D
    }

    Private Shared ReadOnly WebChars() As Byte = {
        &HFA, &HB3, &HB3, &HB3,
        &HC4, &HD9, &HBF, &HB4,
        &HC4, &HC0, &HDA, &HC3,
        &HC4, &HC1, &HC2, &HC5
    }

    Protected Function Draw(ByVal X As Integer, ByVal Y As Integer) As AnsiChar
        Dim T As Tile
        Dim Index As Integer

        T = TileAt(X, Y)

        Select Case T.Element.ID
            Case ElementID.Star
                Index = T.Color + 1
                If (Index > 15) Then
                    Index = 9
                End If
                ColorAt(X, Y) = Index
                Return RenderedChar(TileAt(X, Y), ThingAt(X, Y), 0)
            Case ElementID.Line, ElementID.Web
                Return RenderedChar(T, ThingAt(X, Y), AdjacentIDs(X, Y, T.Element.ID))
        End Select

        Return RenderedChar(T, ThingAt(X, Y), 0)
    End Function

    Public Function RenderedChar(ByVal Tile As Tile, ByVal Thing As Thing, ByVal Adjacent As Integer) As AnsiChar
        Dim Index As Integer

        Select Case Tile.Element.ID

            Case ElementID.Empty, ElementID.None
                Return New AnsiChar(&H20, &HF)

            Case ElementID.BlackText
                Return New AnsiChar(Tile.Color, &HF)
            Case ElementID.BlueText
                Return New AnsiChar(Tile.Color, &H1F)
            Case ElementID.BrownText
                Return New AnsiChar(Tile.Color, &H6F)
            Case ElementID.CyanText
                Return New AnsiChar(Tile.Color, &H3F)
            Case ElementID.GreenText
                Return New AnsiChar(Tile.Color, &H2F)
            Case ElementID.RedText
                Return New AnsiChar(Tile.Color, &H4F)
            Case ElementID.PurpleText
                Return New AnsiChar(Tile.Color, &H5F)

            Case ElementID.BlinkWall
                Return New AnsiChar(&HCE, Tile.Color)
            Case ElementID.Bomb
                With Thing
                    If (.P1 > 1) Then
                        Return New AnsiChar(&H30 + .P1, Tile.Color)
                    Else
                        Return New AnsiChar(&HB, Tile.Color)
                    End If
                End With
            Case ElementID.Clockwise
                Select Case ((State.GameTick \ Tile.Element.Cycle) Mod 4)
                    Case 0
                        Return New AnsiChar(&HB3, Tile.Color)
                    Case 1
                        Return New AnsiChar(&H2F, Tile.Color)
                    Case 2
                        Return New AnsiChar(&HC4, Tile.Color)
                    Case Else
                        Return New AnsiChar(&H5C, Tile.Color)
                End Select
            Case ElementID.Counter
                Select Case ((State.GameTick \ Tile.Element.Cycle) Mod 4)
                    Case 3
                        Return New AnsiChar(&HB3, Tile.Color)
                    Case 2
                        Return New AnsiChar(&H2F, Tile.Color)
                    Case 1
                        Return New AnsiChar(&HC4, Tile.Color)
                    Case Else
                        Return New AnsiChar(&H5C, Tile.Color)
                End Select
            Case ElementID.DragonPup
                Select Case (State.GameTick And 3)
                    Case 0, 2
                        Return New AnsiChar(&H94, Tile.Color)
                    Case 1
                        Return New AnsiChar(&HA2, Tile.Color)
                    Case Else
                        Return New AnsiChar(&H95, Tile.Color)
                End Select
            Case ElementID.Duplicator
                With Thing
                    Select Case .P1
                        Case 1
                            Return New AnsiChar(&HFA, Tile.Color)
                        Case 2
                            Return New AnsiChar(&HF9, Tile.Color)
                        Case 3
                            Return New AnsiChar(&HF8, Tile.Color)
                        Case 4
                            Return New AnsiChar(&H6F, Tile.Color)
                        Case 5
                            Return New AnsiChar(&H4F, Tile.Color)
                        Case Else
                            Return New AnsiChar(&HFA, Tile.Color)
                    End Select
                End With
            Case ElementID.Line
                Return New AnsiChar(LineChars(Adjacent), Tile.Color)
            Case ElementID.Object
                Return New AnsiChar(Thing.P1, Tile.Color)
            Case ElementID.Pusher
                With Thing
                    If (.XStep = 1) Then
                        Return New AnsiChar(&H10, Tile.Color)
                    ElseIf (.XStep = -1) Then
                        Return New AnsiChar(&H11, Tile.Color)
                    ElseIf (.YStep = -1) Then
                        Return New AnsiChar(&H1E, Tile.Color)
                    Else
                        Return New AnsiChar(&H1F, Tile.Color)
                    End If
                End With
            Case ElementID.SpinningGun
                Select Case (State.GameTick Mod 8)
                    Case 0, 1
                        Return New AnsiChar(&H18, Tile.Color)
                    Case 2, 3
                        Return New AnsiChar(&H1A, Tile.Color)
                    Case 4, 5
                        Return New AnsiChar(&H19, Tile.Color)
                    Case Else
                        Return New AnsiChar(&H1B, Tile.Color)
                End Select
            Case ElementID.Star
                Index = Tile.Color
                Return New AnsiChar(StarChars(State.GameTick Mod 4), Index)
            Case ElementID.Stone
                Return New AnsiChar(&H41 + Random(&H1A), Tile.Color)
            Case ElementID.Transporter
                With Thing
                    If (.XStep = 0) Then
                        Index = ((State.GameTick \ .Cycle) Mod 4) + (.YStep << 1) + 2
                        Return New AnsiChar(TransVChars(Index), Tile.Color)
                    Else
                        Index = ((State.GameTick \ .Cycle) Mod 4) + (.XStep << 1) + 2
                        Return New AnsiChar(TransHChars(Index), Tile.Color)
                    End If
                End With
            Case ElementID.Web
                Return New AnsiChar(WebChars(Adjacent), Tile.Color)
        End Select

        'unknown
        If (Tile.Element.DrawCodeEnable) Then
            Return New AnsiChar(&H3F, Tile.Color)
        Else
            Return New AnsiChar(Tile.Element.Char, Tile.Color)
        End If

    End Function

    Protected Sub UpdateAt(ByVal X As Integer, ByVal Y As Integer)
        Dim E As Element = ElementAt(X, Y)
        Dim Ret As AnsiChar

        If (CurrentBoard.IsDark) AndAlso _
            (Not State.EditorMode) AndAlso _
            (Not E.AlwaysShown) AndAlso _
            (Not WithinRadius(PlayerX, PlayerY, X, Y) OrElse (World.TorchCycles <= 0)) Then
            ' dark square
            Ret = New AnsiChar(&HB0, &H7)
        ElseIf (E.DrawCodeEnable) Then
            Ret = Draw(X, Y)
        Else
            Ret = RenderedChar(TileAt(X, Y), ThingAt(X, Y), 0)
        End If

        Video.Plot(X, Y, Ret)
    End Sub

End Class
