﻿' RotonEditor
'
' Contains code that performs editor-related functions such as plotting and board navigation.



Partial Public Class Environment

    Public Sub EditorUpdate(ByVal X As Integer, ByVal Y As Integer)
        UpdateAt(X, Y)
    End Sub

    Public ReadOnly Property EditorChar(ByVal C As Integer, ByVal E As Element) As AnsiChar
        Get
            Dim Ret As AnsiChar

            Ret.Char = E.Char
            If (E.Color < &HF0) Then
                Ret.Color = E.Color
            ElseIf (E.Color = &HFE) Then
                Ret.Color = CByte(((C And &H7) << 4) Or &HF)
            Else
                Ret.Color = CByte(C)
            End If
            Return Ret
        End Get
    End Property

    Public Sub EditorMove(ByVal SourceX As Integer, ByVal SourceY As Integer, ByVal TargetX As Integer, ByVal TargetY As Integer)
        Move(SourceX, SourceY, TargetX, TargetY)
    End Sub

    Public Sub EditorMovePlayer(ByVal TargetX As Integer, ByVal TargetY As Integer)
        MoveThing(0, TargetX, TargetY)
    End Sub

    Public Function EditorIndexAt(ByVal X As Integer, ByVal Y As Integer) As Integer
        Return ThingIndexAt(X, Y)
    End Function

    Public Sub EditorPlot(ByVal X As Integer, ByVal Y As Integer, ByVal T As Tile, ByVal Th As Thing, ByVal Stats As Boolean)
        Dim I As Integer
        I = ThingIndexAt(X, Y)
        If (I <> 0) Then
            If (I > 0) Then
                RemoveThing(I)
            End If
            If (Not Stats) Then
                TileAt(X, Y) = T
            Else
                If Not (ElementAt(X, Y).EditorFloor) Then
                    ElementAt(X, Y) = Library.GetElementFromID(ElementID.Empty)
                End If
                If (Th Is Nothing) Then
                    SpawnThing(X, Y, T.Element, T.Color, T.Element.Cycle, Th)
                Else
                    SpawnThing(X, Y, T.Element, T.Color, Th.Cycle, Th)
                End If
            End If
        End If
        UpdateAt(X, Y)
        UpdateAt(X - 1, Y)
        UpdateAt(X + 1, Y)
        UpdateAt(X, Y + 1)
        UpdateAt(X, Y - 1)
    End Sub

    Public Sub EditorRemoveBoard(ByVal Index As Integer)
        Dim BoardIndex As Integer
        Dim ThingIndex As Integer

        For BoardIndex = 0 To Boards.Count - 1
            With Boards(BoardIndex)
                If (.ExitN > Index) Then
                    .ExitN -= 1
                ElseIf (.ExitN = Index) Then
                    .ExitN = 0
                End If
                If (.ExitS > Index) Then
                    .ExitS -= 1
                ElseIf (.ExitS = Index) Then
                    .ExitS = 0
                End If
                If (.ExitW > Index) Then
                    .ExitW -= 1
                ElseIf (.ExitW = Index) Then
                    .ExitW = 0
                End If
                If (.ExitE > Index) Then
                    .ExitE -= 1
                ElseIf (.ExitE = Index) Then
                    .ExitE = 0
                End If
                For ThingIndex = 0 To .Things.Count - 1
                    If (.TileAt(.Things(ThingIndex).X, .Things(ThingIndex).Y).Element.ID = ElementID.Passage) Then
                        With .ThingAt(.Things(ThingIndex).X, .Things(ThingIndex).Y)
                            If (.P3 > Index) Then
                                .P3 -= 1
                            ElseIf (.P3 = Index) Then
                                .P3 = 0
                            End If
                        End With
                    End If
                Next
            End With
        Next
        Boards.RemoveAt(Index)
    End Sub

    Public Sub EditorSetBoard(Index As Integer)
        SetBoard(Index)
        RedrawBoard()
    End Sub

    Public Sub SetEditorMode(ByVal Enabled As Boolean)
        If (Enabled) Then
            With Library.GetElementFromID(ElementID.Invisible)
                .Char = 176
                .Color = 255
            End With
            State.PlayerElement = Library.GetElementFromID(ElementID.Player)
        Else
            With Library.GetElementFromID(ElementID.Invisible)
                .Char = 32
                .Color = 255
            End With
        End If
        State.EditorMode = Enabled
    End Sub

End Class

