﻿' RotonElement
'
' Contains information about element libraries for both ZZT and Super ZZT.



Public Enum ElementID As Integer
    None = 0
    Ammo
    Bear
    BlinkRayH
    BlinkRayV
    BlinkWall
    BoardEdge
    Bomb
    Boulder
    Breakable
    Bullet
    Clockwise
    Counter
    Door
    DragonPup
    Duplicator
    Empty
    Energizer
    Fake
    Floor
    Forest
    Gem
    Head
    Invisible
    Key
    Lava
    Line
    Lion
    Messenger
    Monitor
    Normal
    [Object]
    Pairer
    Passage
    Player
    Pusher
    Ricochet
    Roton
    Ruffian
    Scroll
    Segment
    Shark
    SliderEW
    SliderNS
    Slime
    Solid
    Spider
    SpinningGun
    Star
    Stone
    Tiger
    Torch
    Transporter
    Water
    WaterE
    WaterN
    WaterS
    WaterW
    Web

    BlueText
    GreenText
    CyanText
    RedText
    PurpleText
    BrownText
    BlackText
End Enum

Public Class Element
    Public ReadOnly OOPString As String
    Public ID As ElementID
    Public [Char] As Byte
    Public [Color] As Byte
    Public Destructible As Boolean
    Public Pushable As Boolean
    Public AlwaysShown As Boolean
    Public EditorFloor As Boolean
    Public Floor As Boolean
    Public DrawCodeEnable As Boolean
    Public Cycle As Integer
    Public Menu As Integer
    Public Key As Byte
    Public Name As String
    Public Category As String
    Public P1 As String
    Public P2 As String
    Public P3 As String
    Public Board As String
    Public [Step] As String
    Public Code As String
    Public Score As Integer

    Public Sub New(ByVal NewID As ElementID, Optional ByVal NewChar As Byte = 32, Optional ByVal NewColor As Byte = 255, Optional ByVal NewDestructible As Boolean = False, Optional ByVal NewPushable As Boolean = False, Optional ByVal NewAlwaysShown As Boolean = False, Optional ByVal NewEditorFloor As Boolean = False, Optional ByVal NewFloor As Boolean = False, Optional ByVal NewDrawCodeEnable As Boolean = False, Optional ByVal NewCycle As Integer = -1, Optional ByVal NewMenu As Byte = 0, Optional ByVal NewKey As Byte = 0, Optional ByVal NewName As String = "", Optional ByVal NewCategory As String = "", Optional ByVal NewP1 As String = "", Optional ByVal NewP2 As String = "", Optional ByVal NewP3 As String = "", Optional ByVal NewBoard As String = "", Optional ByVal NewStep As String = "", Optional ByVal NewCode As String = "", Optional ByVal NewScore As Integer = 0)
        ID = NewID
        [Char] = NewChar
        [Color] = NewColor
        Destructible = NewDestructible
        Pushable = NewPushable
        AlwaysShown = NewAlwaysShown
        EditorFloor = NewEditorFloor
        Floor = NewFloor
        DrawCodeEnable = NewDrawCodeEnable
        Cycle = NewCycle
        Menu = NewMenu
        Key = NewKey
        Name = NewName
        Category = NewCategory
        P1 = NewP1
        P2 = NewP2
        P3 = NewP3
        Board = NewBoard
        [Step] = NewStep
        Code = NewCode
        Score = NewScore
        OOPString = GetOOPString(Name)
    End Sub

    Private Shared Function GetOOPString(ByVal InName As String) As String
        Dim Index As Integer
        Dim Ret As String = ""
        Dim Max As Integer = InName.Length - 1
        Dim Ch As String

        For Index = 0 To Max
            Ch = InName.Substring(Index, 1).ToUpper
            Select Case Ch
                Case "A" To "Z"
                    Ret &= Ch
            End Select
        Next

        Return Ret
    End Function

    Public Overrides Function ToString() As String
        Return ID.ToString & ": [" & Menu & "/" & Chr(Key) & "] " & Name
    End Function
End Class

' stock elements that are built in to ZZT
Public NotInheritable Class ElementDefault
    Public Shared Function Ammo() As Element
        Return New Element(ElementID.Ammo, 132, 3, , True, , , , , , 1, 65, "Ammo")
    End Function
    Public Shared Function Bear() As Element
        Return New Element(ElementID.Bear, 153, 6, True, True, , , , , 3, 2, 66, "Bear", "Creatures:", "Sensitivity?", , , , , , 1)
    End Function
    Public Shared Function BearSuperZZT() As Element
        Return New Element(ElementID.Bear, 235, 2, True, True, , , , , 3, 2, 66, "Bear", "Creatures:", "Sensitivity?", , , , , , 1)
    End Function
    Public Shared Function BlackText() As Element
        Return New Element(ElementID.BlackText, , , , , , , , , , 6)
    End Function
    Public Shared Function BlinkRayH() As Element
        Return New Element(ElementID.BlinkRayH, 205, , , , , , , , , 3)
    End Function
    Public Shared Function BlinkRayV() As Element
        Return New Element(ElementID.BlinkRayV, 186, , , , , , , , , 3)
    End Function
    Public Shared Function BlinkWall() As Element
        Return New Element(ElementID.BlinkWall, 206, , , , , , , True, 1, 3, 76, "Blink wall", "Starting time", "Period", , , , "Wall direction")
    End Function
    Public Shared Function BlueText() As Element
        Return New Element(ElementID.BlueText, , , , , , , , , , 6)
    End Function
    Public Shared Function BoardEdge() As Element
        Return New Element(ElementID.BoardEdge, , , , , , , , , , 3)
    End Function
    Public Shared Function Bomb() As Element
        Return New Element(ElementID.Bomb, 11, , , True, , , , True, 6, 1, 66, "Bomb")
    End Function
    Public Shared Function Boulder() As Element
        Return New Element(ElementID.Boulder, 254, , , True, , , , , , 3, 79, "Boulder")
    End Function
    Public Shared Function Breakable() As Element
        Return New Element(ElementID.Breakable, 177, , , , , , , , , 3, 66, "Breakable")
    End Function
    Public Shared Function BrownText() As Element
        Return New Element(ElementID.BrownText, , , , , , , , , , 6)
    End Function
    Public Shared Function Bullet() As Element
        Return New Element(ElementID.Bullet, 248, 15, True, , , , , , 1, , , "Bullet")
    End Function
    Public Shared Function Clockwise() As Element
        Return New Element(ElementID.Clockwise, 47, , , , , , , True, 3, 1, 49, "Clockwise", "Conveyors:")
    End Function
    Public Shared Function Counter() As Element
        Return New Element(ElementID.Counter, 92, , , , , , , True, 2, 1, 50, "Counter")
    End Function
    Public Shared Function CyanText() As Element
        Return New Element(ElementID.CyanText, , , , , , , , , , 6)
    End Function
    Public Shared Function Door() As Element
        Return New Element(ElementID.Door, 10, 254, , , , , , , , 1, 68, "Door")
    End Function
    Public Shared Function DragonPup() As Element
        Return New Element(ElementID.DragonPup, 237, 4, True, True, , , , True, 2, 4, 68, "Dragon Pup", , "Intelligence?", "Switch Rate?", , , , , 1)
    End Function
    Public Shared Function Duplicator() As Element
        Return New Element(ElementID.Duplicator, 250, 15, , , , , , True, 2, 1, 85, "Duplicator", , , "Duplication rate?;SF", , , "Source direction?")
    End Function
    Public Shared Function Empty() As Element
        Return New Element(ElementID.Empty, , 112, , True, , , True, , , 3, , "Empty")
    End Function
    Public Shared Function Energizer() As Element
        Return New Element(ElementID.Energizer, 127, 5, , , , , , , , 1, 69, "Energizer")
    End Function
    Public Shared Function Fake() As Element
        Return New Element(ElementID.Fake, 178, , , , , True, True, , , 3, 65, "Fake")
    End Function
    Public Shared Function Floor() As Element
        Return New Element(ElementID.Floor, 176, , , , , True, True, , , 5, 70, "Floor", "Terrains:")
    End Function
    Public Shared Function Forest() As Element
        Return New Element(ElementID.Forest, 176, 32, , , , , , , , 3, 70, "Forest")
    End Function
    Public Shared Function Gem() As Element
        Return New Element(ElementID.Gem, 4, , True, True, , , , , , 1, 71, "Gem")
    End Function
    Public Shared Function GreenText() As Element
        Return New Element(ElementID.GreenText, , , , , , , , , , 6)
    End Function
    Public Shared Function Head() As Element
        Return New Element(ElementID.Head, 233, , True, , , , , , 2, 2, 72, "Head", "Centipedes", "Intelligence?", "Deviance?", , , , , 1)
    End Function
    Public Shared Function Invisible() As Element
        Return New Element(ElementID.Invisible, 32, , , , , , , , , 3, 73, "Invisible")
    End Function
    Public Shared Function Key() As Element
        Return New Element(ElementID.Key, 12, , , True, , , , , , 1, 75, "Key")
    End Function
    Public Shared Function Lava() As Element
        Return New Element(ElementID.Lava, 111, 78, , , , True, , , , 3, 76, "Lava", "Terrains:")
    End Function
    Public Shared Function Line() As Element
        Return New Element(ElementID.Line, 206, , , , , , , True, , 3, , "Line")
    End Function
    Public Shared Function Lion() As Element
        Return New Element(ElementID.Lion, 234, 12, True, True, , , , , 2, 2, 76, "Lion", "Beasts:", "Intelligence?", , , , , , 1)
    End Function
    Public Shared Function Messenger() As Element
        Return New Element(ElementID.Messenger)
    End Function
    Public Shared Function Monitor() As Element
        Return New Element(ElementID.Monitor, , 7, , , , , , , 1, , , "Monitor")
    End Function
    Public Shared Function MonitorSuperZZT() As Element
        Return New Element(ElementID.Monitor, 2, 31, , True, , , , , 1, , , "Monitor")
    End Function
    Public Shared Function None() As Element
        Return New Element(ElementID.None)
    End Function
    Public Shared Function Normal() As Element
        Return New Element(ElementID.Normal, 178, , , , , , , , , 3, 78, "Normal")
    End Function
    Public Shared Function [Object]() As Element
        Return New Element(ElementID.Object, 2, , , , , , , True, 3, 2, 79, "Object", , "Character?", , , , , "Edit Program")
    End Function
    Public Shared Function Pairer() As Element
        Return New Element(ElementID.Pairer, 229, 1, True, True, , , , , 2, 4, 80, "Pairer", "Intelligence?", , , , , , , 2)
    End Function
    Public Shared Function Passage() As Element
        Return New Element(ElementID.Passage, 240, 254, , , True, , , , 0, 1, 80, "Passage", , , , , "Room thru passage?")
    End Function
    Public Shared Function Player() As Element
        Return New Element(ElementID.Player, 2, 31, True, True, True, , , , 1, 1, 90, "Player", "Items:")
    End Function
    Public Shared Function PurpleText() As Element
        Return New Element(ElementID.PurpleText, , , , , , , , , , 6)
    End Function
    Public Shared Function Pusher() As Element
        Return New Element(ElementID.Pusher, 16, , , , , , , True, 4, 2, 80, "Pusher", , , , , , "Push direction?")
    End Function
    Public Shared Function RedText() As Element
        Return New Element(ElementID.RedText, , , , , , , , , , 6)
    End Function
    Public Shared Function Ricochet() As Element
        Return New Element(ElementID.Ricochet, 42, 10, , , , , , , , 3, 82, "Ricochet")
    End Function
    Public Shared Function Roton() As Element
        Return New Element(ElementID.Roton, 148, 13, True, True, , , , , 1, 4, 82, "Roton", "Uglies:", "Intelligence?", "Switch Rate?", , , , , 2)
    End Function
    Public Shared Function Ruffian() As Element
        Return New Element(ElementID.Ruffian, 5, 13, True, True, , , , , 1, 2, 82, "Ruffian", "Intelligence?", "Resting time?", , , , , , 2)
    End Function
    Public Shared Function Scroll() As Element
        Return New Element(ElementID.Scroll, 232, 15, , True, , , , , 1, 1, 83, "Scroll", , , , , , , "Edit text of scroll")
    End Function
    Public Shared Function Segment() As Element
        Return New Element(ElementID.Segment, 79, , True, , , , , , 2, 2, 83, "Segment", , , , , , , , 3)
    End Function
    Public Shared Function Shark() As Element
        Return New Element(ElementID.Shark, 94, 7, , , , , , , 3, 2, 89, "Shark", , "Intelligence?")
    End Function
    Public Shared Function SliderEW() As Element
        Return New Element(ElementID.SliderEW, 29, , , , , , , , , 3, 50, "Slider (EW)")
    End Function
    Public Shared Function SliderNS() As Element
        Return New Element(ElementID.SliderNS, 18, , , , , , , , , 3, 49, "Slider (NS)")
    End Function
    Public Shared Function Slime() As Element
        Return New Element(ElementID.Slime, 42, , , , , , , , 3, 2, 86, "Slime", , , "Movement speed?;FS")
    End Function
    Public Shared Function Solid() As Element
        Return New Element(ElementID.Solid, 219, , , , , , , , , 3, 83, "Solid", "Walls:")
    End Function
    Public Shared Function Spider() As Element
        Return New Element(ElementID.Spider, 15, , True, , , , , , 1, 4, 83, "Spider", "Intelligence?", , , , , , , 3)
    End Function
    Public Shared Function SpinningGun() As Element
        Return New Element(ElementID.SpinningGun, 24, , , , , , , True, 2, 2, 71, "Spinning Gun", , "Intelligence?", "Firing rate?", "Firing type?")
    End Function
    Public Shared Function Star() As Element
        Return New Element(ElementID.Star, 83, 15, , , , , , True, 1, , , "Star")
    End Function
    Public Shared Function Stone() As Element
        Return New Element(ElementID.Stone, 90, 15, , , , , , True, 1, 5, 90, "Stone")
    End Function
    Public Shared Function Tiger() As Element
        Return New Element(ElementID.Tiger, 227, 11, True, True, , , , , 2, 2, 84, "Tiger", , "Intelligence?", "Firing rate?", "Firing type?", , , , 2)
    End Function
    Public Shared Function Torch() As Element
        Return New Element(ElementID.Torch, 157, 6, , , True, , , , , 1, 84, "Torch")
    End Function
    Public Shared Function Transporter() As Element
        Return New Element(ElementID.Transporter, 197, , , , , , , True, 2, 3, 84, "Transporter", , , , , , "Direction?")
    End Function
    Public Shared Function Water() As Element
        Return New Element(ElementID.Water, 176, 249, , , , True, , , , 3, 87, "Water", "Terrains:")
    End Function
    Public Shared Function WaterE() As Element
        Return New Element(ElementID.WaterE, 16, 25, , , , True, True, , , 5, 54, "Water E")
    End Function
    Public Shared Function WaterN() As Element
        Return New Element(ElementID.WaterN, 30, 25, , , , True, True, , , 5, 54, "Water N")
    End Function
    Public Shared Function WaterS() As Element
        Return New Element(ElementID.WaterS, 31, 25, , , , True, True, , , 5, 54, "Water S")
    End Function
    Public Shared Function WaterW() As Element
        Return New Element(ElementID.WaterW, 17, 25, , , , True, True, , , 5, 54, "Water W")
    End Function
    Public Shared Function Web() As Element
        Return New Element(ElementID.Web, 197, , , , , True, True, True, , 5, 87, "Web")
    End Function
End Class

Public Class Library
    Private Elements As List(Of Element)

    Public Sub New()
        Elements = New List(Of Element)
    End Sub

    Public Sub New(ByVal NewElements() As Element)
        Elements = New List(Of Element)
        Elements.AddRange(NewElements)
    End Sub

    Public ReadOnly Property Count As Integer
        Get
            Return Elements.Count
        End Get
    End Property

    Public ReadOnly Property GetIDFromElement(ByVal E As ElementID) As Byte
        Get
            Dim Index As Integer
            For Index = 0 To Elements.Count - 1
                If Elements(Index).ID = E Then
                    Return CByte(Index)
                End If
            Next
            Return 0
        End Get
    End Property

    Public ReadOnly Property GetElementFromIndex(ByVal Index As Byte) As Element
        Get
            If (Index >= 0) AndAlso (Index < Elements.Count) Then
                Return Elements(Index)
            Else
                Return New Element(ElementID.None)
            End If
        End Get
    End Property

    Public ReadOnly Property GetElementFromID(ByVal ID As ElementID) As Element
        Get
            Dim Index As Integer
            For Index = 0 To Elements.Count - 1
                If Elements(Index).ID = ID Then
                    Return Elements(Index)
                End If
            Next
            Return New Element(ElementID.None)
        End Get
    End Property

    Public ReadOnly Property GetIndexFromElement(ByVal E As Element) As Integer
        Get
            Return Elements.IndexOf(E)
        End Get
    End Property
End Class

Public NotInheritable Class LibraryDefault
    Public Shared Function ZZT() As Library
        Return New Library(
            {
                ElementDefault.Empty,
                ElementDefault.BoardEdge,
                ElementDefault.Monitor,
                ElementDefault.Messenger,
                ElementDefault.Player,
                ElementDefault.Ammo,
                ElementDefault.Torch,
                ElementDefault.Gem,
                ElementDefault.Key,
                ElementDefault.Door,
                ElementDefault.Scroll,
                ElementDefault.Passage,
                ElementDefault.Duplicator,
                ElementDefault.Bomb,
                ElementDefault.Energizer,
                ElementDefault.Star,
                ElementDefault.Clockwise,
                ElementDefault.Counter,
                ElementDefault.Bullet,
                ElementDefault.Water,
                ElementDefault.Forest,
                ElementDefault.Solid,
                ElementDefault.Normal,
                ElementDefault.Breakable,
                ElementDefault.Boulder,
                ElementDefault.SliderNS,
                ElementDefault.SliderEW,
                ElementDefault.Fake,
                ElementDefault.Invisible,
                ElementDefault.BlinkWall,
                ElementDefault.Transporter,
                ElementDefault.Line,
                ElementDefault.Ricochet,
                ElementDefault.BlinkRayH,
                ElementDefault.Bear,
                ElementDefault.Ruffian,
                ElementDefault.Object,
                ElementDefault.Slime,
                ElementDefault.Shark,
                ElementDefault.SpinningGun,
                ElementDefault.Pusher,
                ElementDefault.Lion,
                ElementDefault.Tiger,
                ElementDefault.BlinkRayV,
                ElementDefault.Head,
                ElementDefault.Segment,
                ElementDefault.None,
                ElementDefault.BlueText,
                ElementDefault.GreenText,
                ElementDefault.CyanText,
                ElementDefault.RedText,
                ElementDefault.PurpleText,
                ElementDefault.BrownText,
                ElementDefault.BlackText
            }
        )
    End Function

    Public Shared Function SuperZZT() As Library
        Return New Library(
            {
                ElementDefault.Empty,
                ElementDefault.BoardEdge,
                ElementDefault.MonitorSuperZZT,
                ElementDefault.Messenger,
                ElementDefault.Player,
                ElementDefault.Ammo,
                ElementDefault.None,
                ElementDefault.Gem,
                ElementDefault.Key,
                ElementDefault.Door,
                ElementDefault.Scroll,
                ElementDefault.Passage,
                ElementDefault.Duplicator,
                ElementDefault.Bomb,
                ElementDefault.Energizer,
                ElementDefault.None,
                ElementDefault.Clockwise,
                ElementDefault.Counter,
                ElementDefault.None,
                ElementDefault.Lava,
                ElementDefault.Forest,
                ElementDefault.Solid,
                ElementDefault.Normal,
                ElementDefault.Breakable,
                ElementDefault.Boulder,
                ElementDefault.SliderNS,
                ElementDefault.SliderEW,
                ElementDefault.Fake,
                ElementDefault.Invisible,
                ElementDefault.BlinkWall,
                ElementDefault.Transporter,
                ElementDefault.Line,
                ElementDefault.Ricochet,
                ElementDefault.None,
                ElementDefault.BearSuperZZT,
                ElementDefault.Ruffian,
                ElementDefault.Object,
                ElementDefault.Slime,
                ElementDefault.None,
                ElementDefault.SpinningGun,
                ElementDefault.Pusher,
                ElementDefault.Lion,
                ElementDefault.Tiger,
                ElementDefault.None,
                ElementDefault.Head,
                ElementDefault.Segment,
                ElementDefault.None,
                ElementDefault.Floor,
                ElementDefault.WaterN,
                ElementDefault.WaterS,
                ElementDefault.WaterW,
                ElementDefault.WaterE,
                ElementDefault.None,
                ElementDefault.None,
                ElementDefault.None,
                ElementDefault.None,
                ElementDefault.None,
                ElementDefault.None,
                ElementDefault.None,
                ElementDefault.Roton,
                ElementDefault.DragonPup,
                ElementDefault.Pairer,
                ElementDefault.Spider,
                ElementDefault.Web,
                ElementDefault.Stone,
                ElementDefault.None,
                ElementDefault.None,
                ElementDefault.None,
                ElementDefault.None,
                ElementDefault.Bullet,
                ElementDefault.BlinkRayH,
                ElementDefault.BlinkRayV,
                ElementDefault.Star,
                ElementDefault.BlueText,
                ElementDefault.GreenText,
                ElementDefault.CyanText,
                ElementDefault.RedText,
                ElementDefault.PurpleText,
                ElementDefault.BrownText,
                ElementDefault.BlackText
            }
        )
    End Function

    Public Shared Function Extended() As Library
        Return New Library(
            {
                ElementDefault.Empty,
                ElementDefault.BoardEdge,
                ElementDefault.Monitor,
                ElementDefault.Messenger,
                ElementDefault.Player,
                ElementDefault.Ammo,
                ElementDefault.Torch,
                ElementDefault.Gem,
                ElementDefault.Key,
                ElementDefault.Door,
                ElementDefault.Scroll,
                ElementDefault.Passage,
                ElementDefault.Duplicator,
                ElementDefault.Bomb,
                ElementDefault.Energizer,
                ElementDefault.Star,
                ElementDefault.Clockwise,
                ElementDefault.Counter,
                ElementDefault.Bullet,
                ElementDefault.Water,
                ElementDefault.Forest,
                ElementDefault.Solid,
                ElementDefault.Normal,
                ElementDefault.Breakable,
                ElementDefault.Boulder,
                ElementDefault.SliderNS,
                ElementDefault.SliderEW,
                ElementDefault.Fake,
                ElementDefault.Invisible,
                ElementDefault.BlinkWall,
                ElementDefault.Transporter,
                ElementDefault.Line,
                ElementDefault.Ricochet,
                ElementDefault.BlinkRayH,
                ElementDefault.Bear,
                ElementDefault.Ruffian,
                ElementDefault.Object,
                ElementDefault.Slime,
                ElementDefault.Shark,
                ElementDefault.SpinningGun,
                ElementDefault.Pusher,
                ElementDefault.Lion,
                ElementDefault.Tiger,
                ElementDefault.BlinkRayV,
                ElementDefault.Head,
                ElementDefault.Segment,
                ElementDefault.Floor,
                ElementDefault.WaterN,
                ElementDefault.WaterS,
                ElementDefault.WaterW,
                ElementDefault.WaterE,
                ElementDefault.Roton,
                ElementDefault.DragonPup,
                ElementDefault.Pairer,
                ElementDefault.Spider,
                ElementDefault.Web,
                ElementDefault.Stone,
                ElementDefault.BlueText,
                ElementDefault.GreenText,
                ElementDefault.CyanText,
                ElementDefault.RedText,
                ElementDefault.PurpleText,
                ElementDefault.BrownText,
                ElementDefault.BlackText
            }
        )
    End Function
End Class
