﻿' RotonEnvironment
'
' Contains the Environment class which simulates an instance of the ZZT engine.



Public Enum EnvironmentGameType As Integer
    None = 0
    ZZT = 1
    SuperZZT = 2
    Extended = 3
End Enum

Public NotInheritable Class PascalError
    Public Shared ReadOnly Runtime202 As New StackOverflowException(
        "Runtime error 202.") 'stack overflow
    Public Shared ReadOnly Runtime204 As New AccessViolationException(
        "Runtime error 204.") 'invalid pointer operation
End Class

Public Enum EnvironmentGameExecutionState As Integer
    Init = 0
    TitleScreenInit
    TitleScreenPreLoop
    TitleScreen
    GameInit
    Game
    Quit
End Enum

Public Class Environment
    Public State As New EnvironmentGameState
    Public Parameters As New EnvironmentParameters

    Public SoundBank As SoundBank
    Public Library As Library
    Public World As World
    Public Boards As New List(Of Board)
    Private _Audio As Audio
    Private _Video As Video
    Public Input As Input
    Public Scroll As Scroll
    Public ScrollOwner As Thing
    Public Messenger As Messenger
    Public [Type] As EnvironmentGameType
    Public HighScores As HighScoreList
    Public Cache As New FileCache
    Public Alerts As New Alerts

    Protected Shared ReadOnly TextEncoding As System.Text.Encoding = System.Text.Encoding.GetEncoding(437)

    Protected RNG As Random
    Protected _RNGSeed As Integer
    Public Property RNGSeed As Integer
        Get
            Return _RNGSeed
        End Get
        Set(ByVal value As Integer)
            _RNGSeed = value
            RNG = New Random(value)
        End Set
    End Property

    Public Sub New()
        _Audio = New AudioNull
        _Video = New VideoNull
        Input = New InputNull
        Scroll = New ScrollNull
        Messenger = New MessengerNull
        State = New EnvironmentGameState
        RNGSeed = CInt(Now.Ticks And &H7FFFFFFF)
    End Sub

    Public Property Audio As Audio
        Get
            Return _Audio
        End Get
        Set(ByVal value As Audio)
            _Audio = value
        End Set
    End Property

    Public Property Video As Video
        Get
            Return _Video
        End Get
        Set(ByVal value As Video)
            _Video = value
            _Video.SetSize(Width, Height)
            _Video.Write(0, 0, "Video initialized", &H8F)
        End Set
    End Property

    Public ReadOnly Property CurrentBoard As Board
        Get
            Return Boards(World.Board)
        End Get
    End Property
End Class

