﻿' RotonExecution
'
' Contains all main and inner loop code.



Partial Public Class Environment

    Public Function Execute() As Boolean
        Dim Result As Boolean

        If (Debugger.IsAttached) Then
            ' debugger is enabled, disable error trapping
            Result = ExecuteInit()
        Else
            Try
                Result = ExecuteInit()
            Catch ex As Exception
                StopSound()
                State.QuitZZT = True
                Throw ex
                Return False
            End Try
        End If

        ProcessSound()
        Return Result
    End Function

    Protected Function ExecuteMain() As Boolean
        Dim UpdatedScreen As Boolean = False
        Dim TargetX As Integer
        Dim TargetY As Integer
        Dim E As Element

        If (State.QuitZZT) Then
            Return UpdatedScreen
        End If

        If (State.PlayerElement.ID <> ElementID.Player) OrElse ((State.PlayerElement.ID = ElementID.Player) AndAlso (World.Health > 0)) Then
            State.GameWaitTime = State.GameSpeed << 1
        End If

        If (State.GameWaitTime = 0) Then
            State.GameWaitTime = 1
        End If

        If (Scroll.Visible) Then
            ReadInput()
            Scroll.Update()
            If (Not Scroll.Visible) Then
                If (Scroll.Text.Length > &H8000) Then
                    ' scroll too long
                    Throw PascalError.Runtime204
                End If
                If (Scroll.LineIndex >= 0) Then
                    ExecuteScrollClose(Scroll.LineIndex)
                End If
            End If
            Return (State.KeyPressed <> 0) OrElse (Scroll.Animating)
        End If

        If (State.GamePaused) Then
            E = Library.GetElementFromID(ElementID.Player)
            If (State.MainTimer Mod 25 = 0) Then
                State.GamePausedBlink = Not State.GamePausedBlink
                If (State.GamePausedBlink) Then
                    Video.Plot(PlayerX, PlayerY, New AnsiChar(E.Char, E.Color))
                Else
                    With CurrentBoard.Things(0)
                        If (ElementAt(.X, .Y).ID = ElementID.Player) Then
                            Video.Plot(.X, .Y, New AnsiChar(32, 0))
                        Else
                            Video.Plot(.X, .Y, New AnsiChar(ElementAt(.X, .Y).Char, ColorAt(.X, .Y)))
                        End If
                    End With
                End If
                UpdatedScreen = True
            End If

            ReadInput()
            If (Input.GetKey = &H1B) Then
                RaiseQuitGame()
                StopSound()
            End If

            If (State.KeyVectorX <> 0) OrElse (State.KeyVectorY <> 0) Then
                TargetX = PlayerX + State.KeyVectorX
                TargetY = PlayerY + State.KeyVectorY
                Interact(TargetX, TargetY, 0, State.KeyVectorX, State.KeyVectorY)
                UpdatedScreen = True
            End If

            If (State.KeyVectorX <> 0) OrElse (State.KeyVectorY <> 0) Then
                TargetX = PlayerX + State.KeyVectorX
                TargetY = PlayerY + State.KeyVectorY
                If (ElementAt(TargetX, TargetY).Floor) Then
                    If (Parameters.BuggyPassages) Then
                        If (ElementAt(PlayerX, PlayerY).ID = ElementID.Player) Then
                            MoveThing(0, TargetX, TargetY)
                        Else
                            TileAt(TargetX, TargetY) = New Tile(E, E.Color)
                            UpdateAt(PlayerX, PlayerY)
                            CurrentBoard.Things(0).X = TargetX
                            CurrentBoard.Things(0).Y = TargetY
                            UpdateAt(PlayerX, PlayerY)
                            Radius(PlayerX, PlayerY, RadiusType.Update)
                            Radius(PlayerX - State.KeyVectorX, PlayerY - State.KeyVectorY, RadiusType.Update)
                        End If
                    Else
                        If (ElementAt(PlayerX, PlayerY).ID = ElementID.Player) Then
                            MoveThing(0, TargetX, TargetY)
                        Else
                            UpdateAt(PlayerX, PlayerY)
                            CurrentBoard.Things(0).X = TargetX
                            CurrentBoard.Things(0).Y = TargetY
                            CurrentBoard.Things(0).UnderElement = ElementAt(TargetX, TargetY)
                            CurrentBoard.Things(0).UnderColor = ColorAt(TargetX, TargetY)
                            ElementAt(TargetX, TargetY) = Library.GetElementFromID(ElementID.Player)
                            ColorAt(TargetX, TargetY) = ElementAt(TargetX, TargetY).Color
                            UpdateAt(TargetX, TargetY)
                            Radius(TargetX, TargetY, RadiusType.Update)
                            Radius(TargetX - State.KeyVectorX, TargetY - State.KeyVectorY, RadiusType.Update)
                        End If
                    End If
                    State.GamePaused = False
                    State.GameTick = Random(100)
                    State.ActIndex = CurrentBoard.Things.Count + 1
                    World.SavedGame = True
                    UpdatedScreen = True
                    StatusRedraw()
                End If
            End If
        End If

        If (State.ActIndex < ThingCount) OrElse ((State.MainTimer Mod State.GameWaitTime) = 0) Then

            If (State.ScrollOwnerIndex >= 0) Then
                State.ActIndex = State.ScrollOwnerIndex
                State.ScrollOwnerIndex = -1
            End If

            If (State.ActIndex >= ThingCount) Then
                State.GameTick += 1
                If (State.GameTick > &H1A4) Then
                    State.GameTick = 1
                End If
                State.ActIndex = 0
                ReadInput()
            End If

            Do Until (State.ActIndex >= ThingCount) OrElse (Scroll.Visible) OrElse (State.BreakGameLoop) OrElse (State.GamePaused)
                If (State.ActIndex < ThingCount) Then
                    With CurrentBoard.Things(State.ActIndex)
                        If (.Cycle <> 0) Then
                            If (State.ActIndex Mod .Cycle) = (State.GameTick Mod .Cycle) Then
                                Act(State.ActIndex)
                            End If
                        End If
                    End With
                    If (Not Scroll.Visible) Then
                        State.ActIndex += 1
                    Else
                        State = State
                    End If
                End If
            Loop

            If (State.ActIndex >= ThingCount) Then
                If (Messenger.TicksLeft > 0) Then
                    Messenger.Update()
                    If (Messenger.TicksLeft = 0) Then
                        UpdateBorder()
                    End If
                End If
            End If

            If (State.BreakGameLoop) Then
                Select Case State.PlayerElement.ID
                    Case ElementID.Player
                        If (State.QuitGame) Then
                            If (World.Health <= 0) Then
                                ' show high score dialog
                            End If
                            StopSound()
                        End If
                    Case ElementID.Monitor
                        RaiseEvent RedrawStatus()
                End Select
                State.DisableSound = False
            End If

            UpdatedScreen = True
        End If

        If (Not State.GamePaused) AndAlso (CurrentBoard.TimeLimit > 0) Then
            State.TimeLimitTicks += 1
        End If

        Return UpdatedScreen
    End Function

    Protected Function ExecuteInit() As Boolean
        Dim UpdatedScreen As Boolean = False
        Dim WorldStream As IO.Stream
        Dim NewEnv As Environment

        ' increment main timer
        State.MainTimer += 1

        ' if fading, process that
        If (State.FadeMode <> EnvironmentFadeMode.None) Then
            ExecuteFade()
            Return True
        End If

        ' blinks are processed here
        If (State.MainTimer Mod 20 = 0) Then
            Video.ToggleBlink()
            UpdatedScreen = True
        End If

        If (State.EditorMode) Then
            State.GameWaitTime = State.GameSpeed << 1
            If ((State.MainTimer Mod State.GameWaitTime) = 0) Then
                State.GameTick += 1
                If (State.GameTick > &H1A4) Then
                    State.GameTick = 1
                End If
                For State.ActIndex = 0 To ThingCount - 1
                    With CurrentBoard.Things(State.ActIndex)
                        If (.Cycle <> 0) Then
                            If (State.ActIndex Mod .Cycle) = (State.GameTick Mod .Cycle) Then
                                UpdateAt(.X, .Y)
                            End If
                        End If
                    End With
                Next
                UpdatedScreen = True
            End If
            Return UpdatedScreen
        End If

        State.BreakGameLoop = False

        Select Case State.Execution
            Case EnvironmentGameExecutionState.Init
                LoadHighScores()
                State.StartBoard = World.Board
                State.PlayerElement = Library.GetElementFromID(ElementID.Monitor)
                SetBoard(0)
                RedrawBoard()
                State.Execution = EnvironmentGameExecutionState.TitleScreenInit
                UpdatedScreen = True
                StatusRedraw()
                BoardFade()
                ExecuteInit()
            Case EnvironmentGameExecutionState.TitleScreenInit
                State.PlayerElement = Library.GetElementFromID(ElementID.Monitor)
                State.BreakGameLoop = False
                State.GameTick = Random(100)
                State.ActIndex = CurrentBoard.Things.Count + 1
                State.Execution = EnvironmentGameExecutionState.TitleScreenPreLoop
                UpdatedScreen = UpdatedScreen Or ExecuteInit()
            Case EnvironmentGameExecutionState.TitleScreenPreLoop
                TileAt(PlayerX, PlayerY) = New Tile(State.PlayerElement, State.PlayerElement.Color)
                UpdateAt(PlayerX, PlayerY)
                State.GamePaused = False
                State.Execution = EnvironmentGameExecutionState.TitleScreen
                StatusRedraw()
                UpdatedScreen = UpdatedScreen Or ExecuteInit()
            Case EnvironmentGameExecutionState.TitleScreen, EnvironmentGameExecutionState.Game
                UpdatedScreen = UpdatedScreen Or ExecuteMain()
                If (State.BreakGameLoop) Then
                    Select Case State.KeyPressed
                        Case &H57 'W
                            RaiseEvent LoadWorld(Me)
                        Case &H50 'P
                            If (World.SavedGame) Then
                                WorldStream = Cache.LoadFileStream(World.Name & "." & Parameters.WorldExtension)
                                NewEnv = File.Load(WorldStream)
                                ApplyEnvironment(NewEnv)
                            End If
                            State.Execution = EnvironmentGameExecutionState.GameInit
                            StopSound()
                            SetBoard(State.StartBoard)
                            ExecuteInit()
                            UpdatedScreen = True
                        Case &H41 'A
                            ExecuteHelp("ABOUT")
                        Case &H45 'E
                            'editor
                        Case &H53 'S
                            RaiseEvent SetGameSpeed()
                        Case &H52 'R
                            RaiseEvent LoadGame(Me)
                        Case &H48 'H
                            ExecuteHelp("GAME")
                        Case 191 'cheat
                            RaiseEvent EnterCheat()
                        Case &H1B, &H51 'esc, Q
                            RaiseQuitZZT()
                            StopSound()
                    End Select
                    If (State.Execution = EnvironmentGameExecutionState.TitleScreen) Then
                        State.Execution = EnvironmentGameExecutionState.TitleScreenPreLoop
                    Else
                        ExecuteInit()
                        UpdatedScreen = True
                    End If
                End If
                If (State.QuitGame) Then
                    StopSound()
                    State.QuitGame = False
                    State.Execution = EnvironmentGameExecutionState.Init
                    UpdatedScreen = True
                    ExecuteInit()
                End If
            Case EnvironmentGameExecutionState.GameInit
                State.PlayerElement = Library.GetElementFromID(ElementID.Player)
                State.GamePaused = True
                State.Execution = EnvironmentGameExecutionState.Game
                BoardFade()
                EnterBoard()
                RedrawBoard()
                UpdatedScreen = True
                StatusRedraw()
                ExecuteInit()
            Case EnvironmentGameExecutionState.Quit
                ' return to title screen
        End Select

        Return UpdatedScreen
    End Function

    Protected Sub ExecuteFade()
        Dim Index As Integer
        Dim Max As Integer = State.FadeMatrix.Length
        Dim EmptyChar As New AnsiChar(0, 0)

        Select Case State.FadeMode
            Case EnvironmentFadeMode.None
            Case EnvironmentFadeMode.FadeOut
                For Index = 1 To Parameters.FadeBlocksPerTick
                    With State.FadeMatrix(State.FadeIndex)
                        Video.PlotOverlay(.X, .Y, Parameters.FadeChar)
                    End With
                    State.FadeIndex += 1
                    If (State.FadeIndex = Max) Then
                        State.FadeIndex = 0
                        State.FadeMode = EnvironmentFadeMode.FadeIn
                        Exit For
                    End If
                Next
            Case EnvironmentFadeMode.FadeIn
                For Index = 1 To Parameters.FadeBlocksPerTick
                    With State.FadeMatrix(State.FadeIndex)
                        Video.PlotOverlay(.X, .Y, EmptyChar)
                    End With
                    State.FadeIndex += 1
                    If (State.FadeIndex = Max) Then
                        State.FadeIndex = -1
                        State.FadeMode = EnvironmentFadeMode.None
                        If (State.EnteringBoard) Then
                            EnterBoardPostFade()
                            State.EnteringBoard = False
                        End If
                        Exit For
                    End If
                Next
        End Select
    End Sub

End Class
