﻿' RotonFileLoad
'
' Code for importing files to an Environment.

Imports System.IO



Partial Public NotInheritable Class File
    Private Shared ReadOnly Encoding As System.Text.Encoding = System.Text.Encoding.GetEncoding(437)
    Private Shared ReadOnly Unicode As System.Text.Encoding = System.Text.Encoding.Unicode

    Private Class ZZTBinaryReader
        Inherits BinaryReader

        Public Sub New(ByVal input As Stream)
            MyBase.New(input, Encoding)
        End Sub

        Public Sub New(ByVal Input As Stream, Enc As System.Text.Encoding)
            MyBase.New(Input, Enc)
        End Sub

        Public Function ReadZZTString(ByVal FixedLength As Integer) As String
            Dim Ret As String
            Dim Length As Byte

            Length = ReadByte()
            Ret = ReadChars(FixedLength)
            Return Ret.Substring(0, Length)
        End Function
    End Class

    Public Shared Function Load(ByVal Source As Stream) As Environment
        Dim Reader As New ZZTBinaryReader(Source)
        Dim ID As Int16
        Dim Ret As Environment

        ID = Reader.ReadInt16

        Select Case ID
            Case -1
                Ret = LoadZZT(Reader)
            Case -2
                Ret = LoadSuperZZT(Reader)
            Case -13
                Ret = LoadExtendedZZT(Reader)
            Case Else
                Ret = Nothing
        End Select

        If (Not (Ret Is Nothing)) Then
            ' force set koopo bug (for now)
            'Ret.Parameters.KoopoBug = False

            With Ret.State
                If (Ret.World.SavedGame) Then
                    .TitleScreen = False
                    .Execution = EnvironmentGameExecutionState.GameInit
                Else
                    .TitleScreen = True
                    .Execution = EnvironmentGameExecutionState.Init
                End If
                .WorldLoaded = True
            End With

        End If

        Return Ret
    End Function

    Private Shared Function LoadZZT(ByVal SourceReader As ZZTBinaryReader) As Environment
        Dim Data() As Byte
        Dim DataStream As MemoryStream
        Dim Reader As ZZTBinaryReader
        Dim Ret As New Environment
        Dim BoardCount As Integer
        Dim Index As Integer
        Dim Library As Library
        Dim B As Board

        Library = LibraryDefault.ZZT

        Data = SourceReader.ReadBytes(&H200 - 2)
        DataStream = New MemoryStream(Data)
        Reader = New ZZTBinaryReader(DataStream)

        BoardCount = Reader.ReadInt16
        Ret.Library = Library
        Ret.World = LoadZZTWorld(Reader)
        Ret.SoundBank = SoundBankDefault.ZZT

        For Index = 0 To BoardCount
            B = LoadZZTBoard(SourceReader, Library)
            If (B.Things.Count > 0) Then
                B.TileAt(B.Things(0).X, B.Things(0).Y) = New Tile(Library.GetElementFromID(ElementID.Player), Library.GetElementFromID(ElementID.Player).Color)
            End If
            Ret.Boards.Add(B)
        Next

        Ret.Type = EnvironmentGameType.ZZT
        Ret.Parameters = EnvironmentParameters.ZZT

        Select Case Ret.World.Name
            Case "BQUE", "BQUEST", "BQUEST-B", "BQUEST-C"
                'Ret.SetCodeSet(Environment.OOPCodeSets.BQuest)
                Ret.OOPCodeSet = Environment.CommandSetDefaults.BananaQuest
                Ret.Parameters.DoorLockedMessage = "The %1 Banana required"
                Ret.Parameters.DoorOpenMessage = "The %1 Banana was given."
                Ret.Parameters.KeyMessage = "You have the Ban-%1-ana."
                Ret.Parameters.KeyAlreadyMessage = "Can't carry 2 Ban-%1-ana!"
        End Select

        DataStream.Dispose()

        Return Ret
    End Function

    Private Shared Function LoadZZTWorld(ByVal Reader As ZZTBinaryReader) As World
        Dim Ret As New World
        Dim Index As Integer
        Dim Flag As String

        With Ret
            .Ammo = Reader.ReadInt16
            .Gems = Reader.ReadInt16
            ReDim .Keys(6)
            For Index = 0 To 6
                .Keys(Index) = (Reader.ReadByte <> 0)
            Next
            .Health = Reader.ReadInt16
            .Board = Reader.ReadInt16
            .Torches = Reader.ReadInt16
            .TorchCycles = Reader.ReadInt16
            .EnergyCycles = Reader.ReadInt16
            Reader.ReadInt16()
            .Score = Reader.ReadInt16
            .Name = Reader.ReadZZTString(20)
            For Index = 0 To 9
                Flag = Reader.ReadZZTString(20)
                If Flag <> "" Then
                    .Flags.Add(Flag)
                End If
            Next
            .TimePassed = Reader.ReadInt16
            Reader.ReadInt16()
            .SavedGame = (Reader.ReadByte <> 0)
        End With

        Return Ret
    End Function

    Private Shared Function LoadZZTBoard(ByVal SourceReader As ZZTBinaryReader, ByVal Library As Library) As Board
        Dim Data() As Byte
        Dim DataStream As MemoryStream
        Dim Reader As ZZTBinaryReader
        Dim BoardSize As Int16
        Dim Ret As Board
        Dim ThingCount As Integer
        Dim Index As Integer
        Dim Thing As Thing
        Dim ThingCode As String
        Dim CodeContainer As CodeContainer

        With EnvironmentParameters.ZZT
            Ret = New Board(.BoardWidth, .BoardHeight, .MaxThings, Library.GetElementFromID(ElementID.BoardEdge), Library.GetElementFromID(ElementID.Empty), Library.GetElementFromID(ElementID.Player))
        End With

        BoardSize = SourceReader.ReadInt16
        Data = SourceReader.ReadBytes(BoardSize)
        DataStream = New MemoryStream(Data)
        Reader = New ZZTBinaryReader(DataStream)

        With Ret
            .Name = Reader.ReadZZTString(50)
            Try
                .Tiles = LoadRLE(Reader, Ret.Area, Library)
                .MaxShots = Reader.ReadByte
                .IsDark = (Reader.ReadByte <> 0)
                .ExitN = Reader.ReadByte
                .ExitS = Reader.ReadByte
                .ExitW = Reader.ReadByte
                .ExitE = Reader.ReadByte
                .RestartZap = (Reader.ReadByte <> 0)
                Reader.ReadBytes(59)
                .EnterX = Reader.ReadByte
                .EnterY = Reader.ReadByte
                .TimeLimit = Reader.ReadInt16
                Reader.ReadBytes(16)
                If (.Tiles Is Nothing) Then
                    .Clear()
                    Return Ret
                End If
                ThingCount = Reader.ReadInt16
                For Index = 0 To ThingCount
                    Thing = LoadThing(Reader, Library)
                    Reader.ReadBytes(8)
                    If (Thing.Length > 0) Then
                        ThingCode = Reader.ReadChars(Thing.Length)
                        CodeContainer = New CodeContainer(ThingCode)
                        Thing.Code = CodeContainer
                    End If
                    .Things.Add(Thing)
                Next
                LoadThingPointers(Ret)
            Catch ex As Exception
                Return Ret
            End Try
        End With

        DataStream.Dispose()

        Return Ret
    End Function

    Private Shared Function LoadSuperZZT(ByVal SourceReader As ZZTBinaryReader) As Environment
        Dim Data() As Byte
        Dim DataStream As MemoryStream
        Dim Reader As ZZTBinaryReader
        Dim Ret As New Environment
        Dim BoardCount As Integer
        Dim Index As Integer
        Dim Library As Library
        Dim B As Board

        Library = LibraryDefault.SuperZZT

        Data = SourceReader.ReadBytes(&H400 - 2)
        DataStream = New MemoryStream(Data)
        Reader = New ZZTBinaryReader(DataStream)

        BoardCount = Reader.ReadInt16
        Ret.Library = Library
        Ret.World = LoadSuperZZTWorld(Reader)
        Ret.SoundBank = SoundBankDefault.SuperZZT

        For Index = 0 To BoardCount
            B = LoadSuperZZTBoard(SourceReader, Library)
            B.TileAt(B.Things(0).X, B.Things(0).Y) = New Tile(Library.GetElementFromID(ElementID.Player), Library.GetElementFromID(ElementID.Player).Color)
            Ret.Boards.Add(B)
        Next

        Ret.Type = EnvironmentGameType.SuperZZT
        Ret.Parameters = EnvironmentParameters.SuperZZT

        DataStream.Dispose()

        Return Ret
    End Function

    Private Shared Function LoadSuperZZTWorld(ByVal Reader As ZZTBinaryReader) As World
        Dim Ret As New World
        Dim Index As Integer
        Dim Flag As String

        With Ret
            .Ammo = Reader.ReadInt16
            .Gems = Reader.ReadInt16
            ReDim .Keys(6)
            For Index = 0 To 6
                .Keys(Index) = (Reader.ReadByte <> 0)
            Next
            .Health = Reader.ReadInt16
            .Board = Reader.ReadInt16
            Reader.ReadInt16()
            .EnergyCycles = Reader.ReadInt16
            Reader.ReadInt16()
            .Score = Reader.ReadInt16
            .Name = Reader.ReadZZTString(20)
            For Index = 0 To 15
                Flag = Reader.ReadZZTString(20)
                If Flag <> "" Then
                    .Flags.Add(Flag)
                End If
            Next
            .TimePassed = Reader.ReadInt16
            Reader.ReadInt16()
            .SavedGame = (Reader.ReadByte <> 0)
            .Stones = Reader.ReadInt16
        End With

        Return Ret
    End Function

    Private Shared Function LoadSuperZZTBoard(ByVal SourceReader As ZZTBinaryReader, ByVal Library As Library) As Board
        Dim Data() As Byte
        Dim DataStream As MemoryStream
        Dim Reader As ZZTBinaryReader
        Dim BoardSize As Int16
        Dim Ret As Board
        Dim ThingCount As Integer
        Dim Index As Integer
        Dim Thing As Thing
        Dim ThingCode As String
        Dim CodeContainer As CodeContainer

        With EnvironmentParameters.SuperZZT
            Ret = New Board(.BoardWidth, .BoardHeight, .MaxThings, Library.GetElementFromID(ElementID.BoardEdge), Library.GetElementFromID(ElementID.Empty), Library.GetElementFromID(ElementID.Player))
        End With

        BoardSize = SourceReader.ReadInt16
        Data = SourceReader.ReadBytes(BoardSize)
        DataStream = New MemoryStream(Data)
        Reader = New ZZTBinaryReader(DataStream)

        With Ret
            .Name = Reader.ReadZZTString(60)
            Try
                .Tiles = LoadRLE(Reader, Ret.Area, Library)
                .MaxShots = Reader.ReadByte
                .ExitN = Reader.ReadByte
                .ExitS = Reader.ReadByte
                .ExitW = Reader.ReadByte
                .ExitE = Reader.ReadByte
                .RestartZap = (Reader.ReadByte <> 0)
                .EnterX = Reader.ReadByte
                .EnterY = Reader.ReadByte
                Reader.ReadInt16()
                Reader.ReadInt16()
                .TimeLimit = Reader.ReadInt16
                Reader.ReadBytes(14)
                If (.Tiles Is Nothing) Then
                    .Clear()
                    Return Ret
                End If
                ThingCount = Reader.ReadInt16
                For Index = 0 To ThingCount
                    Thing = LoadThing(Reader, Library)
                    If (Thing.Length > 0) Then
                        ThingCode = Reader.ReadChars(Thing.Length)
                        CodeContainer = New CodeContainer(ThingCode)
                        Thing.Code = CodeContainer
                    End If
                    .Things.Add(Thing)
                Next
                LoadThingPointers(Ret)
            Catch ex As Exception
                Return Ret
            End Try
        End With

        DataStream.Dispose()

        Return Ret
    End Function

    Private Shared Function LoadExtendedZZT(ByVal Reader As BinaryReader) As Environment
        Return Nothing
    End Function

    Private Shared Function LoadExtendedZZTBoard(ByVal Reader As BinaryReader, ByVal Library As Library) As Board
        Return Nothing
    End Function

    Private Shared Function LoadExtendedZZTWorld(ByVal Reader As BinaryReader) As World
        Return Nothing
    End Function

    Private Shared Function LoadThing(ByVal Source As BinaryReader, ByVal Library As Library) As Thing
        Dim Ret As New Thing
        With Ret
            .X = CInt(Source.ReadByte) - 1
            .Y = CInt(Source.ReadByte) - 1
            .XStep = Source.ReadInt16
            .YStep = Source.ReadInt16
            .Cycle = Source.ReadInt16
            .P1 = Source.ReadByte
            .P2 = Source.ReadByte
            .P3 = Source.ReadByte
            .Follower = Source.ReadInt16
            .Leader = Source.ReadInt16
            .UnderElement = Library.GetElementFromIndex(Source.ReadByte)
            .UnderColor = Source.ReadByte
            Source.ReadInt32()
            .CurrentInstruction = Source.ReadInt16
            .Length = Source.ReadInt16
        End With
        Return Ret
    End Function

    Private Shared Sub LoadThingPointers(ByRef Board As Board)
        Dim Index As Integer
        Dim T As Thing

        With Board
            For Index = 0 To .Things.Count - 1
                T = .Things(Index)
                If (T.Length < 0) Then
                    T.Code = .Things(-T.Length).Code
                    T.Length = .Things(-T.Length).Length
                End If
            Next
        End With
    End Sub

    Private Shared Function LoadExtendedThing(ByVal Source As BinaryReader, ByVal Library As Library) As Thing
        Return Nothing
    End Function

    Private Shared Function LoadRLE(ByVal Source As BinaryReader, ByVal Area As Integer, ByVal Library As Library) As Tile()
        Dim Ret As New List(Of Tile)
        Dim Index As Integer
        Dim Length As Integer
        Dim Code As Byte
        Dim Color As Byte

        Do While (Index < Area)
            If (Length = 0) Then
                Length = Source.ReadByte
                Code = Source.ReadByte
                Color = Source.ReadByte
                If (Length = 0) Then
                    Length = 256
                End If
            End If
            Ret.Add(New Tile(Library.GetElementFromIndex(Code), Color))
            Length -= 1
            Index += 1
        Loop

        If (Length > 0) Then
            ' board is corrupted
            Return Nothing
        End If

        Return Ret.ToArray
    End Function

    Public Shared Function ImportBoard(ByVal Source As Stream, ByVal T As EnvironmentGameType) As Board
        Select Case T
            Case EnvironmentGameType.ZZT
                Return ImportZZTBoard(Source)
            Case EnvironmentGameType.SuperZZT
                Return ImportSuperZZTBoard(Source)
        End Select
        Return Nothing
    End Function

    Private Shared Function ImportZZTBoard(ByVal Source As Stream) As Board
        Dim Reader As New ZZTBinaryReader(Source)
        Return LoadZZTBoard(Reader, LibraryDefault.ZZT)
    End Function

    Private Shared Function ImportSuperZZTBoard(ByVal Source As Stream) As Board
        Dim Reader As New ZZTBinaryReader(Source)
        Return LoadSuperZZTBoard(Reader, LibraryDefault.SuperZZT)
    End Function
End Class


