﻿' RotonFileSave
'
' Code to export Environment to a file.

Imports System.IO



Partial Public NotInheritable Class File
    Private Class ZZTBinaryWriter
        Inherits BinaryWriter

        Public Sub New(ByVal output As Stream)
            MyBase.New(output, Encoding)
        End Sub

        Public Sub New(ByVal output As Stream, Enc As System.Text.Encoding)
            MyBase.New(output, Enc)
        End Sub

        Public Sub WriteZZTString(ByVal OutputString As String, ByVal FixedLength As Integer)
            If (OutputString Is Nothing) Then
                OutputString = ""
            End If
            Dim Length As Byte = CByte(OutputString.Length)
            Dim Chars() As Char = CType(OutputString, Char())
            ReDim Preserve Chars(FixedLength - 1)
            Write(Length)
            Write(Chars)
        End Sub

        Public Sub WriteByte(ByVal V As Int32)
            Dim B As Byte = CByte(V And &HFF)
            Write(B)
        End Sub

        Public Sub WriteInt16(ByVal V As Int32)
            Dim I As Int16 = CType((V << 16) >> 16, Int16)
            Write(I)
        End Sub

        Public Sub WriteInt32(ByVal V As Int32)
            Write(V)
        End Sub

        Public Sub WriteChars(ByVal S As String)
            Dim C() As Char = CType(S, Char())
            Write(C)
        End Sub

        Public Sub WriteBoolean(ByVal V As Boolean)
            Dim Result As Byte
            If (V) Then
                Result = 1
            Else
                Result = 0
            End If
            Write(Result)
        End Sub

        Public Sub WriteNulls(ByVal Count As Integer)
            Dim Index As Integer
            Dim N As Byte = 0
            For Index = 0 To Count - 1
                Write(N)
            Next
        End Sub
    End Class

    Public Shared Sub Save(ByVal TargetStream As Stream, ByVal Source As Environment)
        Dim Target As New MemoryStream
        Dim Writer As New ZZTBinaryWriter(Target)

        Select Case Source.Type
            Case EnvironmentGameType.ZZT
                SaveZZT(Writer, Source)
            Case EnvironmentGameType.SuperZZT
                SaveSuperZZT(Writer, Source)
        End Select

        TargetStream.Write(Target.ToArray, 0, CInt(Target.Length))
        Target.Dispose()
    End Sub

    Private Shared Sub SaveZZT(ByVal Target As ZZTBinaryWriter, ByVal Source As Environment)
        Dim Index As Integer

        Target.WriteInt16(-1)
        Target.WriteInt16(Source.Boards.Count - 1)
        SaveZZTWorld(Target, Source.World)
        For Index = 0 To Source.Boards.Count - 1
            SaveZZTBoard(Target, Source.Boards(Index), Source.Library)
        Next
    End Sub

    Private Shared Sub SaveZZTWorld(ByVal Target As ZZTBinaryWriter, ByVal W As World)
        Dim Buffer(&H200 - 5) As Byte
        Dim Mem As New MemoryStream(Buffer)
        Dim Writer As New ZZTBinaryWriter(Mem)
        Dim Index As Integer
        Dim Flags() As String = W.Flags.ToArray
        ReDim Preserve Flags(9)

        Writer.BaseStream.Position = 0
        With W
            Writer.WriteInt16(.Ammo)
            Writer.WriteInt16(.Gems)
            For Index = 0 To 6
                Writer.WriteBoolean(.Keys(Index))
            Next
            Writer.WriteInt16(.Health)
            Writer.WriteInt16(.Board)
            Writer.WriteInt16(.Torches)
            Writer.WriteInt16(.TorchCycles)
            Writer.WriteInt16(.EnergyCycles)
            Writer.WriteInt16(0)
            Writer.WriteInt16(.Score)
            Writer.WriteZZTString(.Name, 20)
            For Index = 0 To 9
                Writer.WriteZZTString(Flags(Index), 20)
            Next
            Writer.WriteInt16(.TimePassed)
            Writer.WriteInt16(0)
            Writer.WriteBoolean(.SavedGame)
        End With

        Target.Write(Mem.ToArray)
        Mem.Dispose()
    End Sub

    Private Shared Sub SaveZZTBoard(ByVal Target As ZZTBinaryWriter, ByVal B As Board, ByVal Library As Library)
        Dim Mem As New MemoryStream
        Dim Writer As New ZZTBinaryWriter(Mem)
        Dim Index As Integer

        With B
            Writer.WriteZZTString(.Name, 50)
            SaveRLE(Writer, .Tiles, Library)
            Writer.WriteByte(.MaxShots)
            Writer.WriteBoolean(.IsDark)
            Writer.WriteByte(.ExitN)
            Writer.WriteByte(.ExitS)
            Writer.WriteByte(.ExitW)
            Writer.WriteByte(.ExitE)
            Writer.WriteBoolean(.RestartZap)
            Writer.WriteNulls(59)
            Writer.WriteByte(.EnterX)
            Writer.WriteByte(.EnterY)
            Writer.WriteInt16(.TimeLimit)
            Writer.WriteNulls(16)
            Writer.WriteInt16(.Things.Count - 1)
            SaveConsolidateThings(B)
            For Index = 0 To .Things.Count - 1
                SaveThing(Writer, .Things(Index), Library)
                Writer.WriteNulls(8)
                If (.Things(Index).Length > 0) Then
                    Writer.WriteChars(.Things(Index).Code.Code)
                End If
            Next
        End With

        Target.WriteInt16(CInt(Mem.Length))
        Target.Write(Mem.ToArray)
        Mem.Dispose()
    End Sub

    Private Shared Sub SaveSuperZZT(ByVal Target As ZZTBinaryWriter, ByVal Source As Environment)
        Dim Index As Integer

        Target.WriteInt16(-2)
        Target.WriteInt16(Source.Boards.Count - 1)
        SaveSuperZZTWorld(Target, Source.World)
        For Index = 0 To Source.Boards.Count - 1
            SaveSuperZZTBoard(Target, Source.Boards(Index), Source.Library)
        Next
    End Sub

    Private Shared Sub SaveSuperZZTWorld(ByVal Target As ZZTBinaryWriter, ByVal W As World)
        Dim Buffer(&H400 - 5) As Byte
        Dim Mem As New MemoryStream(Buffer)
        Dim Writer As New ZZTBinaryWriter(Mem)
        Dim Index As Integer
        Dim Flags() As String = W.Flags.ToArray
        ReDim Preserve Flags(15)

        Writer.BaseStream.Position = 0
        With W
            Writer.WriteInt16(.Ammo)
            Writer.WriteInt16(.Gems)
            For Index = 0 To 6
                Writer.WriteBoolean(.Keys(Index))
            Next
            Writer.WriteInt16(.Health)
            Writer.WriteInt16(.Board)
            Writer.WriteInt16(0)
            Writer.WriteInt16(.EnergyCycles)
            Writer.WriteInt16(0)
            Writer.WriteInt16(.Score)
            Writer.WriteZZTString(.Name, 20)
            For Index = 0 To 15
                Writer.WriteZZTString(Flags(Index), 20)
            Next
            Writer.WriteInt16(.TimePassed)
            Writer.WriteInt16(0)
            Writer.WriteBoolean(.SavedGame)
            Writer.WriteInt16(.Stones)
        End With

        Target.Write(Mem.ToArray)
        Mem.Dispose()
    End Sub

    Private Shared Sub SaveSuperZZTBoard(ByVal Target As ZZTBinaryWriter, ByVal B As Board, ByVal Library As Library)
        Dim Mem As New MemoryStream
        Dim Writer As New ZZTBinaryWriter(Mem)
        Dim Index As Integer

        With B
            Writer.WriteZZTString(.Name, 60)
            SaveRLE(Writer, .Tiles, Library)
            Writer.WriteByte(.MaxShots)
            Writer.WriteByte(.ExitN)
            Writer.WriteByte(.ExitS)
            Writer.WriteByte(.ExitW)
            Writer.WriteByte(.ExitE)
            Writer.WriteBoolean(.RestartZap)
            Writer.WriteByte(.EnterX)
            Writer.WriteByte(.EnterY)
            Writer.WriteInt16(0)
            Writer.WriteInt16(0)
            Writer.WriteInt16(.TimeLimit)
            Writer.WriteNulls(14)
            Writer.WriteInt16(.Things.Count - 1)
            SaveConsolidateThings(B)
            For Index = 0 To .Things.Count - 1
                SaveThing(Writer, .Things(Index), Library)
                If (.Things(Index).Length > 0) Then
                    Writer.WriteChars(.Things(Index).Code.Code)
                End If
            Next
        End With

        Target.WriteInt16(CInt(Mem.Length))
        Target.Write(Mem.ToArray)
        Mem.Dispose()
    End Sub

    Private Shared Sub SaveThing(ByVal Target As ZZTBinaryWriter, ByVal T As Thing, ByVal Library As Library)
        With T
            Target.WriteByte(.X + 1)
            Target.WriteByte(.Y + 1)
            Target.WriteInt16(.XStep)
            Target.WriteInt16(.YStep)
            Target.WriteInt16(.Cycle)
            Target.WriteByte(.P1)
            Target.WriteByte(.P2)
            Target.WriteByte(.P3)
            Target.WriteInt16(.Follower)
            Target.WriteInt16(.Leader)
            Target.WriteByte(Library.GetIDFromElement(.UnderElement.ID))
            Target.WriteByte(.UnderColor)
            Target.WriteInt32(0)
            Target.WriteInt16(.CurrentInstruction)
            Target.WriteInt16(.Length)
        End With
    End Sub

    Private Shared Sub SaveRLE(ByVal Target As ZZTBinaryWriter, ByVal Tiles() As Tile, ByVal Library As Library)
        Dim RLEColor As Byte
        Dim RLECount As Integer
        Dim RLEID As Byte
        Dim ThisColor As Byte
        Dim ThisID As Byte
        Dim Index As Integer
        Dim Max As Integer = Tiles.Length

        RLEColor = CByte(Tiles(0).Color)
        RLEID = Library.GetIDFromElement(Tiles(0).Element.ID)
        RLECount = 1
        Index = 1

        Do
            ThisColor = CByte(Tiles(Index).Color And 255)
            ThisID = Library.GetIDFromElement(Tiles(Index).Element.ID)
            If (ThisColor <> RLEColor) OrElse (ThisID <> RLEID) OrElse (RLECount = 255) Then
                Target.WriteByte(RLECount)
                Target.WriteByte(RLEID)
                Target.WriteByte(RLEColor)
                RLECount = 0
                RLEID = ThisID
                RLEColor = ThisColor
            End If
            RLECount += 1
            Index += 1
        Loop While (Index < Max)
        Target.WriteByte(RLECount)
        Target.WriteByte(RLEID)
        Target.WriteByte(RLEColor)
    End Sub

    Private Shared Sub SaveConsolidateThings(ByVal B As Board)
        Dim J As Integer
        Dim CodeMatch As Boolean
        Dim Index As Integer
        With B
            For Index = 0 To .Things.Count - 1
                If (.Things(Index).Code Is Nothing) Then
                    .Things(Index).Length = 0
                Else
                    CodeMatch = False
                    For J = 0 To Index - 1
                        If (Not (.Things(J).Code Is Nothing)) AndAlso (.Things(J).Code.Equals(.Things(Index).Code)) Then
                            .Things(Index).Length = -J
                            CodeMatch = True
                            Exit For
                        End If
                    Next
                    If (Not CodeMatch) Then
                        .Things(Index).Length = .Things(Index).Code.Code.Length
                    End If
                End If
            Next
        End With
    End Sub

    Public Shared Sub ExportBoard(ByVal Target As Stream, ByVal B As Board, ByVal T As EnvironmentGameType)
        Select Case T
            Case EnvironmentGameType.ZZT
                ExportZZTBoard(Target, B)
            Case EnvironmentGameType.SuperZZT
                ExportSuperZZTBoard(Target, B)
        End Select
    End Sub

    Private Shared Sub ExportZZTBoard(ByVal Target As Stream, ByVal B As Board)
        Dim Writer As New ZZTBinaryWriter(Target)
        SaveZZTBoard(Writer, B, LibraryDefault.ZZT)
    End Sub

    Private Shared Sub ExportSuperZZTBoard(ByVal Target As Stream, ByVal B As Board)
        Dim Writer As New ZZTBinaryWriter(Target)
        SaveSuperZZTBoard(Writer, B, LibraryDefault.SuperZZT)
    End Sub

End Class
