﻿' RotonInteract
'
' Contains all interaction code. This code is executed when the player "touches" tiles on the board.



Partial Public Class Environment

    Private Shared ReadOnly ColorStrings() As String = {
        "Black",
        "Blue",
        "Green",
        "Cyan",
        "Red",
        "Purple",
        "Yellow",
        "White"
    }

    Protected Sub Interact(ByVal X As Integer, ByVal Y As Integer, ByVal Index As Integer, ByRef XVector As Integer, ByRef YVector As Integer)
        Dim E As Element = ElementAt(X, Y)

        Select Case E.ID
            Case ElementID.Ammo
                InteractAmmo(X, Y, Index, XVector, YVector)
            Case ElementID.BoardEdge
                InteractBoardEdge(X, Y, Index, XVector, YVector)
            Case ElementID.Bomb
                InteractBomb(X, Y, Index, XVector, YVector)
            Case ElementID.Door
                InteractDoor(X, Y, Index, XVector, YVector)
            Case ElementID.Energizer
                InteractEnergizer(X, Y, Index, XVector, YVector)
            Case ElementID.Fake
                InteractFake(X, Y, Index, XVector, YVector)
            Case ElementID.Forest
                InteractForest(X, Y, Index, XVector, YVector)
            Case ElementID.Gem
                InteractGem(X, Y, Index, XVector, YVector)
            Case ElementID.Invisible
                InteractInvisible(X, Y, Index, XVector, YVector)
            Case ElementID.Key
                InteractKey(X, Y, Index, XVector, YVector)
            Case ElementID.Lava
                InteractLava(X, Y, Index, XVector, YVector)
            Case ElementID.Object
                InteractObject(X, Y, Index, XVector, YVector)
            Case ElementID.Passage
                InteractPassage(X, Y, Index, XVector, YVector)
            Case ElementID.Scroll
                InteractScroll(X, Y, Index, XVector, YVector)
            Case ElementID.Slime
                InteractSlime(X, Y, Index, XVector, YVector)
            Case ElementID.Stone
                InteractStone(X, Y, Index, XVector, YVector)
            Case ElementID.Torch
                InteractTorch(X, Y, Index, XVector, YVector)
            Case ElementID.Transporter
                InteractTransporter(X, Y, Index, XVector, YVector)
            Case ElementID.Water
                InteractWater(X, Y, Index, XVector, YVector)
            Case ElementID.SliderNS,
                ElementID.SliderEW,
                ElementID.Boulder
                InteractPuzzle(X, Y, Index, XVector, YVector)
            Case ElementID.Bear,
                ElementID.Bullet,
                ElementID.DragonPup,
                ElementID.Head,
                ElementID.Lion,
                ElementID.Pairer,
                ElementID.Roton,
                ElementID.Ruffian,
                ElementID.Segment,
                ElementID.Spider,
                ElementID.Star,
                ElementID.Tiger
                InteractEnemy(X, Y, Index, XVector, YVector)
        End Select
    End Sub

    Protected Sub InteractAmmo(ByVal X As Integer, ByVal Y As Integer, ByVal Index As Integer, ByRef XVector As Integer, ByRef YVector As Integer)
        World.Ammo += Parameters.AmmoPerPickup
        RemoveItem(X, Y)
        StatusUpdate()
        PlaySound(2, SoundBank.Ammo)
        If (Alerts.Ammo) Then
            SetMessage(&HC8, Parameters.AmmoMessage.Replace("%1", Parameters.AmmoPerPickup.ToString))
            Alerts.Ammo = False
        End If
        UpdateAt(X, Y)
    End Sub

    Protected Sub InteractBoardEdge(ByVal X As Integer, ByVal Y As Integer, ByVal Index As Integer, ByRef XVector As Integer, ByRef YVector As Integer)
        Dim TargetX As Integer
        Dim TargetY As Integer
        Dim TargetBoard As Integer
        Dim OldBoard As Integer
        Dim E As Element

        TargetX = X
        TargetY = Y
        If (YVector = -1) Then
            TargetY = Height - 1
            TargetBoard = CurrentBoard.ExitN
        ElseIf (YVector = 1) Then
            TargetY = 0
            TargetBoard = CurrentBoard.ExitS
        ElseIf (XVector = -1) Then
            TargetX = Width - 1
            TargetBoard = CurrentBoard.ExitW
        Else
            TargetX = 0
            TargetBoard = CurrentBoard.ExitE
        End If
        If (TargetBoard > 0) Then
            OldBoard = World.Board
            SetBoard(TargetBoard)
            If (ElementAt(TargetX, TargetY).ID <> ElementID.Player) Then
                Interact(TargetX, TargetY, Index, State.KeyVectorX, State.KeyVectorY)
            End If
            E = ElementAt(TargetX, TargetY)
            If (E.Floor) OrElse (E.ID = ElementID.Player) Then
                If (E.ID <> ElementID.Player) Then
                    MoveThing(0, TargetX, TargetY)
                End If
                BoardFade()
                XVector = 0
                YVector = 0
                EnterBoard()
                State.EnteringBoard = True
                If (Parameters.EnterAlert) Then
                    'ALL:ENTER
                    SendLabel(0, OOPCodeSet(KnownCommand.All) & ":" & OOPCodeSet(KnownCommand.Enter), False)
                End If
            Else
                SetBoard(OldBoard)
            End If
        End If
    End Sub

    Protected Sub InteractBomb(ByVal X As Integer, ByVal Y As Integer, ByVal Index As Integer, ByRef XVector As Integer, ByRef YVector As Integer)
        Dim T As Thing

        T = ThingAt(X, Y)
        With T
            If (.P1 > 0) Then
                Push(X, Y, New Vector(XVector, YVector))
                If (Not Parameters.KoopoBug) Then
                    Index = CurrentBoard.Things.IndexOf(T)
                End If
                T = CurrentBoard.Things(Index)
            Else
                .P1 = 9
                UpdateAt(X, Y)
                SetMessage(&HC8, Parameters.BombActiveMessage)
                PlaySound(4, SoundBank.BombActivate)
            End If
        End With
    End Sub

    Protected Sub InteractDoor(ByVal X As Integer, ByVal Y As Integer, ByVal Index As Integer, ByRef XVector As Integer, ByRef YVector As Integer)
        Dim I As Integer
        Dim Success As Boolean

        I = (ColorAt(X, Y) >> 4) Mod 8
        If (I > 0) Then
            Success = World.Keys(I - 1)
            If (Success) Then
                World.Keys(I - 1) = False
            End If
        Else
            Success = ((World.Gems And &HFF00) <> 0)
            If (Success) Then
                World.Gems = (World.Gems And 255)
            End If
        End If
        If (Success) Then
            RemoveItem(X, Y)
            StatusUpdate()
            SetMessage(&HC8, Parameters.DoorOpenMessage.Replace("%1", ColorStrings(I)))
            PlaySound(3, SoundBank.DoorOpen)
        Else
            SetMessage(&HC8, Parameters.DoorLockedMessage.Replace("%1", ColorStrings(I)))
            PlaySound(3, SoundBank.DoorLocked)
        End If
        UpdateAt(X, Y)
    End Sub

    Protected Sub InteractEnemy(ByVal X As Integer, ByVal Y As Integer, ByVal Index As Integer, ByRef XVector As Integer, ByRef YVector As Integer)
        Attack(Index, X, Y)
    End Sub

    Protected Sub InteractEnergizer(ByVal X As Integer, ByVal Y As Integer, ByVal Index As Integer, ByRef XVector As Integer, ByRef YVector As Integer)
        PlaySound(9, SoundBank.Energizer)
        RemoveItem(X, Y)
        World.EnergyCycles = &H4B
        StatusUpdate()
        If (Alerts.Energy) Then
            Alerts.Energy = False
            SetMessage(&HC8, Parameters.EnergizerMessage)
        End If
        'ALL:ENERGIZE
        SendLabel(0, OOPCodeSet(KnownCommand.All) & ":" & OOPCodeSet(KnownCommand.Energize), False)
        UpdateAt(X, Y)
    End Sub

    Protected Sub InteractFake(ByVal X As Integer, ByVal Y As Integer, ByVal Index As Integer, ByRef XVector As Integer, ByRef YVector As Integer)
        If (Alerts.Fake) Then
            Alerts.Fake = False
            SetMessage(&H96, Parameters.FakeWallMessage)
        End If
    End Sub

    Protected Sub InteractForest(ByVal X As Integer, ByVal Y As Integer, ByVal Index As Integer, ByRef XVector As Integer, ByRef YVector As Integer)
        Dim ForestSound(1) As Byte

        If (Parameters.ForestToFloor) Then
            TileAt(X, Y) = New Tile(Library.GetElementFromID(ElementID.Floor), &H2)
        Else
            ElementAt(X, Y) = Library.GetElementFromID(ElementID.Empty)
        End If
        ForestSound(0) = SoundBank.Forest(State.ForestMusicIndex)
        ForestSound(1) = 1
        State.ForestMusicIndex = (State.ForestMusicIndex + 1) Mod SoundBank.Forest.Length
        PlaySound(3, ForestSound)
        If (Alerts.Forest) Then
            Alerts.Forest = False
            SetMessage(&HC8, Parameters.ForestMessage)
        End If
        UpdateAt(X, Y)
    End Sub

    Protected Sub InteractGem(ByVal X As Integer, ByVal Y As Integer, ByVal Index As Integer, ByRef XVector As Integer, ByRef YVector As Integer)
        World.Gems += 1
        World.Health += Parameters.HealthPerGem
        World.Score += 10
        RemoveItem(X, Y)
        StatusUpdate()
        PlaySound(2, SoundBank.Gem)
        If (Alerts.Gem) Then
            Alerts.Gem = False
            SetMessage(&HC8, Parameters.GemMessage)
        End If
        UpdateAt(X, Y)
    End Sub

    Protected Sub InteractInvisible(ByVal X As Integer, ByVal Y As Integer, ByVal Index As Integer, ByRef XVector As Integer, ByRef YVector As Integer)
        ElementAt(X, Y) = Library.GetElementFromID(ElementID.Normal)
        UpdateAt(X, Y)
        PlaySound(3, SoundBank.Invisible)
        SetMessage(&H64, Parameters.InvisibleMessage)
    End Sub

    Protected Sub InteractKey(ByVal X As Integer, ByVal Y As Integer, ByVal Index As Integer, ByRef XVector As Integer, ByRef YVector As Integer)
        Dim I As Integer
        Dim Success As Boolean

        I = ColorAt(X, Y) Mod 8
        If (I > 0) Then
            Success = Not World.Keys(I - 1)
            If (Success) Then
                World.Keys(I - 1) = True
            End If
        Else
            Success = ((World.Gems And &HFF00) = 0)
            If (Success) Then
                World.Gems = (World.Gems And 255) Or 256
            End If
        End If
        If (Success) Then
            RemoveItem(X, Y)
            StatusUpdate()
            SetMessage(&HC8, Parameters.KeyMessage.Replace("%1", ColorStrings(I)))
            PlaySound(2, SoundBank.Key)
        Else
            SetMessage(&HC8, Parameters.KeyAlreadyMessage.Replace("%1", ColorStrings(I)))
            PlaySound(2, SoundBank.KeyAlready)
        End If
        UpdateAt(X, Y)
    End Sub

    Protected Sub InteractLava(ByVal X As Integer, ByVal Y As Integer, ByVal Index As Integer, ByRef XVector As Integer, ByRef YVector As Integer)
        PlaySound(3, SoundBank.Water)
        SetMessage(&H64, Parameters.LavaMessage)
    End Sub

    Protected Sub InteractObject(ByVal X As Integer, ByVal Y As Integer, ByVal Index As Integer, ByRef XVector As Integer, ByRef YVector As Integer)
        'TOUCH
        SendLabel(-ThingIndexAt(X, Y), OOPCodeSet(KnownCommand.Touch), False)
    End Sub

    Protected Sub InteractPassage(ByVal X As Integer, ByVal Y As Integer, ByVal Index As Integer, ByRef XVector As Integer, ByRef YVector As Integer)
        Dim Col As Integer
        Dim TargetX As Integer = -1
        Dim TargetY As Integer = -1
        Dim IndexX As Integer
        Dim IndexY As Integer

        Col = ColorAt(X, Y)
        SetBoard(ThingAt(X, Y).P3)

        With CurrentBoard.Things(0)
            For IndexX = 0 To Width - 1
                For IndexY = 0 To Height - 1
                    With TileAt(IndexX, IndexY)
                        If (.Color = Col) AndAlso (.Element.ID = ElementID.Passage) Then
                            If (Parameters.BuggyPassages) OrElse ((IndexX <> X) OrElse (IndexY <> Y)) Then
                                TargetX = IndexX
                                TargetY = IndexY
                            End If
                        End If
                    End With
                Next
            Next
            If (Parameters.BuggyPassages) Then
                TileAt(PlayerX, PlayerY) = New Tile(Library.GetElementFromID(ElementID.Empty), 0)
            Else
                TileAt(PlayerX, PlayerY) = New Tile(CurrentBoard.Things(0).UnderElement, CurrentBoard.Things(0).UnderColor)
            End If
            If (TargetX >= 0) Then
                .X = TargetX
                .Y = TargetY
            End If
        End With

        State.GamePaused = True
        StatusUpdate()
        PlaySound(4, SoundBank.Passage)
        BoardFade()
        EnterBoard()

        State.EnteringBoard = True
        XVector = 0
        YVector = 0
        If (Parameters.EnterAlert) Then
            'ALL:ENTER
            SendLabel(0, OOPCodeSet(KnownCommand.All) & ":" & OOPCodeSet(KnownCommand.Enter), False)
        End If
    End Sub

    Protected Sub InteractPuzzle(ByVal X As Integer, ByVal Y As Integer, ByVal Index As Integer, ByRef XVector As Integer, ByRef YVector As Integer)
        Push(X, Y, New Vector(XVector, YVector))
        PlaySound(2, SoundBank.Push)
    End Sub

    Protected Sub InteractScroll(ByVal X As Integer, ByVal Y As Integer, ByVal Index As Integer, ByRef XVector As Integer, ByRef YVector As Integer)
        Dim T As Thing
        Dim I As Integer
        PlaySound(2, ProcessMusic("c-c+d-d+e-e+f-f+g-g"))
        I = ThingIndexAt(X, Y)
        T = CurrentBoard.Things(I)
        T.CurrentInstruction = 0
        ExecuteOOP(I, T.CurrentInstruction, "Scroll")
        MainOOPState = Nothing
        If (Scroll.Visible) Then
            State.KeyVectorX = 0
            State.KeyVectorY = 0
        End If
        RemoveThing(ThingIndexAt(X, Y))
        UpdateAt(X, Y)
    End Sub

    Protected Sub InteractSlime(ByVal X As Integer, ByVal Y As Integer, ByVal Index As Integer, ByRef XVector As Integer, ByRef YVector As Integer)
        Dim I As Integer
        Dim T As Thing = CurrentBoard.Things(Index)

        I = ColorAt(X, Y)
        HarmThing(ThingIndexAt(X, Y))
        TileAt(X, Y) = New Tile(Library.GetElementFromID(ElementID.Breakable), I)
        UpdateAt(X, Y)
        PlaySound(2, SoundBank.SlimeDie)
    End Sub

    Protected Sub InteractStone(ByVal X As Integer, ByVal Y As Integer, ByVal Index As Integer, ByRef XVector As Integer, ByRef YVector As Integer)
        If (World.Stones < 0) Then
            World.Stones = 0
        End If
        World.Stones += 1
        DestroyAt(X, Y)
        StatusUpdate()
        SetMessage(&HC8, Parameters.StoneMessage)
    End Sub

    Protected Sub InteractTorch(ByVal X As Integer, ByVal Y As Integer, ByVal Index As Integer, ByRef XVector As Integer, ByRef YVector As Integer)
        World.Torches += 1
        RemoveItem(X, Y)
        StatusUpdate()
        If (Alerts.Torch) Then
            SetMessage(&HC8, Parameters.TorchMessage)
            Alerts.Torch = False
        End If
        PlaySound(3, SoundBank.Torch)
        UpdateAt(X, Y)
    End Sub

    Protected Sub InteractTransporter(ByVal X As Integer, ByVal Y As Integer, ByVal Index As Integer, ByRef XVector As Integer, ByRef YVector As Integer)
        PushTransporter(X - XVector, Y - YVector, New Vector(XVector, YVector))
        XVector = 0
        YVector = 0
    End Sub

    Protected Sub InteractWater(ByVal X As Integer, ByVal Y As Integer, ByVal Index As Integer, ByRef XVector As Integer, ByRef YVector As Integer)
        PlaySound(3, SoundBank.Water)
        SetMessage(&H64, Parameters.WaterMessage)
    End Sub

End Class
