﻿' RotonMessenger
'
' Contains the plugin interface for displaying messages.


Public Interface Messenger
    Sub SetMessage(ByVal Ticks As Integer, ByVal Message As String)
    Sub Update()
    Sub Clear()
    ReadOnly Property TicksLeft As Integer
End Interface

Public Class MessengerNull
    Implements Messenger

    Public Sub SetMessage(ByVal Ticks As Integer, ByVal Message As String) Implements Messenger.SetMessage
        Return
    End Sub

    Public ReadOnly Property TicksLeft As Integer Implements Messenger.TicksLeft
        Get
            Return 0
        End Get
    End Property

    Public Sub Update() Implements Messenger.Update
        Return
    End Sub

    Public Sub Clear() Implements Messenger.Clear
        Return
    End Sub
End Class

Public Class MessengerVideo
    Implements Messenger

    Private Video As Video
    Private YPos As Integer
    Private MessageTicks As Integer
    Private MessageStrings() As String
    Private Params As EnvironmentParameters

    Public Sub New(ByVal NewVideo As Video, ByVal NewParams As EnvironmentParameters)
        Video = NewVideo
        YPos = NewParams.ViewHeight - 1
        MessageStrings = {""}
        Params = NewParams
    End Sub

    Public Sub SetMessage(ByVal Ticks As Integer, ByVal Message As String) Implements Messenger.SetMessage
        MessageTicks = Ticks
        MessageStrings = Message.Split(Chr(13))
        Video.ClearOverlay()
    End Sub

    Public ReadOnly Property TicksLeft As Integer Implements Messenger.TicksLeft
        Get
            Return MessageTicks
        End Get
    End Property

    Public Sub Update() Implements Messenger.Update
        Dim Index As Integer
        Dim St As String
        Dim Y As Integer = YPos - (MessageStrings.Length - 1)

        If (MessageTicks > 0) Then
            For Index = 0 To MessageStrings.Length - 1
                St = " " & MessageStrings(Index) & " "
                If (St.Length > 2) Then
                    Video.WriteOverlay(((Params.ViewWidth - St.Length) \ 2) + Params.MessageCenterOffset, Y + Index, St, (MessageTicks Mod 7) + 9)
                End If
            Next
            MessageTicks -= 1
            If (MessageTicks = 0) Then
                Video.ClearOverlay()
            End If
        End If
    End Sub

    Public Sub Clear() Implements Messenger.Clear
        MessageTicks = 0
        Video.ClearOverlay()
    End Sub
End Class
