﻿' RotonOOP
'
' Handles all ZZT-OOP programming and functions.


Partial Public Class Environment

    Public OOPCodeSet As CommandSet = CommandSetDefaults.Standard
    Protected MainOOPState As OOPState

    Public Enum KnownCommand As Integer
        [Then]
        [Go]
        [Try]
        Walk
        [Set]
        [Clear]
        [If]
        Shoot
        Throwstar
        Give
        Take
        [End]
        Endgame
        Idle
        Restart
        Zap
        Restore
        Lock
        Unlock
        Send
        Become
        Put
        Change
        Play
        Cycle
        [Char]
        Die
        Bind
        All
        Others
        Self
        [Not]
        Alligned
        Contact
        Blocked
        Energized
        Any
        N
        North
        S
        South
        E
        East
        W
        West
        I
        IdleDirection
        Seek
        Flow
        [Rnd]
        RndNS
        RndNE
        CW
        CCW
        RNDP
        OPP
        Blue
        Green
        Cyan
        Red
        Purple
        Yellow
        White
        Health
        Ammo
        Gems
        Torches
        Score
        Time
        Z
        Touch
        Thud
        Bombed
        Energize
        Enter
        Hint
        Shot
    End Enum

    Public NotInheritable Class CommandSetDefaults
        Public Shared Function Standard() As CommandSet
            Return New CommandSet({
                "THEN", "GO", "TRY", "WALK", "SET",
                "CLEAR", "IF", "SHOOT", "THROWSTAR", "GIVE",
                "TAKE", "END", "ENDGAME", "IDLE", "RESTART",
                "ZAP", "RESTORE", "LOCK", "UNLOCK", "SEND",
                "BECOME", "PUT", "CHANGE", "PLAY", "CYCLE",
                "CHAR", "DIE", "BIND", "ALL", "OTHERS",
                "SELF", "NOT", "ALLIGNED", "CONTACT", "BLOCKED",
                "ENERGIZED", "ANY", "N", "NORTH", "S",
                "SOUTH", "E", "EAST", "W", "WEST",
                "I", "IDLE", "SEEK", "FLOW", "RND",
                "RNDNS", "RNDNE", "CW", "CCW", "RNDP",
                "OPP", "BLUE", "GREEN", "CYAN", "RED",
                "PURPLE", "YELLOW", "WHITE", "HEALTH", "AMMO",
                "GEMS", "TORCHES", "SCORE", "TIME", "Z",
                "TOUCH", "THUD", "BOMBED", "ENERGIZE", "ENTER",
                "HINT", "SHOT"
            })
        End Function

        Public Shared Function BananaQuest() As CommandSet
            Return New CommandSet({
                "THEN", "GO", "TRY", "WALK", "SET",
                "CLEAR", "IF", "SHOOT", "THROWSTAR", "GIVE",
                "TAKE", "STP", "OHHDREA", "IDLE", "RESTART",
                "ZAP", "RESTORE", "LOCK", "UNLOCK", "SEND",
                "BECOME", "PUT", "CHANGE", "MUZK", "SPEED",
                "CHAR", "DIE", "BIND", "ALL", "OTHERS",
                "SELF", "NOT", "ALLIGNED", "CONTACT", "BLOCKED",
                "ENERGIZED", "ANY", "N", "NORTH", "S",
                "SOUTH", "E", "EAST", "W", "WEST",
                "I", "IDLE", "SEEK", "FLOW", "RND",
                "RNDNS", "RNDNE", "CW", "CCW", "RNDP",
                "OPP", "BLUE", "GREEN", "CYAN", "RED",
                "PURPLE", "YELLOW", "WHITE", "HEALTH", "AMMO",
                "GEMS", "TORCHES", "SCORE", "TIME", "Z",
                "TCTVT", "THUD", "BOMBED", "ENERGIZE", "ENTER",
                "HINT", "SHOT"
            })
        End Function
    End Class

    Public Class CommandSet
        Public Commands() As String

        Public Sub New(NewCommands() As String)
            Commands = NewCommands
        End Sub

        Default Public ReadOnly Property GetIndex(KC As KnownCommand) As String
            Get
                Return Commands(CInt(KC))
            End Get
        End Property
    End Class

    Protected Class Kind
        Public Element As Element
        Public Color As Integer
    End Class

    Protected Class OOPState
        Public OOPByte As String
        Public OOPNumber As Integer
        Public OOPWord As String
        Public Initted As Boolean
        Public ThingName As String
        Public SendLabel As String
        Public BoardMessage As String
        Public Message As New List(Of String)
        Public TargetX As Integer
        Public TargetY As Integer
        Public Moved As Boolean
        Public Died As Boolean
        Public Finished As Boolean
        Public TargetTileData As Tile
        Public PreviousInst As Integer
        Public Repeat As Boolean
        Public AutoNextLine As Boolean
        Public OOPSearchOffs As Integer
        Public SearchOffset As Integer
        Public GiveTakeOffset As Integer
        Public CommandsExecuted As Integer
        Public ThingData As Thing
        Public DeathX As Integer
        Public DeathY As Integer
        Public Vector As Vector
        Public ThingIndex As Integer
        Public ResumeCommand As Boolean
        Public Kind As Kind
    End Class

    Protected Function ReadOOPByte(ByVal Thing As Thing, ByRef Offset As Integer) As String
        Dim Ret As String
        Dim Code As CodeContainer

        Code = Thing.Code
        If (Code Is Nothing) OrElse (Offset >= Thing.Length) OrElse (Offset < 0) Then
            Ret = ""
        Else
            Ret = Code.Code.Substring(Offset, 1)
            Offset += 1
        End If
        Return Ret
    End Function

    Protected Function ReadOOPWord(ByVal Thing As Thing, ByRef Offset As Integer) As String
        Dim Ch As String
        Dim Ret As String = ""

        Do
            Ch = ReadOOPByte(Thing, Offset)
        Loop While (Ch = " ")
        Ch = Ch.ToUpper
        If (Ch < "0") OrElse (Ch > "9") Then
            Do
                Select Case Ch
                    Case "A" To "Z", "0" To "9", ":", "_"
                        Ret &= Ch
                    Case Else
                        Exit Do
                End Select
                Ch = ReadOOPByte(Thing, Offset).ToUpper
            Loop
        End If
        If (Offset > 0) Then
            Offset -= 1
        End If
        Return Ret
    End Function

    Protected Function ReadOOPNumber(ByVal Thing As Thing, ByRef Offset As Integer) As Integer
        Dim Ch As String
        Dim Ret As String = ""
        Dim RetInt As Int16

        Do
            Ch = ReadOOPByte(Thing, Offset)
        Loop While (Ch = " ")
        Do
            If (Ch >= "0") AndAlso (Ch <= "9") Then
                Ret &= Ch
                Ch = ReadOOPByte(Thing, Offset)
            Else
                Exit Do
            End If
        Loop
        If (Offset > 0) Then
            Offset -= 1
        End If
        If (Ret = "") Then
            Return -1
        Else
            Try
                RetInt = CType(Ret, Int16)
            Catch ex As Exception
                RetInt = -1
            End Try
            Return CInt(RetInt)
        End If
    End Function

    Protected Function ReadOOPDirection(ByVal Thing As Thing, ByRef Offset As Integer) As Vector
        Dim V As Vector
        If (Not EvalDirection(Thing, Offset, V)) Then
            OOPError(CurrentBoard.Things.IndexOf(Thing), "Bad direction")
        End If
        Return V
    End Function

    Protected Function ReadOOPLine(ByVal Thing As Thing, ByRef Offset As Integer) As String
        Dim Ret As String = ""
        Dim Ch As String

        If (Offset < 0) Then
            Return ""
        End If

        Do
            Ch = ReadOOPByte(Thing, Offset)
            If (Ch = "") OrElse (Ch = Chr(0)) OrElse (Ch = Chr(13)) Then
                Exit Do
            End If
            Ret &= Ch
        Loop
        Return Ret
    End Function

    Protected Function EvalCondition(ByVal Thing As Thing, ByRef CurrentInst As Integer) As Boolean
        Dim V As Vector = Nothing
        Dim K As Kind = Nothing
        Dim W As String
        W = ReadOOPWord(Thing, CurrentInst)

        Select Case W
            Case OOPCodeSet(KnownCommand.Not)
                'NOT
                Return (Not EvalCondition(Thing, CurrentInst))
            Case OOPCodeSet(KnownCommand.Alligned)
                'ALLIGNED
                With Thing
                    Return ((.X = PlayerX) OrElse (.Y = PlayerY))
                End With
            Case OOPCodeSet(KnownCommand.Contact)
                'CONTACT
                With Thing
                    Return (((.Y - PlayerY) ^ 2) + ((.X - PlayerX) ^ 2) = 1)
                End With
            Case OOPCodeSet(KnownCommand.Blocked)
                'BLOCKED
                With Thing
                    V = ReadOOPDirection(Thing, CurrentInst)
                    Return (Not ElementAt(.X + V.X, .Y + V.Y).Floor)
                End With
            Case OOPCodeSet(KnownCommand.Energized)
                'ENERGIZED
                Return (World.EnergyCycles > 0)
            Case OOPCodeSet(KnownCommand.Any)
                'ANY
                If (EvalKind(Thing, CurrentInst, K)) Then
                    Return IsAny(K, -1, 0)
                Else
                    OOPError(CurrentBoard.Things.IndexOf(Thing), "Bad object kind")
                    Return False
                End If
            Case Else
                Return FindFlag(W)
        End Select
    End Function

    Protected Function EvalDirection(ByVal Thing As Thing, ByRef CurrentInst As Integer, ByRef ReturnVector As Vector) As Boolean
        Dim Success As Boolean
        Select Case ReadOOPWord(Thing, CurrentInst)
            Case OOPCodeSet(KnownCommand.N), OOPCodeSet(KnownCommand.North)
                'N, NORTH
                ReturnVector = Vector.North
                Success = True
            Case OOPCodeSet(KnownCommand.S), OOPCodeSet(KnownCommand.South)
                'S, SOUTH
                ReturnVector = Vector.South
                Success = True
            Case OOPCodeSet(KnownCommand.E), OOPCodeSet(KnownCommand.East)
                'E, EAST
                ReturnVector = Vector.East
                Success = True
            Case OOPCodeSet(KnownCommand.W), OOPCodeSet(KnownCommand.West)
                'W, WEST
                ReturnVector = Vector.West
                Success = True
            Case OOPCodeSet(KnownCommand.I), OOPCodeSet(KnownCommand.IdleDirection)
                'I, IDLE
                ReturnVector = Vector.Idle
                Success = True
            Case OOPCodeSet(KnownCommand.Seek)
                'SEEK
                With Thing
                    ReturnVector = Seek(.X, .Y)
                End With
                Success = True
            Case OOPCodeSet(KnownCommand.Flow)
                'FLOW
                With Thing
                    ReturnVector = .Flow
                End With
                Success = True
            Case OOPCodeSet(KnownCommand.Rnd)
                'RND
                ReturnVector = Rnd()
                Success = True
            Case OOPCodeSet(KnownCommand.RndNS)
                'RNDNS
                ReturnVector = RndNS()
                Success = True
            Case OOPCodeSet(KnownCommand.RndNE)
                'RNDNE
                ReturnVector = RndNE()
                Success = True
            Case OOPCodeSet(KnownCommand.CW)
                'CW
                Success = EvalDirection(Thing, CurrentInst, ReturnVector)
                ReturnVector = New Vector(-ReturnVector.Y, ReturnVector.X)
            Case OOPCodeSet(KnownCommand.CCW)
                'CCW
                Success = EvalDirection(Thing, CurrentInst, ReturnVector)
                ReturnVector = New Vector(ReturnVector.Y, -ReturnVector.X)
            Case OOPCodeSet(KnownCommand.RNDP)
                'RNDP
                Success = EvalDirection(Thing, CurrentInst, ReturnVector)
                ReturnVector = RndP(ReturnVector)
            Case OOPCodeSet(KnownCommand.OPP)
                'OPP
                Success = EvalDirection(Thing, CurrentInst, ReturnVector)
                ReturnVector = ReturnVector.Opp
        End Select
        Return Success
    End Function

    Protected Function EvalKind(ByVal Thing As Thing, ByRef CurrentInst As Integer, ByRef ReturnKind As Kind) As Boolean
        Dim ElementIndex As Integer
        Dim Word As String

        ReturnKind = New Kind
        ReturnKind.Color = 0
        ReturnKind.Element = Nothing
        Do
            Word = ReadOOPWord(Thing, CurrentInst)
            Select Case Word
                Case OOPCodeSet(KnownCommand.Blue)
                    'BLUE
                    ReturnKind.Color = 9
                    Continue Do
                Case OOPCodeSet(KnownCommand.Green)
                    'GREEN
                    ReturnKind.Color = 10
                    Continue Do
                Case OOPCodeSet(KnownCommand.Cyan)
                    'CYAN
                    ReturnKind.Color = 11
                    Continue Do
                Case OOPCodeSet(KnownCommand.Red)
                    'RED
                    ReturnKind.Color = 12
                    Continue Do
                Case OOPCodeSet(KnownCommand.Purple)
                    'PURPLE
                    ReturnKind.Color = 13
                    Continue Do
                Case OOPCodeSet(KnownCommand.Yellow)
                    'YELLOW
                    ReturnKind.Color = 14
                    Continue Do
                Case OOPCodeSet(KnownCommand.White)
                    'WHITE
                    ReturnKind.Color = 15
                    Continue Do
                Case Else
                    For ElementIndex = 0 To Library.Count - 1
                        If (Word = Library.GetElementFromIndex(CByte(ElementIndex)).OOPString) Then
                            ReturnKind.Element = Library.GetElementFromIndex(CByte(ElementIndex))
                            Exit For
                        End If
                    Next
            End Select
            Exit Do
        Loop
        Return (Not (ReturnKind.Element Is Nothing))
    End Function

    Protected Function IsAny(ByVal K As Kind, ByRef X As Integer, ByRef Y As Integer) As Boolean
        Dim H As Integer = Height
        Dim W As Integer = Width
        Dim T As Tile
        X += 1

        Do Until (Y = H)
            Do Until (X = W)
                T = TileAt(X, Y)
                If (T.Element.ID = K.Element.ID) Then
                    If (K.Color = 0) OrElse (K.Color = ColorMatch(T)) Then
                        Return True
                    End If
                End If
                X += 1
            Loop
            X = 0
            Y += 1
        Loop
        Return False
    End Function

    Protected Function ColorMatch(ByVal Tile As Tile) As Integer
        If (Tile.Element.Color < &HF0) Then
            Return (Tile.Element.Color And 7)
        ElseIf (Tile.Element.Color = &HFE) Then
            Return (((Tile.Color >> 4) And &HF) + 8)
        End If
        Return (Tile.Color And &HF)
    End Function

    Protected Sub OOPError(Index As Integer, ErrorMessage As String)
        SetMessage(&HC8, "ERR: " & ErrorMessage)
        PlaySound(5, SoundBank.OOPError)
        CurrentBoard.Things(Index).CurrentInstruction = -1
    End Sub

    Protected Sub ExecuteScroll(Owner As Thing, Title As String, Text As String)
        If Not (Scroll.Visible) Then
            ScrollOwner = Owner
            Scroll.Text = Text
            Scroll.Title = Title
            Scroll.Open(State, Parameters)
        End If
    End Sub

    Protected Sub ExecuteScrollClose(Index As Integer)
        Dim Message() As String = Scroll.Text.Split(Chr(13))
        Dim Label As String

        If (Message(Index) <> "") AndAlso (Message(Index).Substring(0, 1) = "!") AndAlso (Message(Index).IndexOf(";") > 0) Then
            Label = Message(Index).Substring(1)
            Label = Label.Substring(0, Label.IndexOf(";")).ToUpper
            If (Label.Length > 0) AndAlso (Label.StartsWith("-")) Then
                Label = Label.Substring(1)
                If (Label = "CON") Then
                    Throw New NotSupportedException("Console activated")
                End If
                Scroll.CancelClose()
                ExecuteHelp(Label)
            Else
                State.ScrollOwnerLink = Label
            End If
        Else
            State.ScrollOwnerLink = ""
        End If

        If (MainOOPState Is Nothing) Then
            State.ScrollOwnerLink = ""
        End If
    End Sub

    Public Function FindFlag(ByVal Name As String) As Boolean
        Dim Result As Boolean
        If (World.Flags.Contains(Name)) Then
            Result = True
        ElseIf (Name = "") AndAlso (World.Flags.Count < Parameters.MaxFlags) Then
            Result = True
        End If
        Return Result
    End Function

    Public Sub SetFlag(ByVal Name As String)
        If (Not FindFlag(Name)) Then
            If (World.Flags.Count < Parameters.MaxFlags) Then
                World.Flags.Add(Name.ToUpper)
            Else
                World.Flags(Parameters.MaxFlags - 1) = Name
            End If
        End If
    End Sub

    Public Sub ClearFlag(ByVal Name As String)
        Dim Index As Integer
        Index = World.Flags.IndexOf(Name)
        If (Index >= 0) Then
            World.Flags.RemoveAt(Index)
        End If
    End Sub

    Protected Sub ExecuteOOP(Index As Integer, ByRef Instruction As Integer, Name As String)
        Dim OOP As OOPState
        Dim Count As Integer
        Dim Skip As Boolean

        If (MainOOPState Is Nothing) Then
            MainOOPState = New OOPState
            MainOOPState.ThingIndex = Index
        Else
            If (State.ScrollOwnerLink <> "") Then
                Skip = (Not SendLabel(Index, State.ScrollOwnerLink, True))
            Else
                Skip = True
            End If
            State.ScrollOwnerLink = ""
        End If

        OOP = MainOOPState

        With OOP
            .Message = New List(Of String)
            If (Not Skip) Then
                .Moved = False
                .Repeat = False
                .Died = False
                .Finished = False
                .CommandsExecuted = 0
                .ThingData = CurrentBoard.Things(Index)

                Do
                    If (Instruction < 0) Then
                        Exit Do
                    End If

                    .AutoNextLine = True
                    .PreviousInst = Instruction
                    .OOPByte = ReadOOPByte(.ThingData, Instruction)
                    Select Case .OOPByte
                        Case ":"
                            ReadOOPLine(.ThingData, Instruction)
                        Case "'"
                            ReadOOPLine(.ThingData, Instruction)
                        Case "@"
                            ReadOOPLine(.ThingData, Instruction)
                        Case "/", "?"
                            If (.OOPByte = "/") Then
                                .Repeat = True
                            End If
                            If (EvalDirection(.ThingData, Instruction, .Vector)) Then
                                If (Parameters.MultiMovement) Then
                                    Count = ReadOOPNumber(.ThingData, Instruction)
                                    If (Count < 0) Then
                                        Count = 1
                                    End If
                                    If (.OOPByte = "?") Then
                                        Count = (-Count) And 255
                                    End If
                                    .ThingData.P2 = Count
                                    .ThingData.Vector = .Vector
                                    .Repeat = False
                                Else
                                    If (.Vector.X = 0) AndAlso (.Vector.Y = 0) Then
                                        .Repeat = False
                                    Else
                                        .TargetX = .Vector.X + .ThingData.X
                                        .TargetY = .Vector.Y + .ThingData.Y
                                        If (Not ElementAt(.TargetX, .TargetY).Floor) Then
                                            Push(.TargetX, .TargetY, .Vector)
                                        End If
                                        If (ElementAt(.TargetX, .TargetY).Floor) Then
                                            MoveThing(.ThingIndex, .TargetX, .TargetY)
                                            .Repeat = False
                                        End If
                                    End If
                                End If
                                .OOPByte = ReadOOPByte(.ThingData, Instruction)
                                If (.OOPByte <> Chr(13)) Then
                                    Instruction -= 1
                                End If
                                .Moved = True
                            Else
                                OOPError(.ThingIndex, "Bad direction")
                            End If
                        Case "#"
                            Do
                                .ResumeCommand = False
                                .OOPWord = ReadOOPWord(.ThingData, Instruction)
                                If (.OOPWord = OOPCodeSet(KnownCommand.Then)) Then
                                    .OOPWord = ReadOOPWord(.ThingData, Instruction)
                                End If
                                If (.OOPWord.Length = 0) Then
                                    Exit Do
                                End If
                                .CommandsExecuted += 1
                                Select Case .OOPWord
                                    Case OOPCodeSet(KnownCommand.Become)
                                        ExecuteOOPBecome(OOP, Instruction)
                                    Case OOPCodeSet(KnownCommand.Bind)
                                        ExecuteOOPBind(OOP, Instruction)
                                    Case OOPCodeSet(KnownCommand.Change)
                                        ExecuteOOPChange(OOP, Instruction)
                                    Case OOPCodeSet(KnownCommand.Char)
                                        ExecuteOOPChar(OOP, Instruction)
                                    Case OOPCodeSet(KnownCommand.Clear)
                                        ExecuteOOPClear(OOP, Instruction)
                                    Case OOPCodeSet(KnownCommand.Cycle)
                                        ExecuteOOPCycle(OOP, Instruction)
                                    Case OOPCodeSet(KnownCommand.Die)
                                        ExecuteOOPDie(OOP, Instruction)
                                    Case OOPCodeSet(KnownCommand.End)
                                        ExecuteOOPEnd(OOP, Instruction)
                                    Case OOPCodeSet(KnownCommand.Endgame)
                                        ExecuteOOPEndGame(OOP, Instruction)
                                    Case OOPCodeSet(KnownCommand.Give)
                                        ExecuteOOPGive(OOP, Instruction)
                                    Case OOPCodeSet(KnownCommand.Go)
                                        ExecuteOOPGo(OOP, Instruction)
                                    Case OOPCodeSet(KnownCommand.Idle)
                                        ExecuteOOPIdle(OOP, Instruction)
                                    Case OOPCodeSet(KnownCommand.If)
                                        ExecuteOOPIf(OOP, Instruction)
                                    Case OOPCodeSet(KnownCommand.Lock)
                                        ExecuteOOPLock(OOP, Instruction)
                                    Case OOPCodeSet(KnownCommand.Play)
                                        ExecuteOOPPlay(OOP, Instruction)
                                    Case OOPCodeSet(KnownCommand.Put)
                                        ExecuteOOPPut(OOP, Instruction)
                                    Case OOPCodeSet(KnownCommand.Restart)
                                        ExecuteOOPRestart(OOP, Instruction)
                                    Case OOPCodeSet(KnownCommand.Restore)
                                        ExecuteOOPRestore(OOP, Instruction)
                                    Case OOPCodeSet(KnownCommand.Send)
                                        ExecuteOOPSend(OOP, Instruction)
                                    Case OOPCodeSet(KnownCommand.Set)
                                        ExecuteOOPSet(OOP, Instruction)
                                    Case OOPCodeSet(KnownCommand.Shoot)
                                        ExecuteOOPShoot(OOP, Instruction)
                                    Case OOPCodeSet(KnownCommand.Take)
                                        ExecuteOOPTake(OOP, Instruction)
                                    Case OOPCodeSet(KnownCommand.Throwstar)
                                        ExecuteOOPThrowstar(OOP, Instruction)
                                    Case OOPCodeSet(KnownCommand.Try)
                                        ExecuteOOPTry(OOP, Instruction)
                                    Case OOPCodeSet(KnownCommand.Unlock)
                                        ExecuteOOPUnlock(OOP, Instruction)
                                    Case OOPCodeSet(KnownCommand.Walk)
                                        ExecuteOOPWalk(OOP, Instruction)
                                    Case OOPCodeSet(KnownCommand.Zap)
                                        ExecuteOOPZap(OOP, Instruction)
                                    Case Else
                                        ExecuteOOPUserCommand(OOP, Instruction)
                                End Select

                                If (.ResumeCommand) Then
                                    .ResumeCommand = False
                                    Continue Do
                                End If

                                If (.AutoNextLine) AndAlso (Instruction >= 0) Then
                                    ReadOOPLine(.ThingData, Instruction)
                                End If
                                Exit Do
                            Loop

                            If (.OOPByte = "") Then
                                Continue Do
                            End If
                        Case Else
                            If (.OOPByte = Chr(13)) Then
                                If (.Message.Count > 0) Then
                                    .Message.Add(.OOPByte)
                                End If
                            ElseIf (.OOPByte = "") Then
                                .Finished = True
                            Else
                                .Message.Add(.OOPByte & ReadOOPLine(.ThingData, Instruction))
                            End If
                    End Select
                Loop Until (.Finished) OrElse (.Moved) OrElse (.Repeat) OrElse (.Died) OrElse (.CommandsExecuted > Parameters.MaxCommands)

                If (.Repeat) Then
                    Instruction = .PreviousInst
                End If

                If (.OOPByte = "") Then
                    Instruction = -1
                End If

                If (.Message.Count > Parameters.MessageLines) Then
                    If (ReadOOPByte(.ThingData, 0) = "@") Then
                        Name = ReadOOPLine(.ThingData, 1)
                    End If
                    If (Name = "") Then
                        Name = "Interaction"
                    End If
                    ExecuteScroll(.ThingData, Name, ExecuteOOPBuildMessage(.Message))
                    Exit Sub
                ElseIf (.Message.Count > 0) Then
                    SetMessage(&HC8, ExecuteOOPBuildMessage(.Message))
                End If
            End If

            If (.Died) Then
                .DeathX = .ThingData.X
                .DeathY = .ThingData.Y
                If (Parameters.PreKillObject) Then
                    HarmThing(.ThingIndex)
                End If
                PlotTile(.DeathX, .DeathY, .TargetTileData)
            End If

            State.ScrollOwnerLink = ""
            MainOOPState = Nothing
        End With
    End Sub

    Private Function ExecuteOOPBuildMessage(Source As List(Of String)) As String
        Dim Ret As String = ""
        Dim Index As Integer

        For Index = 0 To Source.Count - 1
            If (Index > 0) Then
                Ret &= Chr(13)
            End If
            Ret &= Source(Index)
        Next

        Return Ret
    End Function

    Private Sub ExecuteOOPGo(OOP As OOPState, ByRef Instruction As Integer)
        With OOP
            .Vector = ReadOOPDirection(.ThingData, Instruction)
            If (Not ElementAt(.Vector.X + .ThingData.X, .Vector.Y + .ThingData.Y).Floor) Then
                Push(.ThingData.X + .Vector.X, .ThingData.Y + .Vector.Y, .Vector)
            End If
            If (ElementAt(.Vector.X + .ThingData.X, .Vector.Y + .ThingData.Y).Floor) Then
                MoveThing(.ThingIndex, .Vector.X + .ThingData.X, .Vector.Y + .ThingData.Y)
                .Moved = True
            Else
                .Repeat = True
            End If
        End With
    End Sub

    Private Sub ExecuteOOPTry(OOP As OOPState, ByRef Instruction As Integer)
        With OOP
            .Vector = ReadOOPDirection(.ThingData, Instruction)
            If (Not ElementAt(.Vector.X + .ThingData.X, .Vector.Y + .ThingData.Y).Floor) Then
                Push(.ThingData.X + .Vector.X, .ThingData.Y + .Vector.Y, .Vector)
            End If
            If (ElementAt(.Vector.X + .ThingData.X, .Vector.Y + .ThingData.Y).Floor) Then
                MoveThing(.ThingIndex, .Vector.X + .ThingData.X, .Vector.Y + .ThingData.Y)
                .Moved = True
            Else
                .ResumeCommand = True
            End If
        End With
    End Sub

    Private Sub ExecuteOOPWalk(OOP As OOPState, ByRef Instruction As Integer)
        With OOP
            .Vector = ReadOOPDirection(.ThingData, Instruction)
            .ThingData.XStep = .Vector.X
            .ThingData.YStep = .Vector.Y
        End With
    End Sub

    Private Sub ExecuteOOPSet(OOP As OOPState, ByRef Instruction As Integer)
        With OOP
            .OOPWord = ReadOOPWord(.ThingData, Instruction)
            SetFlag(.OOPWord)
        End With
    End Sub

    Private Sub ExecuteOOPClear(OOP As OOPState, ByRef Instruction As Integer)
        With OOP
            .OOPWord = ReadOOPWord(.ThingData, Instruction)
            ClearFlag(.OOPWord)
        End With
    End Sub

    Private Sub ExecuteOOPIf(OOP As OOPState, ByRef Instruction As Integer)
        With OOP
            If (EvalCondition(.ThingData, Instruction)) Then
                .ResumeCommand = True
            End If
        End With
    End Sub

    Private Sub ExecuteOOPShoot(OOP As OOPState, ByRef Instruction As Integer)
        With OOP
            .Vector = ReadOOPDirection(.ThingData, Instruction)
            If (SpawnProjectile(Library.GetElementFromID(ElementID.Bullet), .ThingData.X, .ThingData.Y, .Vector, True)) Then
                PlaySound(2, SoundBank.EnemyShoot)
            End If
            .Moved = True
        End With
    End Sub

    Private Sub ExecuteOOPThrowstar(OOP As OOPState, ByRef Instruction As Integer)
        With OOP
            .Vector = ReadOOPDirection(.ThingData, Instruction)
            SpawnProjectile(Library.GetElementFromID(ElementID.Star), .ThingData.X, .ThingData.Y, .Vector, True)
            .Moved = True
        End With
    End Sub

    Private Sub ExecuteOOPGive(OOP As OOPState, ByRef Instruction As Integer)
        ExecuteOOPTransaction(OOP, Instruction, False)
    End Sub

    Private Sub ExecuteOOPTake(OOP As OOPState, ByRef Instruction As Integer)
        ExecuteOOPTransaction(OOP, Instruction, True)
    End Sub

    Private Sub ExecuteOOPTransaction(OOP As OOPState, ByRef Instruction As Integer, ByVal IsTake As Boolean)
        With OOP
            .OOPWord = ReadOOPWord(.ThingData, Instruction)
            .OOPNumber = ReadOOPNumber(.ThingData, Instruction)
            If (.OOPNumber > 0) Then
                If (IsTake) Then
                    .OOPNumber = -.OOPNumber
                End If
                Select Case .OOPWord
                    Case OOPCodeSet(KnownCommand.Health)
                        ExecuteOOPTransactionApply(OOP, World.Health, .OOPNumber)
                    Case OOPCodeSet(KnownCommand.Ammo)
                        ExecuteOOPTransactionApply(OOP, World.Ammo, .OOPNumber)
                    Case OOPCodeSet(KnownCommand.Gems)
                        ExecuteOOPTransactionApply(OOP, World.Gems, .OOPNumber)
                    Case OOPCodeSet(KnownCommand.Torches)
                        If (Parameters.TorchesEnabled) Then
                            ExecuteOOPTransactionApply(OOP, World.Torches, .OOPNumber)
                        Else
                            .ResumeCommand = True
                        End If
                    Case OOPCodeSet(KnownCommand.Score)
                        ExecuteOOPTransactionApply(OOP, World.Score, .OOPNumber)
                    Case OOPCodeSet(KnownCommand.Time)
                        ExecuteOOPTransactionApply(OOP, World.TimePassed, .OOPNumber)
                    Case OOPCodeSet(KnownCommand.Z)
                        If (Parameters.StonesEnabled) Then
                            ExecuteOOPTransactionApply(OOP, World.Stones, .OOPNumber)
                        Else
                            .ResumeCommand = True
                        End If
                    Case Else
                        .ResumeCommand = True
                End Select
            End If
        End With
        StatusUpdate()
    End Sub

    Private Sub ExecuteOOPTransactionApply(ByVal OOP As OOPState, ByRef Value As Integer, Amount As Integer)
        Dim ValueTest As Integer = Value

        With OOP
            ValueTest += Amount

            ' a little hack for 16-bit numbers
            If (ValueTest > 32767) Then
                ValueTest -= 65536
            ElseIf (ValueTest < -32768) Then
                ValueTest += 65536
            End If

            If (ValueTest < 0) Then
                .ResumeCommand = True
            Else
                Value = ValueTest
            End If
        End With
    End Sub

    Private Sub ExecuteOOPEnd(OOP As OOPState, ByRef Instruction As Integer)
        With OOP
            .OOPByte = ""
            Instruction = -1
        End With
    End Sub

    Private Sub ExecuteOOPEndGame(OOP As OOPState, ByRef Instruction As Integer)
        With OOP
            World.Health = 0
        End With
    End Sub

    Private Sub ExecuteOOPIdle(OOP As OOPState, ByRef Instruction As Integer)
        With OOP
            .Moved = True
        End With
    End Sub

    Private Sub ExecuteOOPRestart(OOP As OOPState, ByRef Instruction As Integer)
        With OOP
            Instruction = 0
            .AutoNextLine = False
        End With
    End Sub

    Private Sub ExecuteOOPZap(OOP As OOPState, ByRef Instruction As Integer)
        Dim Pre As String
        Dim Post As String
        With OOP
            .OOPWord = ReadOOPWord(.ThingData, Instruction)
            .SearchOffset = 0
            Do While (OOPSendSearch(.ThingIndex, .OOPWord, .SearchOffset, .OOPSearchOffs, Chr(13) & ":"))
                With CurrentBoard.Things(.SearchOffset).Code
                    Pre = .Code.Substring(0, OOP.OOPSearchOffs + 1)
                    Post = .Code.Substring(OOP.OOPSearchOffs + 2)
                    .Code = Pre & "'" & Post
                End With
            Loop
        End With
    End Sub

    Private Sub ExecuteOOPRestore(OOP As OOPState, ByRef Instruction As Integer)
        Dim Pre As String
        Dim Post As String
        With OOP
            .OOPWord = ReadOOPWord(.ThingData, Instruction)
            .SearchOffset = 0
            If (OOPSendSearch(.ThingIndex, .OOPWord, .SearchOffset, .OOPSearchOffs, Chr(13) & "'")) Then
                Do
                    With CurrentBoard.Things(.SearchOffset).Code
                        Pre = .Code.Substring(0, OOP.OOPSearchOffs + 1)
                        Post = .Code.Substring(OOP.OOPSearchOffs + 2)
                        .Code = Pre & ":" & Post
                    End With
                    .OOPSearchOffs = OOPSearch(.SearchOffset, Chr(13) & "'" & .OOPWord & Chr(13))
                Loop While (.OOPSearchOffs > 0)
            End If
        End With
    End Sub

    Private Sub ExecuteOOPLock(OOP As OOPState, ByRef Instruction As Integer)
        With OOP
            If (Parameters.MultiMovement) Then
                .ThingData.P3 = 1
            Else
                .ThingData.P2 = 1
            End If
        End With
    End Sub

    Private Sub ExecuteOOPUnlock(OOP As OOPState, ByRef Instruction As Integer)
        With OOP
            If (Parameters.MultiMovement) Then
                .ThingData.P3 = 0
            Else
                .ThingData.P2 = 0
            End If
        End With
    End Sub

    Private Sub ExecuteOOPSend(OOP As OOPState, ByRef Instruction As Integer)
        With OOP
            .OOPWord = ReadOOPWord(.ThingData, Instruction)
            .AutoNextLine = (Not SendLabel(.ThingIndex, .OOPWord, False))
        End With
    End Sub

    Private Sub ExecuteOOPBecome(OOP As OOPState, ByRef Instruction As Integer)
        With OOP
            If (EvalKind(.ThingData, Instruction, .Kind)) Then
                .Died = True
                .TargetTileData = New Tile(.Kind.Element, .Kind.Color)
            Else
                OOPError(.ThingIndex, "Bad #BECOME")
            End If
        End With
    End Sub

    Private Sub ExecuteOOPPut(OOP As OOPState, ByRef Instruction As Integer)
        Dim ValidHeight As Integer

        ValidHeight = CurrentBoard.Height
        If (Parameters.BottomRowPutBug) Then
            ValidHeight -= 1
        End If

        With OOP
            .Vector = ReadOOPDirection(.ThingData, Instruction)
            If (.Vector.X = 0) AndAlso (.Vector.Y = 0) Then
                OOPError(.ThingIndex, "Bad #PUT")
            Else
                If (EvalKind(.ThingData, Instruction, .Kind)) Then
                    .TargetX = .Vector.X + .ThingData.X
                    .TargetY = .Vector.Y + .ThingData.Y
                    If (.TargetX >= 0) AndAlso (.TargetX < CurrentBoard.Width) Then
                        If (.TargetY >= 0) AndAlso (.TargetY < ValidHeight) Then
                            If (Not ElementAt(.TargetX, .TargetY).Floor) Then
                                Push(.TargetX, .TargetY, .Vector)
                            End If
                            PlotTile(.TargetX, .TargetY, New Tile(.Kind.Element, .Kind.Color))
                        End If
                    End If
                Else
                    OOPError(.ThingIndex, "Bad #PUT")
                End If
            End If
        End With
    End Sub

    Private Sub ExecuteOOPChange(OOP As OOPState, ByRef Instruction As Integer)
        Dim SourceKind As New Kind
        Dim TargetKind As New Kind
        Dim X As Integer = -1
        Dim Y As Integer = 0

        With OOP
            If (Not EvalKind(.ThingData, Instruction, SourceKind)) Then
                OOPError(.ThingIndex, "Bad #CHANGE")
            End If
            If (Not EvalKind(.ThingData, Instruction, TargetKind)) Then
                OOPError(.ThingIndex, "Bad #CHANGE")
            End If
            If (TargetKind.Color = 0) Then
                If (TargetKind.Element.Color < &HF0) Then
                    TargetKind.Color = TargetKind.Element.Color
                End If
            End If
            Do While (IsAny(SourceKind, X, Y))
                PlotTile(X, Y, New Tile(TargetKind.Element, TargetKind.Color))
            Loop
        End With
    End Sub

    Private Sub ExecuteOOPPlay(OOP As OOPState, ByRef Instruction As Integer)
        With OOP
            PlayMusic(ReadOOPLine(.ThingData, Instruction))
            .AutoNextLine = False
        End With
    End Sub

    Private Sub ExecuteOOPCycle(OOP As OOPState, ByRef Instruction As Integer)
        With OOP
            .OOPNumber = ReadOOPNumber(.ThingData, Instruction)
            If (.OOPNumber > 0) Then
                .ThingData.Cycle = .OOPNumber
            End If
        End With
    End Sub

    Private Sub ExecuteOOPChar(OOP As OOPState, ByRef Instruction As Integer)
        With OOP
            .OOPNumber = ReadOOPNumber(.ThingData, Instruction)
            If (.OOPNumber >= 0) Then
                .ThingData.P1 = .OOPNumber
                UpdateAt(.ThingData.X, .ThingData.Y)
            End If
        End With
    End Sub

    Private Sub ExecuteOOPDie(OOP As OOPState, ByRef Instruction As Integer)
        With OOP
            .Died = True
            .TargetTileData = New Tile(Library.GetElementFromID(ElementID.Empty), &HF)
        End With
    End Sub

    Private Sub ExecuteOOPBind(OOP As OOPState, ByRef Instruction As Integer)
        Dim SearchIndex As Integer

        SearchIndex = 0
        With OOP
            .OOPWord = ReadOOPWord(.ThingData, Instruction)
            If (EvalSend(.ThingIndex, SearchIndex, .OOPWord)) Then
                .ThingData.Code = CurrentBoard.Things(SearchIndex).Code
                .ThingData.Length = CurrentBoard.Things(SearchIndex).Length
            Else
                OOP = OOP
            End If
            Instruction = 0
        End With
    End Sub

    Private Sub ExecuteOOPUserCommand(OOP As OOPState, ByRef Instruction As Integer)
        With OOP
            If (Not SendLabel(.ThingIndex, .OOPWord, False)) Then
                If (Not .OOPWord.Contains(":")) Then
                    OOPError(.ThingIndex, "Bad command " & .OOPWord)
                End If
            Else
                .AutoNextLine = False
            End If
        End With
    End Sub

    Private Function OOPSendSearch(Sender As Integer, SearchLabel As String, ByRef Index As Integer, ByRef Instruction As Integer, LabelPrefix As String) As Boolean
        Dim SearchLabelCopy As String
        Dim LabelPrefixCopy As String
        Dim SplitterOffset As Integer
        Dim Target As String = ""
        Dim Label As String
        Dim Success As Boolean

        SearchLabelCopy = SearchLabel
        LabelPrefixCopy = LabelPrefix
        Success = False
        SplitterOffset = SearchLabelCopy.IndexOf(":")
        Label = ""

        If (SplitterOffset > 0) Then
            Target = SearchLabelCopy.Substring(0, SplitterOffset)
            Label = SearchLabelCopy.Substring(SplitterOffset + 1)
            Success = EvalSend(Sender, Index, Target)
        Else
            If (Index < Sender) Then
                Label = SearchLabelCopy
                SplitterOffset = 0
                Success = True
                Index = Sender
            End If
        End If

        ' ZZT has a very strange loop structure here
        ' possibly caused by compiler optimization
        Do
            If (Success) Then
                If (Label = OOPCodeSet(KnownCommand.Restart)) Then
                    Instruction = 0
                Else
                    Instruction = OOPSearch(Index, LabelPrefixCopy & Label)
                    If (Instruction < 0) AndAlso (SplitterOffset > 0) Then
                        Success = EvalSend(Sender, Index, Target)
                        Continue Do
                    End If
                End If
                Success = (Instruction >= 0)
            End If
            Exit Do
        Loop

        Return Success
    End Function

    Private Function EvalSend(SourceIndex As Integer, ByRef SearchIndex As Integer, TargetString As String) As Boolean
        Dim Success As Boolean = False
        Dim Code As String

        SearchIndex += 1
        Select Case TargetString
            Case OOPCodeSet(KnownCommand.All)
                Success = (SearchIndex < CurrentBoard.Things.Count)
            Case OOPCodeSet(KnownCommand.Others)
                If (SearchIndex < CurrentBoard.Things.Count) Then
                    If (SearchIndex = SourceIndex) Then
                        SearchIndex += 1
                        Success = (SearchIndex < CurrentBoard.Things.Count)
                    Else
                        Success = True
                    End If
                End If
            Case OOPCodeSet(KnownCommand.Self)
                If (SourceIndex > 0) Then
                    If (SearchIndex <= SourceIndex) Then
                        SearchIndex = SourceIndex
                        Success = True
                    End If
                End If
            Case Else
                Do Until Success
                    If (SearchIndex >= CurrentBoard.Things.Count) Then
                        Exit Do
                    End If
                    With CurrentBoard.Things(SearchIndex)
                        If (.Code IsNot Nothing) Then
                            Code = CurrentBoard.Things(SearchIndex).Code.Code.ToUpper
                            If (Code.StartsWith("@")) Then
                                If (ReadOOPWord(CurrentBoard.Things(SearchIndex), 1) = TargetString) Then
                                    Success = True
                                End If
                            End If
                        End If
                    End With
                    If (Not Success) Then
                        SearchIndex += 1
                    End If
                Loop
        End Select

        Return Success
    End Function

    Private Function OOPSearch(ByVal Index As Integer, ByVal SearchString As String) As Integer
        Dim SearchIndex As Integer
        Dim Code As String = ""

        SearchString = SearchString.ToUpper

        With CurrentBoard.Things(Index)
            If (.Length <= 0) Then
                Return -1
            End If
            If (.Code IsNot Nothing) Then
                Code = CurrentBoard.Things(Index).Code.Code.ToUpper & Chr(13)
            End If
        End With

        Do
            SearchIndex = Code.IndexOf(SearchString, SearchIndex)
            If (SearchIndex < 0) Then
                Exit Do
            End If
            Select Case Code.Substring(SearchIndex + SearchString.Length, 1)
                Case "A" To "Z", "_"
                    SearchIndex += 1
                    Continue Do
                Case Else
                    Exit Do
            End Select
        Loop

        Return SearchIndex
    End Function

    Protected Function SendLabel(Sender As Integer, InputLabel As String, Force As Boolean) As Boolean
        Dim ModifiedSelf As Boolean
        Dim Label As String
        Dim SearchOffset As Integer = 0
        Dim ThingIndex As Integer = 0
        Dim NotSelf As Boolean
        Dim Locked As Boolean

        Label = InputLabel
        If (Sender < 0) Then
            Sender = -Sender
            NotSelf = True
        Else
            NotSelf = False
        End If

        ModifiedSelf = False
        ThingIndex = 0

        Do Until (Not OOPSendSearch(Sender, Label, ThingIndex, SearchOffset, Chr(13) & ":"))
            With CurrentBoard.Things(ThingIndex)
                If (Parameters.MultiMovement) Then
                    Locked = (.P3 <> 0)
                Else
                    Locked = (.P2 <> 0)
                End If
                'If (Not Locked) OrElse (Force = True) OrElse (Results(Index).Target.Equals(Sender) AndAlso (NotSelfIfLocked = False)) Then
                If (Not Locked) OrElse (Force) OrElse ((Sender = ThingIndex) AndAlso (Not NotSelf)) Then
                    If (ThingIndex = Sender) Then
                        ModifiedSelf = True
                    End If
                    .CurrentInstruction = SearchOffset
                End If
            End With
        Loop

        Return ModifiedSelf
    End Function

End Class
