﻿' RotonScroll
'
' Contains interface for displaying and interacting with scrolls.

Imports System.Windows.Forms

Public Interface Scroll
    ReadOnly Property Visible As Boolean
    Property Title As String
    Property Text() As String
    ReadOnly Property LineIndex As Integer
    Sub Open(ES As EnvironmentGameState, EP As EnvironmentParameters)
    Function Update() As Boolean
    Function Animating() As Boolean
    Sub CancelClose()
    Sub Reset()
End Interface

Public Class ScrollNull
    Implements Scroll

    Public Sub Open(ES As EnvironmentGameState, EP As EnvironmentParameters) Implements Scroll.Open
        Return
    End Sub

    Public Property Text As String Implements Scroll.Text
        Get
            Return ""
        End Get
        Set(ByVal value As String)
            Return
        End Set
    End Property

    Public Property Title As String Implements Scroll.Title
        Get
            Return ""
        End Get
        Set(ByVal value As String)
            Return
        End Set
    End Property

    Public ReadOnly Property Visible As Boolean Implements Scroll.Visible
        Get
            Return False
        End Get
    End Property

    Public Function Update() As Boolean Implements Scroll.Update
        Return False
    End Function

    Public ReadOnly Property LineIndex As Integer Implements Scroll.LineIndex
        Get
            Return -1
        End Get
    End Property

    Public Function Animating() As Boolean Implements Scroll.Animating
        Return False
    End Function

    Public Sub CancelClose() Implements Scroll.CancelClose
        Return
    End Sub

    Public Sub Reset() Implements Scroll.Reset
        Return
    End Sub
End Class

Public Class ScrollVideo
    Implements Scroll

    Private Shared ReadOnly ScrollCharsTop() As Byte = {
        &HC6, &HD1, &HCD, &HD1, &HB5
    }

    Private Shared ReadOnly ScrollCharsMid() As Byte = {
        &H20, &HB3, &H20, &HB3, &H20
    }

    Private Shared ReadOnly ScrollCharsSplit() As Byte = {
        &H20, &HC6, &HCD, &HB5, &H20
    }

    Private Shared ReadOnly ScrollCharsBottom() As Byte = {
        &HC6, &HCF, &HCD, &HCF, &HB5
    }

    Private _Visible As Boolean
    Private _Text As String
    Private _Title As String
    Private _Message() As String
    Private _MessageLength As Integer
    Private _State As EnvironmentGameState
    Private _Parameters As EnvironmentParameters
    Private _Index As Integer
    Private _X As Integer
    Private _Y As Integer
    Private _CenterX As Integer
    Private _CenterY As Integer
    Private _Width As Integer
    Private _Height As Integer
    Private _Label As String
    Private _Video As Video
    Private _FrameTopY As Integer
    Private _FrameBottomY As Integer
    Private _Opening As Boolean
    Private _Closing As Boolean
    Private _Draw As Boolean

    Private Shared CP437 As System.Text.Encoding = System.Text.Encoding.GetEncoding(437)

    Public Sub New(ByVal V As Video)
        _Video = V
    End Sub

    Public Sub Open(ES As EnvironmentGameState, EP As EnvironmentParameters) Implements Scroll.Open
        _State = ES
        _Parameters = EP
        _Message = _Text.Split({Chr(13)})
        _MessageLength = _Message.Length
        If (_Message(_Message.Length - 1) = "") Then
            _MessageLength -= 1
        End If
        _X = (_Parameters.ViewWidth - _Parameters.ScrollWidth) \ 2
        _Y = (_Parameters.ViewHeight - _Parameters.ScrollHeight) \ 2
        _Width = _Parameters.ScrollWidth
        _Height = _Parameters.ScrollHeight
        _CenterX = _X + (_Width \ 2)
        _CenterY = _Y + 2 + ((_Height - 3) \ 2)
        _Index = 0
        _FrameTopY = _Height \ 2
        _FrameBottomY = (_Height - _FrameTopY) - 1
        _Opening = True
        _Visible = True
    End Sub

    Public Property Text As String Implements Scroll.Text
        Get
            Return _Text
        End Get
        Set(ByVal value As String)
            _Text = value
        End Set
    End Property

    Public Property Title As String Implements Scroll.Title
        Get
            Return _Title
        End Get
        Set(ByVal value As String)
            _Title = value
        End Set
    End Property

    Public Function Update() As Boolean Implements Scroll.Update
        Dim Updated As Boolean = False

        _Draw = False

        If (_Visible) Then

            If (_Opening) Then
                _Draw = True
                If (_FrameTopY = 0) Then
                    _Opening = False
                    DrawFrame()
                    DrawContents()
                    Updated = True
                Else
                    _FrameTopY -= 1
                    _FrameBottomY += 1
                    DrawFrame()
                    Return True
                End If
            ElseIf (_Closing) Then
                _Draw = True
                EraseOverlayRow(_FrameTopY)
                EraseOverlayRow(_FrameBottomY)
                _FrameTopY += 1
                _FrameBottomY -= 1
                If (_FrameTopY >= _FrameBottomY) Then
                    _Closing = False
                    _Visible = False
                    _Video.ClearOverlay()
                Else
                    DrawFrame()
                End If
                Return True
            End If

            If (Not _Closing) AndAlso (Not _Opening) Then
                Select Case CType(_State.KeyPressed, Windows.Forms.Keys)
                    Case Keys.Up
                        _Index -= 1
                        Updated = True
                    Case Keys.Down
                        _Index += 1
                        Updated = True
                    Case Keys.Escape
                        _Index = -1
                        _Label = ""
                        _Closing = True
                    Case Keys.Enter
                        _Label = ""
                        _Closing = True
                    Case Keys.PageUp
                        _Index -= _Height - 5
                        Updated = True
                    Case Keys.PageDown
                        _Index += _Height - 5
                        Updated = True
                End Select
                If (_Index >= _MessageLength) Then
                    _Index = _MessageLength - 1
                ElseIf (Not _Closing) AndAlso (_Index < 0) Then
                    _Index = 0
                End If
                If (Updated) Then
                    DrawContents()
                End If
            End If
        End If
        Return Updated
    End Function

    Private Sub DrawFrame()
        Dim Y As Integer

        If (_Opening) Then
            For Y = _FrameTopY + 1 To _FrameBottomY - 1
                DrawRow(Y, ScrollCharsMid)
            Next
            DrawRow(_FrameTopY, ScrollCharsTop)
            DrawRow(_FrameBottomY, ScrollCharsBottom)
        ElseIf (_Closing) Then
            DrawRow(_FrameTopY, ScrollCharsTop)
            DrawRow(_FrameBottomY, ScrollCharsBottom)
        Else
            DrawRow(0, ScrollCharsTop)
            DrawRow(1, ScrollCharsMid)
            DrawRow(2, ScrollCharsSplit)
            For Y = 3 To _Height - 2
                DrawRow(Y, ScrollCharsMid)
            Next
            DrawRow(_Height - 1, ScrollCharsBottom)
        End If
    End Sub

    Private Sub DrawContents()
        Dim StartX As Integer
        Dim EndX As Integer
        Dim StartY As Integer
        Dim EndY As Integer
        Dim BackColor As AnsiChar
        Dim X As Integer
        Dim Y As Integer
        Dim I As Integer
        Dim MessageString As String
        Dim DisplayTitle As String

        StartY = (_Y + 3)
        EndY = (_Y + _Height - 2)
        StartX = (_X + 2)
        EndX = (_X + _Width - 3)
        BackColor = New AnsiChar(32, &H1E)

        For X = StartX To EndX
            For Y = StartY To EndY
                _Video.PlotOverlay(X, Y, BackColor)
            Next
            _Video.PlotOverlay(X, _Y + 1, BackColor)
        Next

        _Video.PlotOverlay(StartX, _CenterY, New AnsiChar(175, &H1C))
        _Video.PlotOverlay(EndX, _CenterY, New AnsiChar(174, &H1C))

        For Y = StartY To EndY
            I = (Y - _CenterY) + _Index
            Select Case I
                Case Is < -1, Is > _MessageLength
                    'write nothing
                Case -1, _MessageLength
                    'dividers
                    For X = StartX + 4 To EndX Step 5
                        _Video.PlotOverlay(X, Y, New AnsiChar(&H7, &H1E))
                    Next
                Case Else
                    If (_Message(I) <> "") Then
                        Select Case _Message(I).Substring(0, 1)
                            Case "$"
                                MessageString = _Message(I).Substring(1)
                                _Video.WriteOverlay(_CenterX - (MessageString.Length \ 2), Y, MessageString, &H1F)
                            Case "!"
                                MessageString = _Message(I).Substring(1)
                                _Video.PlotOverlay(StartX + 4, Y, New AnsiChar(16, &H1D))
                                _Video.WriteOverlay(StartX + 6, Y, MessageString.Substring(MessageString.IndexOf(";") + 1), &H1F)
                            Case Else
                                _Video.WriteOverlay(StartX + 2, Y, _Message(I), &H1E)
                        End Select

                    End If
            End Select
        Next

        MessageString = _Message(_Index)
        If (MessageString <> "") AndAlso (MessageString.Substring(0, 1) = "!") Then
            DisplayTitle = CP437.GetChars({174}) & _Parameters.PressEnterMessage & CP437.GetChars({175})
        Else
            DisplayTitle = _Title
        End If
        _Video.WriteOverlay(_CenterX - (DisplayTitle.Length \ 2), _Y + 1, DisplayTitle, &H1E)
    End Sub

    Private Sub DrawRow(ByVal Y As Integer, ByVal Graphics() As Byte)
        Dim X As Integer
        Y += _Y
        _Video.PlotOverlay(_X, Y, New AnsiChar(Graphics(0), &HF))
        _Video.PlotOverlay(_X + 1, Y, New AnsiChar(Graphics(1), &HF))
        For X = _X + 2 To (_X + _Width) - 3
            _Video.PlotOverlay(X, Y, New AnsiChar(Graphics(2), &HF))
        Next
        _Video.PlotOverlay((_X + _Width) - 2, Y, New AnsiChar(Graphics(3), &HF))
        _Video.PlotOverlay((_X + _Width) - 1, Y, New AnsiChar(Graphics(4), &HF))
    End Sub

    Private Sub EraseOverlayRow(ByVal Y As Integer)
        Dim X As Integer
        Dim XW As Integer = (_X + _Width) - 1
        Dim ClearChar As AnsiChar = New AnsiChar(0, 0)
        For X = _X To XW
            _Video.PlotOverlay(X, Y + _Y, ClearChar)
        Next
    End Sub

    Public ReadOnly Property Visible As Boolean Implements Scroll.Visible
        Get
            Return _Visible
        End Get
    End Property

    Public ReadOnly Property LineIndex As Integer Implements Scroll.LineIndex
        Get
            Return _Index
        End Get
    End Property

    Public Function Animating() As Boolean Implements Scroll.Animating
        Return _Draw
    End Function

    Public Sub CancelClose() Implements Scroll.CancelClose
        _Closing = False
        _Index = 0
    End Sub

    Public Sub Reset() Implements Scroll.Reset
        _Closing = False
        _Opening = False
        _Visible = False
        _Index = -1
        If (_Video IsNot Nothing) Then
            _Video.ClearOverlay()
        End If
    End Sub
End Class
