﻿' RotonSound
'
' Contains all internal sound and music processing code.



Partial Public Class Environment
    Protected Sub PlayMusic(ByVal MusicString As String)
        PlaySound(-1, ProcessMusic(MusicString))
    End Sub

    Protected Sub PlaySound(ByVal Priority As Integer, ByVal Sound() As Byte)
        Dim Index As Integer = 0
        Dim Max As Integer = Sound.Length - 1

        If (State.DisableSound) Then
            Exit Sub
        End If

        If (Not State.SoundPlaying) OrElse (Priority = -1) OrElse ((State.SoundPriority <> -1) AndAlso (Priority >= State.SoundPriority)) Then
            If (State.SoundPriority >= 0) Then
                State.SoundPriority = Priority
                If (Priority >= 0) Then
                    State.SoundData.Clear()
                End If
            End If
            Do While (Index < Max) AndAlso (State.SoundData.Count < 256)
                State.SoundData.Enqueue(Sound(Index))
                State.SoundData.Enqueue(Sound(Index + 1))
                Index += 2
            Loop
            State.SoundPlaying = True
        End If
    End Sub

    Protected Sub ProcessSound()
        Dim Sound As Byte
        If (Not State.SoundPlaying) Then
            Exit Sub
        End If
        If (State.SoundTicks > 0) Then
            State.SoundTicks -= 1
            Exit Sub
        End If
        If (State.SoundData.Count = 0) Then
            State.SoundPlaying = False
            State.SoundPriority = 0
            Audio.Stop()
            Exit Sub
        End If
        Sound = State.SoundData.Dequeue
        State.SoundTicks = (State.SoundData.Dequeue * 4) - 1
        If (Not State.GameQuiet) Then
            Select Case Sound
                Case Is < &HF0
                    Audio.PlayTone(Sound)
                Case &HF0 To &HF2, &HF4 To &HF9
                    Audio.PlaySound(Sound And &HF)
                Case Else
                    Audio.Stop()
            End Select
        End If
    End Sub

    Protected Sub PlayerStepSound()
        If (Not State.GameQuiet) AndAlso (Not State.DisableSound) Then
            If (Not State.SoundPlaying) AndAlso (State.SoundData.Count = 0) AndAlso (State.SoundTicks = 0) Then
                Audio.PlaySound(3)
            End If
        End If
    End Sub

    Protected Function ProcessMusic(ByVal InputString As String) As Byte()
        Dim ThisChar As String
        Dim ThisNote As Integer = -1
        Dim Speed As Integer = 1
        Dim Octave As Integer = 3
        Dim Ret As New Queue(Of Byte)

        Do While (InputString.Length > 0)
            ThisChar = InputString.Substring(0, 1).ToUpper
            InputString = InputString.Substring(1)
            Select Case ThisChar
                Case "T" : Speed = 1
                Case "S" : Speed = 2
                Case "I" : Speed = 4
                Case "Q" : Speed = 8
                Case "H" : Speed = 16
                Case "W" : Speed = 32
                Case "." : Speed = (Speed \ 2) * 3
                Case "3" : Speed = (Speed \ 3)
                Case "+"
                    If (Octave < 6) Then
                        Octave += 1
                    End If
                Case "-"
                    If (Octave > 1) Then
                        Octave -= 1
                    End If
                Case "C" : ThisNote = 0
                Case "D" : ThisNote = 2
                Case "E" : ThisNote = 4
                Case "F" : ThisNote = 5
                Case "G" : ThisNote = 7
                Case "A" : ThisNote = 9
                Case "B" : ThisNote = 11
                Case "0" : ThisNote = &HF0
                Case "1" : ThisNote = &HF1
                Case "2" : ThisNote = &HF2
                Case "4" : ThisNote = &HF4
                Case "5" : ThisNote = &HF5
                Case "6" : ThisNote = &HF6
                Case "7" : ThisNote = &HF7
                Case "8" : ThisNote = &HF8
                Case "9" : ThisNote = &HF9
                Case "X" : ThisNote = &HFF
            End Select
            If (ThisNote > -1) Then
                If (InputString.Length > 0) Then
                    Select Case InputString.Substring(0, 1)
                        Case "!"
                            ThisNote -= 1
                        Case "#"
                            ThisNote += 1
                    End Select
                End If
                If (ThisNote < &HF0) Then
                    ThisNote += (Octave << 4)
                End If
                Ret.Enqueue(CByte(ThisNote And 255))
                Ret.Enqueue(CByte(Speed And 255))
                ThisNote = -1
            End If
        Loop
        Return Ret.ToArray
    End Function

    Protected Sub StopSound()
        Audio.Stop()
        State.SoundData.Clear()
        State.SoundPlaying = False
        State.SoundTicks = 0
    End Sub

End Class
