﻿' RotonZipFile
'
' Implements a ZIP archive reader which will read into a FileCache for use by Roton.

Imports System.IO
Imports System.Text



Partial Public Class FileCache

    Public Sub ImportZipArchive(Source As Stream)
        Dim ZipFile As Zip
        Dim Index As Integer

        ZipFile = Zip.Read(Source)
        For Index = 0 To ZipFile.Files.Count - 1
            With ZipFile.Files(Index)
                SaveFile(.FileName, .Data)
            End With
        Next
    End Sub

    Private Enum ZipCompressionMethod As Int16
        Stored = 0
        Shrunk = 1
        Reduced1 = 2
        Reduced2 = 3
        Reduced3 = 4
        Reduced4 = 5
        Imploded = 6
        Tokenized = 7
        Deflated = 8
        EnhancedDeflate = 9
        IBMTerseOld = 10
        Reserved11 = 11
        BZip2 = 12
        Reserved13 = 13
        LZMA = 14
        Reserved15 = 15
        Reserved16 = 16
        Reserved17 = 17
        IBMTerseNew = 18
        IBMLZ77 = 19
        WavPack = 97
        PPMd = 98
    End Enum

    Private Structure ZipLocalFile
        Private Shared Encoding437 As Encoding = Encoding.GetEncoding(437)
        Private Shared EncodingUTF8 As Encoding = Encoding.UTF8

        Public Version As Int16
        Public Flags As Int16
        Public Compression As ZipCompressionMethod
        Public LastModifiedTime As Int16
        Public LastModifiedDate As Int16
        Public Crc32 As Int32
        Public CompressedSize As Int32
        Public UncompressedSize As Int32
        Public FileName As String
        Public ExtraField() As Byte
        Public Data() As Byte

        Public Overrides Function ToString() As String
            Return "ZipLocalFile: " & FileName & "; " & UncompressedSize & " bytes"
        End Function

        Public Property FlagBit(ByVal Index As Integer) As Boolean
            Get
                Dim BitValue As Integer
                BitValue = (1 << Index)
                Return ((Flags And BitValue) <> 0)
            End Get
            Set(ByVal value As Boolean)
                Dim BitValue As Int16
                Dim MaskValue As Int16
                BitValue = CType(1 << Index, Int16)
                MaskValue = CType(-1, Int16) Xor BitValue
                Flags = Flags And MaskValue
                If (value) Then
                    Flags = Flags Or BitValue
                End If
            End Set
        End Property

        Public Property FlagsEncrypted As Boolean
            Get
                Return FlagBit(0)
            End Get
            Set(ByVal value As Boolean)
                FlagBit(0) = value
            End Set
        End Property

        Public Property FlagsCompression1 As Boolean
            Get
                Return FlagBit(1)
            End Get
            Set(ByVal value As Boolean)
                FlagBit(1) = value
            End Set
        End Property

        Public Property FlagsCompression2 As Boolean
            Get
                Return FlagBit(2)
            End Get
            Set(ByVal value As Boolean)
                FlagBit(2) = value
            End Set
        End Property

        Public Property FlagsDataDescriptorFollows As Boolean
            Get
                Return FlagBit(3)
            End Get
            Set(ByVal value As Boolean)
                FlagBit(3) = value
            End Set
        End Property

        Public Property FlagsEnhancedDeflate As Boolean
            Get
                Return FlagBit(4)
            End Get
            Set(ByVal value As Boolean)
                FlagBit(4) = value
            End Set
        End Property

        Public Property FlagsCompressedPatch As Boolean
            Get
                Return FlagBit(5)
            End Get
            Set(ByVal value As Boolean)
                FlagBit(5) = value
            End Set
        End Property

        Public Property FlagsStrongEncryption As Boolean
            Get
                Return FlagBit(6)
            End Get
            Set(ByVal value As Boolean)
                FlagBit(6) = value
            End Set
        End Property

        Public Property FlagsUseUTF8 As Boolean
            Get
                Return FlagBit(11)
            End Get
            Set(ByVal value As Boolean)
                FlagBit(11) = value
            End Set
        End Property

        Public Property FlagsEnhancedCompression As Boolean
            Get
                Return FlagBit(12)
            End Get
            Set(ByVal value As Boolean)
                FlagBit(12) = value
            End Set
        End Property

        Public Property FlagsEncryptedDirectory As Boolean
            Get
                Return FlagBit(13)
            End Get
            Set(ByVal value As Boolean)
                FlagBit(13) = value
            End Set
        End Property

        Public Sub New(ByVal Source As Stream)
            Dim Reader As New BinaryReader(Source)
            Dim FileNameLength As Int16
            Dim ExtraFieldLength As Int16
            Dim Enc As Encoding = Encoding437

            Version = Reader.ReadInt16
            Flags = Reader.ReadInt16
            Compression = CType(Reader.ReadInt16, ZipCompressionMethod)
            LastModifiedTime = Reader.ReadInt16
            LastModifiedDate = Reader.ReadInt16
            Crc32 = Reader.ReadInt32
            CompressedSize = Reader.ReadInt32
            UncompressedSize = Reader.ReadInt32
            FileNameLength = Reader.ReadInt16
            ExtraFieldLength = Reader.ReadInt16

            If (FlagsUseUTF8) Then
                Enc = EncodingUTF8
            End If

            If (FileNameLength > 0) Then
                FileName = Enc.GetString(Reader.ReadBytes(FileNameLength))
            Else
                FileName = ""
            End If

            If (ExtraFieldLength > 0) Then
                ExtraField = Reader.ReadBytes(ExtraFieldLength)
            Else
                ExtraField = {}
            End If

            Decode(Source)
        End Sub

        Public Shared Function Read(ByVal Source As Stream) As ZipLocalFile
            Return New ZipLocalFile(Source)
        End Function

        Public Sub Write(ByVal Target As Stream)
            Dim Writer As BinaryWriter
            Dim FileNameLength As Int16
            Dim ExtraFieldLength As Int16
            Dim FileNameData() As Byte
            Dim Enc As Encoding = Encoding437

            If (FlagsUseUTF8) Then
                Enc = EncodingUTF8
            End If

            FileNameData = Enc.GetBytes(FileName)
            FileNameLength = CType(FileNameData.Length, Int16)
            ExtraFieldLength = CType(ExtraField.Length, Int16)

            Writer = New BinaryWriter(Target)
            Writer.Write(Version)
            Writer.Write(Flags)
            Writer.Write(CType(Compression, Int16))
            Writer.Write(LastModifiedTime)
            Writer.Write(LastModifiedDate)
            Writer.Write(Crc32)
            Writer.Write(CompressedSize)
            Writer.Write(UncompressedSize)
            Writer.Write(FileNameLength)
            Writer.Write(ExtraFieldLength)
            If (FileNameLength > 0) Then
                Writer.Write(FileNameData)
            End If
            If (ExtraFieldLength > 0) Then
                Writer.Write(ExtraField)
            End If

            Writer.Flush()
        End Sub

        Private Sub Decode(ByVal Source As Stream)
            Dim Reader As BinaryReader

            Select Case Compression
                Case ZipCompressionMethod.Stored
                    Reader = New BinaryReader(Source)
                    Data = Reader.ReadBytes(UncompressedSize)
                Case ZipCompressionMethod.Deflated
                    DecodeDeflate(Source)
                Case Else
                    Reader = New BinaryReader(Source)
                    Data = Reader.ReadBytes(CompressedSize)
            End Select
        End Sub

        Private Sub DecodeDeflate(ByVal Source As Stream)
            Dim Mem As MemoryStream
            Dim Decoder As Compression.DeflateStream
            Dim Bytes() As Byte

            ReDim Bytes(CompressedSize - 1)
            Source.Read(Bytes, 0, CompressedSize)
            Mem = New MemoryStream(Bytes)
            Decoder = New Compression.DeflateStream(Mem, IO.Compression.CompressionMode.Decompress)

            ReDim Data(UncompressedSize - 1)
            Decoder.Read(Data, 0, UncompressedSize)
            Decoder.Close()
            Mem.Dispose()
        End Sub
    End Structure

    Private Class Zip
        Public Files As New List(Of ZipLocalFile)

        Public Shared Function Read(ByVal Source As Stream) As Zip
            Dim ThisFile As New ZipLocalFile
            Dim Ret As New Zip
            Dim ID As Int32
            Dim Reader As New BinaryReader(Source)

            ThisFile.Version = 1

            Do Until (Source.Position >= Source.Length)
                ID = Reader.ReadInt32

                Select Case ID
                    Case &H4034B50 'file
                        ThisFile = ZipLocalFile.Read(Source)
                        Ret.Files.Add(ThisFile)
                    Case Else
                        Exit Do
                End Select
            Loop

            Return Ret
        End Function

        Public Sub Write(ByVal Target As Stream)

        End Sub

        Public Function GetFileStreamByName(ByVal FileName As String) As Stream
            Dim Index As Integer
            Dim Selected As Integer = -1

            FileName = FileName.ToUpper

            For Index = 0 To Files.Count - 1
                If (Files(Index).FileName.ToUpper = FileName) Then
                    Selected = Index
                    Exit For
                End If
            Next

            If (Selected >= 0) Then
                Return New MemoryStream(Files(Selected).Data)
            End If

            Return Nothing
        End Function
    End Class

End Class



