﻿' RotonFileCache
'
' Implements a cache of files. If files are absent from the cache, they are loaded from disk.

Imports System.IO



Public Class FileCacheEntry
    Public Name As String
    Public Data() As Byte
End Class

Public Class FileCache

    Private Shared TextEncoding As System.Text.Encoding = System.Text.Encoding.GetEncoding(437)
    Public ReadOnly Files As New List(Of FileCacheEntry)
    Public Path As String

    Public Sub New()
        Init()
    End Sub

    Public Sub New(NewPath As String)
        Path = NewPath
        Init()
    End Sub

    Private Sub Init()
        ' initialize
    End Sub

    Public ReadOnly Property FileList As String()
        Get
            Dim Ret As New List(Of String)
            Dim Index As Integer

            For Index = 0 To Files.Count - 1
                Ret.Add(Files(Index).Name)
            Next

            Return Ret.ToArray
        End Get
    End Property

    Private Function FindFileIndex(Name As String) As Integer
        Dim Ret As Integer = -1
        Dim Index As Integer

        Name = Name.ToUpper

        For Index = 0 To Files.Count - 1
            With Files(Index)
                If (.Name.ToUpper = Name) Then
                    Ret = Index
                    Exit For
                End If
            End With
        Next

        Return Ret
    End Function

    Private Function FindFileLocal(Name As String) As String
        Dim Ret As String = ""
        Dim Target As String

        Try
            If (Name.Contains("\") OrElse Name.Contains(":")) Then
                ' to protect against exploits
                Return ""
            End If
            With My.Computer.FileSystem
                Target = .CombinePath(Path, Name)
                If (.FileExists(Target)) Then
                    Ret = Target
                End If
            End With
        Catch ex As Exception
            ' failed to find file
        End Try

        Return Ret
    End Function

    Public Function LoadFile(FileName As String) As Byte()
        Dim Ret As Byte() = {}
        Dim Index As Integer
        Dim Target As String

        Index = FindFileIndex(FileName)
        If (Index >= 0) Then
            Ret = Files(Index).Data
        Else
            Target = FindFileLocal(FileName)
            If (Target <> "") Then
                Ret = My.Computer.FileSystem.ReadAllBytes(Target)
            End If
        End If

        Return Ret
    End Function

    Public Function LoadFileStream(FileName As String) As MemoryStream
        Dim Data As Byte()

        Data = LoadFile(FileName)
        Return New MemoryStream(Data)
    End Function

    Public Sub SaveFile(FileName As String, Bytes() As Byte)
        Dim Entry As New FileCacheEntry

        With Entry
            .Name = FileName
            .Data = Bytes
        End With

        DeleteFile(FileName)
        Files.Add(Entry)
    End Sub

    Public Sub SaveFileStream(FileName As String, SourceStream As Stream, Length As Integer)
        Dim Bytes(Length - 1) As Byte

        SourceStream.Read(Bytes, 0, Length)
        SaveFile(FileName, Bytes)
    End Sub

    Public Sub DeleteFile(FileName As String)
        Dim Index As Integer

        Index = FindFileIndex(FileName)
        If (Index >= 0) Then
            Files.RemoveAt(Index)
        End If
    End Sub

    Public Sub ImportHelpLibrary(Source As Stream, Length As Integer)
        Dim SourceReader As New BinaryReader(Source)
        Dim Names As New List(Of String)
        Dim Offsets As New List(Of Int32)
        Dim ThisName() As Byte
        Dim ThisNameLength As Integer
        Dim Index As Integer
        Dim Process As Boolean
        Dim MemStream As New MemoryStream
        Dim Mem() As Byte
        Dim Reader As BinaryReader
        Dim LineLength As Byte
        Dim ThisLine() As Byte
        Dim LineData As List(Of Byte)
        Dim Count As Integer

        ' buffer the archive
        Mem = SourceReader.ReadBytes(Length)
        MemStream = New MemoryStream(Mem)
        Reader = New BinaryReader(MemStream)

        ' dunno what this is
        Count = Reader.ReadInt16()

        ' file names
        Process = True
        For Index = 0 To 23
            ThisNameLength = Reader.ReadByte
            ThisName = Reader.ReadBytes(50)
            If (ThisNameLength > 0) AndAlso (Index < Count) Then
                ReDim Preserve ThisName(ThisNameLength - 1)
                Names.Add(TextEncoding.GetString(ThisName))
            End If
        Next

        ' file offsets
        For Index = 0 To 23
            Offsets.Add(Reader.ReadInt32)
        Next

        ' load the files
        For Index = 0 To Count - 1
            MemStream.Position = Offsets(Index)
            LineData = New List(Of Byte)
            Do
                LineLength = Reader.ReadByte
                If (LineLength > 0) Then
                    ThisLine = Reader.ReadBytes(LineLength)
                    If (LineLength = 1) AndAlso (ThisLine(0) = &H40) Then
                        Exit Do
                    End If
                    LineData.AddRange(ThisLine)
                End If
                LineData.Add(&HD)
            Loop
            SaveFile(Names(Index), LineData.ToArray)
        Next

        MemStream.Dispose()

    End Sub

End Class

